/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

//#define SHOW_RUNCMD
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "interface/build.h"

/* cteni databaze pro dotazy*/
#include "database/database.h"
#include "database/memsave.h"

#include "cmd.h"
#include "gf.h"
#include "xalloc.h"
#include "options.h"
#include "config.h"
#include "macros.h"

//struct symbol_list *act_list;
//const  char* active_file;
#ifdef SHOW_ACTIVE_FILE
    int show_active_file=1;
#else
    int show_active_file=0;
#endif

extern int test_already_inserted;
extern int remember_strings;

/* sparse-interface */
extern int   verbose_spot;

extern void     list_strings();
extern uint32_t db_list_strings();
extern int      add_predefined_file(const char* file);

char*    db_generator_next_string(int reset);

void mem_find_include         ( const char*  file );
void mem_find_variable_use    ( const char*  identifier );
void mem_find_variable_dec    ( const char*  identifier );
void mem_find_function_dec    ( const char*  identifier );
void mem_find_function_def    ( const char*  identifier );
void mem_find_macro_define    ( const char*  identifier );
void mem_find_macro_use       ( const char*  identifier );
void mem_find_struct_dec      ( const char*  identifier );
void mem_delete_file          ( const char** identifier );
void mem_find_unused_variable ( );
void mem_find_unused_include  ( );
void build_save();

void sparse_interface_run   ( const char* );
void sparse_interface_init  ();
void sparse_interface_clear ();
void consider_file          (const char* str);


/* sparse-build.c */
void  get_structure_content_cmd (/* const char* file, */ const char* ident);

extern const char* spot_config_filename; /* options.c */

//static xalloc_context cmd_context;

static int mem_build;

/* po skonceni bude *char** ident dany identifikator predtim nemusi byt definovano  */

#include "cmd/auxiliary.h"
#include "cmd/functions.h"
#include "cmd/commands.h"


#define HELP_CMD "help"
static void  help_command(const char* command, int unused);
struct COMMAND handlers[] = {
  { "def"    , set_predefined_command , 0, 
    "prepend content of file to beginning of every compiled file", 
    "file with preprocessor directives", 
    "def path_to_file_with_preprocessor_directives", 
    "content of this file shall be prepended to every builded file\n\
\t\thowever you may use different file for every source\
 more info in documentation", 0,1,0},

  { "db-ro"    , set_database_command   , 0, 
    "open database read-only mode", 
    "path to database", 
    0, "db-ro spot.db", 0, 0, 1, 0},

  { "db-rw"    , set_database_command   , 1, 
    "open database read-write-create mode", 
    "path to database", 
    0, "db-rw spot.db", 0, 0,1,0},

  { "ip"          , set_predefined_command , 0, 
    "set directories to be searched for include file", 
    "file with line separated path to directory", 0, 0, 0, 0,0,0},

  { "fl"          , set_files_command      , 1, 
    "files to be included in database", 
    0, "file with lines: path to file:ip for this file:def for this file \n\t\twhere [] means not mandatory", 
    0, 0, 0,1,0},

  { "bf"          , build_file_command     , 0, 
    "add this file to database", 
    "filename", 
    "db-rw", 0, 0, 0,1,0},

  { "uf"          , build_file_command     , 1, 
    "update database record of this file", 
    "filename" , 
    "db-rw", 0, 0, 0,1,0}, 

  { "rf"          , build_file_command     , 2, 
    "remove all records of this file from database", 
    "filename", 
    "db-rw", 0, 0, 0,1,0},

  { "ba"          , build_all_command      , 0, 
    "Build database for all files and save", 0, 
    "fl db-rw", 
    0, "if building more files at once, optimalization better is possible", 0, 0,0},


  { "ua"          , build_all_command      , 1, 
    "Update database for all files and save", 0, 
    "fl db-rw", 0, 0,0,0},

  { "p"          , process_command      , 0, 
    "Process all lines in file list", 0, 
    "fl db-rw", 
    0, "if building database for the first time(no update) use ba-is faster", 0, 0,0},
  
  { "bm"          , build_mem_command      , 0, 
    "Build database for all files, but no save", 0, 
    "fl", 0, 0, 0, 0, 0},

  { "sm"          , save_mem_command , 0, 
    "Save memory build database", 0, 
    "db-rw and bm", 0, 0, 0,0,0},

  { "ls"   , list_strings_command   , 0, 
    "List all strings saved in database", 0, 
    "db-ro or db-rw or bm", 0, 0, 0,0,0},

  { "fc" , get_function_declaration_command , 0, 
    "List all function forward declarations", 
    "identifier", "db-ro or db-rw",0 , 0, 1,0,0},

  { "ff" , get_function_definition_command  , 0, 
    "List all function definitions", 
    "identifier", "db-ro or db-rw", 0,0,  1,0,0},

  { "fu" , get_function_usage_command , 0, 
    "List all function usages",     
    "identifier", 
    "db-ro or db-rw", 0, 0, 1,0,0},

  { "uff" , get_unused_function_def_command, 0, 
    "List all probably unused function definitions", 0, 
    "db-ro or db-rw or bm", 0, 0, 0,0,0},

  { "mf" , get_macro_definition_command , 0, 
    "List all macro definitions", 
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "mu" , get_macro_use_command  , 0, 
    "List all macro usages", 
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "tu" , get_typedef_use_command  , 0, 
    "List all typedef usages", 
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "td" , get_typedef_declaration_command , 0, 
    "List all typedef declarations", 
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "vc" , get_global_var_declaration_command, 0, 
    "List all variable declarations",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "vu" , get_global_var_usage_command, 0, 
    "List all variable usages",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "uv" , get_unused_variable_command, 0, 
    "List all probably unused variables", 0, 
    "db-ro or db-rw or bm", 0, 0, 0,0,0},

  { "sc"  , get_struct_dec_command  , 0, 
    "List all struct forward declarations",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "sf"  , get_struct_def_command  , 0, 
    "List all struct definitions",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "si"  , get_struct_content_command, 0, 
    "List all struct definitions including identifiers content of struct", 
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "su"  , get_struct_use_command  , 0, 
    "List all struct usages",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "sym"  , get_symbol_use_command  , 0, 
    "List symbol usages",
    "identifier", "db-ro or db-rw", 0, 0, 1,0,0},

  { "in"      , get_files_include_command, 0, 
    "List all files including file", 
    "filename", "db-ro or db-rw", 0,0,  0,1,0},

  { "lvd"   , get_local_variable_dec_command    , 0,
    "Local variable declaration", 
    "file line pos", "ro access to file", 0,
    "expects usage of a variable at specified position and returns position of declaratin" ,  0,1,0},

  { "lvu"   , get_local_variable_use_command    , 0,
    "Local variable usage", 
    "file line pos", "ro access to file", 0,
    "expects usage or declaration of a variable at specified position and returns position of usages",  0,1,0},

  { "va"   , get_visible_command    , 0,
    "List all identifiers visible at", 
    "file line pos", "ro access to file", 0,0,  0,1,0},


  { "vla"  , get_visible_command    , 1,
    "List all local identifiers visible at", 
    "file line pos", "ro access to file", 0, 
    "local means declared in this file", 0,1,0},

  { "vfa"  , get_visible_function_command , 0,
    "List all functions visible at", 
    "file line pos", "ro access to file", 0, 0, 0,1,0},

  { "vlfa" , get_visible_function_command , 1, 
    "List all local functions visible at", 
    "file line pos", "ro access to file", 0, 0, 0,1,0},

  { "vma"  , get_visible_macro_command , 0, 
    "List all macros visible at", 
    "file line pos", "ro access to file", 0, 
    0, 0,1,0},

  { "vlma" , get_visible_macro_command , 1, 
    "List all local macros visible at", 
    "file line pos", "ro access to file", 0, 
    "local means defined in this file", 0,1,0},

  { "vva"  , get_visible_variable_command , 0, 
    "List all variables visible at", 
    "file line pos", "ro access to file", 0, 0, 0,1,0},

  { "vlva" , get_visible_variable_command , 1, 
    "List all local variables visible at", 
    "file line pos", "ro access to file", 0, 
    "local means defined in this file", 0,1,0},

  { "cfa" , get_called_by_function_at_command, 0,
    "List all functions called by function at", 
    "file line pos", "ro access to file", 0, 0, 0,1,0},

  { "ufa" , get_used_by_function_at_command, 0,
    "Variables used by this function", 
    "file line pos", "ro access to file", 0, 0, 0,1,0},

  { "bye"  , exit_command, 0 , 
    "Exit program", 0, 0, 0, 0, 0,0,0},

  { "regexp"    , regexp_command, 0, 
    "find strings matching to regular expression", 
    "extended regular expression", 
    "db-ro or db-rw or bm", 0, 0, 0,0,0},

  { "stat"    , stat_command, 0, 
    "Print some statistics", 0, 0, 0, 0, 0,0,0},

  { "verbose"    , verbose_command, 0, 
    "", 0, 0, 0, 0, 0,0,0},

  { "nsi"    , nsi_command, 0, 
    "", 0, 0, 0, 0, 0,0,0},

  { "so"    , sparse_output_command, 0, 
    "prints sparse warnings to file", 
    "filename"
    , 0, 0, 0, 0,0,0},

  { "ns"    , nosave_command   , 0, 
    "doesn't save any information", 
    "together with so gives sparse output consuming less space and faster", 
    0, 0, 0, 0, 1, 0
    },

  {
    "sys-in", system_includes_command, 0,
    "all includes are system includes - useful for kernel building",
    0,
    0,0,0,0,0,0
  },

  { HELP_CMD    , help_command, 0, 
    "print help", 0, 0, 0, 0, 0,0,1},
  { 0 , 0, 0, 0, 0, 0, 0, 0, 0,0},
};
struct COMMAND* opts = (struct COMMAND*) handlers;
int opts_len = (sizeof (handlers) / sizeof (handlers[0]));
#include "cmd/readline.h"

static void help_command(const char* command, int unused)
{
  char *ident;  get_identifier(command,&ident);
  int skip=0;
  if (strlen(ident)>0) skip=1;
  int i;
  printf("usage cspot -command or cspot for interactive mode\n");
  for (i = 0; i < (sizeof (handlers) / sizeof (handlers[0])); i++)
    {
      if (!handlers[i].command)  break;
      if (skip && strcmp(ident,handlers[i].command) ) continue;
      printf("command: %s\n",   handlers[i].command );
      if (handlers[i].doc)         printf("\tdoc: %s\n", handlers[i].doc );
      if (handlers[i].requires   ) printf("\trequires     : %s\n", handlers[i].requires  );
      if (handlers[i].prereq     ) printf("\tprerequisity : %s\n", handlers[i].prereq    );
      if (handlers[i].example    ) printf("\texample      : %s\n", handlers[i].example   );
      if (handlers[i].other_info ) printf("\tother_info   : %s\n", handlers[i].other_info);
    }
}

int run_command(const char* command)
{
  int ret_val=1;
  print_bad_result=1; /* databaze rika zda nalezen ci nenalezen identifikator/zaznam*/

  xalloc_context lc; memset(&lc,0,sizeof(xalloc_context));
  xalloc_context* c = xalloc_get_context();
  xalloc_set_context(&lc);
#ifdef SHOW_RUNCMD
  printf("running command : '%s'\n",command);
#endif
  if (!command) 
    { ret_val=0; goto end; }
  if (!strlen(command)) 
    { ret_val=0; goto end; }

  int i;
  for (i = 0; i < (sizeof (handlers) / sizeof (handlers[0])); i++) {
    if (!handlers[i].command) break;

    int o=strlen(handlers[i].command);
    if (!strncasecmp(handlers[i].command,command,o)
	&& (!command[o] || command[o]==' ')) /* prikaz je stejne dlouhy (prikaz je po prvni mezeru) */
      {
	int j=strlen(handlers[i].command);
	const char* cmd=command+j;
	handlers[i].handler(cmd,handlers[i].extension);
	clean_temp_strings_context();
	goto end;
      }
  }
  ret_val=0;
  
 end:
  xalloc_set_context(&lc); 
  xfree_all();
  xalloc_set_context(c); 
  return ret_val;
}

