/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#define MAX_PATH_LEN 2048

//#define DEFAULT_SPOT_CONFIG_NAME "spot_config"
//#define MY_ERRORS

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>

#include "base/macros.h"

#include "database/database.h"
#include "interface/build.h"

int system_includes_only;

int building;
int updating;
int deleting;

extern int remember_strings;
extern int optimize_struct;
extern int optimize_string;

#include "config.h"
#include "xalloc.h"
#include "cmd.h"

extern const char* additional_includepath[MAX_ADDITIONAL_INCLUDE];
extern const char* additional_system_includepath[MAX_ADDITIONAL_INCLUDE];

extern char* spot_config_file;
extern int   spot_config_file_uk;
extern const char* spot_config_filename;

int   no_system_includes;

struct stat config_sb;
int config_fd;
int init_config(const char* config_file)
{
  end_config();
  spot_config_file_uk=0;
  spot_config_file=0;
  if ((config_fd = open(config_file,O_RDONLY))<0)
  {
    fprintf(stderr,"couldn't open config file '%s'\n", config_file);
    FPRINTF_ERROR;

    exit(1);
  }
  /* file to be mmapped */
  if (fstat(config_fd,&config_sb)<0)
    {
      fprintf(stderr,"coulnd't stat a file\n");
      FPRINTF_ERROR
      exit(2);
    }
  if (!config_sb.st_size) exit(3);

  if (    ( spot_config_file = (char*) mmap (NULL, config_sb.st_size
					     , PROT_READ
					     , MAP_PRIVATE
					     , config_fd, 0) ) == (void*)-1  )
    {
      fprintf(stderr,"couldn't map the file %s\n", config_file);
      FPRINTF_ERROR
      exit(4);
    }
  return 0;
}

int end_config()
{
  if (spot_config_file) munmap(spot_config_file,config_sb.st_size);
  spot_config_file=0;
  if (config_fd)   close(config_fd);
  config_fd=0;
  return 0;
}

void reset_config()
{
  spot_config_file_uk=0;
}

int is_endl(c) 
{
  switch (c)
    {
    case '\n':
      return 1;
    default:
      return 0;
    }
}

void skip_endl() 
{
  for (;'\n'==*(spot_config_file+spot_config_file_uk);spot_config_file_uk++);
}

int is_endname(char c) { return c==':'; }
void skip_endname() { *(spot_config_file+spot_config_file_uk)==':'?spot_config_file_uk++:0; }
int is_blank(char c)
{
  switch (c)
    {
    case ' ':
    case '\t':
      return 1;
    default:
      return 0;
    }
}

void skip_blank()
{
  for (;
       spot_config_file_uk < config_sb.st_size
	 && is_blank(*(spot_config_file+spot_config_file_uk))
	 ; spot_config_file_uk++);
}


int word_length()
{
  int i;
  for (i=0;spot_config_file_uk+i<config_sb.st_size 
	 && ( !is_endname(*(spot_config_file+spot_config_file_uk+i))
	      && !is_endl(*(spot_config_file+spot_config_file_uk+i)) )
	 ; i++);
  return i;
}

int nacti(char* s)
{
  skip_blank(); 
  if (is_endl(*(spot_config_file+spot_config_file_uk))) return 0;

  int  len = word_length(); 
  if (!len) 
    {
      *s=0;
      skip_endname();
      return 0; 
    }

  memcpy(s,spot_config_file+spot_config_file_uk,len);   
  s[len]=0; 
  { int i; for (i=len-1;i>=0&&is_blank(s[i]);i--) s[i]=0; }
  spot_config_file_uk+=len; 
  skip_blank(); 
  skip_endname();
  skip_blank(); 
  return len;
}

#if 0
int match_some(const char* file)
{
  printf("match some ... called\n"); exit(1);
  int file_backup=spot_config_file_uk;
  spot_config_file_uk=0;
  const char* s;
  int res=0; int unused;
  while ((s=next_file(0,&unused)))
  {
    if (!strcmp(s,file))
      {
	res=1;
	goto end;
      }
  }
  res=0;
 end:
  spot_config_file_uk=file_backup;
  return res;
}
#endif

typedef struct {
  int fd;
  char* realfile;
  struct stat sb;
}  mmap_file_ro_t;

static void mumap_file_ro(mmap_file_ro_t* t)
{
  if (t->fd) close(t->fd);
  if (t->realfile) munmap(t->realfile,t->sb.st_size);
}
static mmap_file_ro_t mmap_file_ro(const char* file)
{
  mmap_file_ro_t t;
  if (( t.fd=open(file,O_RDONLY))<0)
    {
      fprintf(stderr,"mmap_file_ro: couldn't open file '%s'\n",file);
      FPRINTF_ERROR;
      exit(1);
    }
  /* file to be mmapped */
  if (fstat(t.fd,&t.sb)<0)
    {
      fprintf(stderr,"coulnd't stat a file %s\n",file);
      FPRINTF_ERROR;
      exit(2);
    }

  if ( !t.sb.st_size) 
    {
      close(t.fd);
      t.fd=0;
      t.realfile=0;
      return t;
    }

  if (    ( t.realfile = (char*) mmap (NULL, t.sb.st_size
				       , PROT_READ
				       , MAP_PRIVATE
				       , t.fd, 0) ) == (void*)-1      )
    {
      fprintf(stderr,"couldn't map the file '%s'\n",file);
      FPRINTF_ERROR;
      exit(3);
    }
  return t;
}

void set_macros_file(const char* m);
static int include_paths(const char* file);
static const char *includes, *macros;
static char *macro_file;

const char* read_macros_file()
{
#ifdef MY_ERRORS
  printf("READ_MACROS_FILE\n");
#endif
  static char* file_copy;
  if (file_copy)
    {
      free(file_copy);
      file_copy=0;
    }

  if (!macros) 
    {
#ifdef MY_ERRORS
      printf("NO FILE\n");
#endif
      file_copy = malloc(2);
      strcpy(file_copy,"\n");
      return file_copy;
    }

  int len;
  mmap_file_ro_t t = mmap_file_ro(macros);
  macro_file=t.realfile; 
  len=t.sb.st_size;

  if (!macro_file)
    {
#ifdef MY_ERRORS
      printf("NOT MAPPED\n");
#endif
      file_copy = malloc(2);
      strcpy(file_copy,"\n");
      mumap_file_ro(&t);
      return file_copy;
    }

#ifdef MY_ERRORS
  printf("EVERYTHING OK\n");
#endif
  file_copy = malloc(len+2);
  strcpy(file_copy,macro_file);
  file_copy[len]='\n';
  file_copy[len+1]=0;

  mumap_file_ro(&t);

  return file_copy;

#if 0
  if (*(macro_file+len-1) != '\n')
    {
      fprintf(stderr,"file %s should end with newline but ends with %X\n"
	      , file
	      , *(macro_file+sb.st_size-1));
      exit(1);
    }

  return macro_file;
#endif
}

void destroy_macros_file()
{
  return;
#if 0
  munmap(macro_file,sb.st_size);
  macro_file=0;
#endif
}

void set_macros_file(const char* m)
{
  if (!m || !*m) macros=0;
  else           macros=m;
}

void set_includes_file(const char* i)
{
  includes=i;
  if (!*i) includes=0;
}


const char* get_macros_file()
{
  return macros;
}

const char* get_includes_file()
{
  return includes;
}

static int include_paths(const char* file)
{
  static xalloc_context ipc;
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&ipc);

#if 0
   xfree_all();
   int j; for (j=0;j<MAX_ADDITIONAL_INCLUDE;j++)
     additional_includepath[j] = 0;

#endif

  if (!file) return 1;

/*     printf("include paths...\n"); */
  char* realfile;

  int len;

  mmap_file_ro_t t =  mmap_file_ro(file); 
  realfile=t.realfile; 
  len=t.sb.st_size;
  if (!realfile) return 0;

  //printf("realfile =%p\n",realfile);

  int i = 0; int j=0;
  for(;i < len;i++)
    {
      if (    *(realfile+i) == ' '
	   || *(realfile+i) == '\n'
	   || *(realfile+i) == '\t'
	  ) continue;
      else 
	{
	  int o=i;
	  for ( ;    *(realfile+o) != ' '
		  && *(realfile+o) != '\n'
		  && *(realfile+o) != '\t'
		  ; o++ );
	  char* s = (char*) xalloc(o-i+1);
	  strncpy(s,realfile+i,o-i);
	  s[o-i]=0;
	  if (!system_includes_only) 
	    additional_includepath[j++] = s;
	  else
	    additional_system_includepath[j++] = s;
	  i=o;
	}
    }
  mumap_file_ro(&t);
  xalloc_set_context(context);

  return 0;

}

void reset_config_source_compilation()
{
  int j; for (j=0;j<MAX_ADDITIONAL_INCLUDE;j++)
    additional_includepath[j++] = 0;
  for (j=0;j<MAX_ADDITIONAL_INCLUDE;j++)
    additional_system_includepath[j++] = 0;
   set_macros_file(0);
}

void config_source_compilation(const char* source
			       , const char* include_path
			       , const char* macro)
{

  reset_config_source_compilation();


  if (!strcmp(source,"*"))
    {
      includes = !strcmp(include_path , "-")?include_path:0;
      macros   = !strcmp(macro        , "-")?macro:0;
      return;
    }

  /* nejsou li definovany lokalne pouzije globalni nastaveni */

  if (!include_path || !*include_path) include_path=includes;
  if (!macro        || !*macro       ) macro=macros;

  
  
  if (include_path &&strcmp(include_path , "-"))
    {
#if 0
      int fd;
      if ((fd = open(include_path,O_RDONLY))<0)
	{
	  include_path=0;
	  fprintf("cannot open file %s\n",include_path);
	} else {
	  struct stat sb;
	  if (fstat(fd,&sb)<0)
	    {
	      include_path=0;
	    }
	  if (!sb.st_size)
	    include_path=0;
	  close(fd);
	}
#endif

#ifdef MY_ERRORS
      printf("include_path='%s'\n",include_path);
#warning CHYBA1
#endif
      if (include_path) include_paths(include_path); 
    }

  if (macro && strcmp(macro , "-"))
    {
#if 0
      int fd;
      if ((fd = open(macro,O_RDONLY))<0)
	{
	  macro=0;
	  fprintf("cannot open file %s\n",macro);
 	} else {
	  struct stat sb;
	  if (fstat(fd,&sb)<0)
	    {
	      macro=0;
	    }
	  if (!sb.st_size)
	    macro=0;
	  close(fd);
	}
#endif

#ifdef MY_ERRORS
      printf("macro='%s'\n",macro);
#warning CHYBA2
#endif
      if (macro) set_macros_file(macro);
    }
}

#include "base/filelist.h"
const char* next_file(int reset, int* command)
{
  if (reset)
    {
      spot_config_file_uk=0;
      return 0;
    }
  if (!spot_config_filename)     return 0;

#ifndef ADVANCED_FILELIST
  static char file[MAX_LEN];
  static char ip  [MAX_LEN];
  static char mac [MAX_LEN];

  char *f, *i=0, *m=0;
  int len; 
  f=file;
  if ( !(len=nacti(file)) )  { return 0; } 
  if (  (len=nacti(ip)  ) )  i=ip;
  if (  (len=nacti(mac) ) )  m=mac;
  config_source_compilation(f,i,m);
  skip_endl();

  if (!strcmp(file,"*")) 
    return next_file(reset);
  return file;
#else

  static file_record r;
  int next_uk;
  next_uk=get_record(spot_config_filename,&r,spot_config_file_uk);
  if (next_uk==-1) return 0;

  if ( (r.todo== build_now && building)
       || (r.todo==update_now && updating  ) 
       || (r.todo==delete_now && deleting  ) 
       || (r.todo==delete_now_recursive && deleting  ) )
    unset_record(spot_config_filename,spot_config_file_uk);

  *command=r.todo;
  spot_config_file_uk=next_uk;
  
#if 0
  int cont = 0;
  if ( r.todo == build_now && building ) cont = 1;
  if ( ( r.todo == update_now || r.todo == update_always ) 
       && updating ) cont = 1;
  if (!cont)
    return next_file(reset,command);
#endif
  
  char *f, *i=0, *m=0;
  f=r.filename; i=r.include; m=r.macro;
  config_source_compilation(f,i,m);

  if (!strcmp(f,"*")) 
    return next_file(reset,command);

  return f;

#endif
  
}

