/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include "base/filelist.h"
//#define ADVANCED_FILELIST

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>

typedef struct {
  const char* filename;
  int fd;
  struct stat sb;
  char* realfile;
  int pos;
} File_t;

static FILE* open_file(const char* file, const char* flags)
{
  char mark=1;
  FILE* f;
  f = fopen(file,flags);
  if (!f)
    {
      int i;
      int l=strlen(file);
      char* s = malloc(l+1);
      strcpy(s,file);
      for (i=l;i>0;i--)
	{
	  if (s[i]=='/')
	    {
	      s[i]=0;
	      if (!mkdir(s,S_IRWXU))
		{
		  printf("directory %s\n",s); 
		  s[i]=mark;
		  goto ok;
		}
	    }
	}
    ko:
      fprintf(stderr, "couldn't create directory %s \n",s);
      return NULL;
    ok:
      printf("OK\n"); 
      for (;i<l;i++)
	{
	  if (s[i]==mark)
	    {
	      s[i]='/';
	      if ((f = fopen(file,flags))) return f;
	      if (  mkdir(s,S_IRWXU)/* mkdir(s,S_IRUSR|S_IWUSR) */  )
		goto ko;
	    }
	}     
      f = fopen(file,flags);
    }
  return f;
}

static int fmapped;
static File_t fmap(const char* filename)
{
  static File_t f;
  //  printf("fmap %s\n",filename);

  if (fmapped) 
    {
      f.pos=0;
      //      printf("skipped\n");
      return f; /* the file was left mapped */
    }
  else fmapped++;

  f.fd=open(filename,O_RDWR);
  f.realfile=NULL;
  f.filename=filename;
  f.pos=0;

  if (f.fd<0)
    {
      fprintf(stderr, "fmap: couldn't open file '%s'\n",filename);
      FPRINTF_ERROR;

      //      abort();
      exit(1);
    }
  if (fstat(f.fd,&f.sb)<0)
    {
      fprintf(stderr, "couldn't stat file '%s'\n",filename);
      FPRINTF_ERROR;
      exit(1);
    }
  if (((f.realfile) = (char*) mmap(f.realfile,
				   f.sb.st_size,
				   PROT_WRITE|PROT_READ,
				   MAP_SHARED,
				   f.fd,
				   0))==(void*)-1)
    {
      fprintf(stderr, "coulnd't map file '%s'\n",filename);
      exit(1);
    }
  if (!f.realfile) 
    {
      fprintf(stderr,"file not mapped !!!\n");
      exit(1);
    }
  return f;
}

static void fumap(File_t* f)
{
  /* i leave the file mapped for next time */
  return; 

  //  printf("fumap\n");
  if ( f->realfile && (munmap(f->realfile,f->sb.st_size) < 0 ))
    {
      fprintf(stderr, "coulnd't unmap file '%s'\n",f->filename);
    }
  f->realfile=NULL;
  close(f->fd);
}

static int read_line(File_t* f, char* line, int len)
{
  for (;f->realfile[f->pos]=='\n';f->pos++);
  for (;f->sb.st_size<f->pos 
	 && ( f->realfile[f->pos]==' ' 
	      || f->realfile[f->pos]=='\t' 
	      || f->realfile[f->pos]=='\n' )
	 ; f->pos++);


  /* at the end of file */
  if (f->sb.st_size<=f->pos) return 0;

  { 
    int i;
    for (i=0;
	 f->sb.st_size>f->pos && f->realfile[f->pos]!='\n'
	   ; f->pos++, i++)
      {
	if (i+1<len)
	  {
	    line[i]=f->realfile[f->pos];
	    line[i+1]=0;
	  }
	else
	  {
	    fprintf(stderr,"record too long at char %d\n",f->pos);
	    exit(1);
	  }
      }
  }
  for (;f->sb.st_size<f->pos 
	 && ( f->realfile[f->pos]==' ' 
	      || f->realfile[f->pos]=='\t' 
	      || f->realfile[f->pos]=='\n')
	 ; f->pos++);  

  return 1;
}

static file_record parse_line(const char* l)
{
  int   len  = strlen(l);
  char* line = malloc(len + 1);
  file_record r;
  int p;

  strcpy(line,l);
  r.todo=not_set;
  if (len<2 )
    {
      fprintf(stderr,"bad line in file list len=%d\n",len);
      exit(1);
    }

  p=1;
  if (line[p] == ':' ) goto advanced_filelist;
  if (line[p] == ' ' ) goto skip_blank;
  if (line[p] == '\t') goto skip_blank;
  goto normal_filelist;

 skip_blank:
  for (;p<len;p++)
    {
      if (line[p]!=' ' && line[p]!='\t')
	{
	  if (line[p]==':') goto advanced_filelist;
	  else goto normal_filelist;
	}
    }
  free(line);
  return r;
  
 advanced_filelist:
  r.todo = line[0];
  p++;

 normal_filelist:
  {
    int f,i,m;
    f=p;

    for (;p<len;p++) { if (line[p]==':') { 
      int k; line[p]=0; 
      for (k=p;k>0;k--) 
	if (line[k]==' ' || line[k]=='\t') line[k]=0;
	else break;
      for (k=p;k<len;k++) 
	if (line[k]==' ' || line[k]=='\t') line[k]=0;
	else break;
      break; } }

    if (p<len) i=p; else i=len-1;
    
    for (;p<len;p++) { if (line[p]==':') { 
      int k; line[p]=0; 
      for (k=p;k>0;k--) 
	if (line[k]==' ' || line[k]=='\t') line[k]=0;
	else break;
      for (k=p;k<len;k++) 
	if (line[k]==' ' || line[k]=='\t') line[k]=0;
	else break;
      break; } }
    
    if (p<len) m=p; else m=len-1;

    strcpy(r.filename,line+f);
    strcpy(r.include,line+i+1);
    strcpy(r.macro,line+m+1);
#if 0
    printf("\n\nfilename '%s'\n",r.filename);
    printf("include '%s'\n",r.include);
    printf("macro '%s'\n\n",r.macro);
#endif
  }
  free(line);
  return r;
}

static void append_record(const char* file, file_record rec)
{
  FILE* spot = open_file(file ,"a");
  if (!spot) { exit(1); }

#ifdef ADVANCED_FILELIST
  fputc(rec.todo,spot);
  fputs(":",spot);
#endif
  fputs(rec.filename,spot);
  fputs(":",spot);
  fputs(rec.include,spot); 
  fputs(":",spot);
  fputs(rec.macro,spot);
  fputs("\n",spot);

  fclose(spot);

}

static int file_exists(const char* filename)
{
  struct stat s;
  int exists=0;
  int fd = open(filename,O_RDONLY);
  if (fd<0) goto end; /* soubor neexistuje - pouzito zrejme pro generovani globalnich .i a .i*/
  if (fstat(fd,&s)) goto end; /* opet soubor nelze spravne otevrit */
  exists=1;
 end:
  close(fd);
  return exists;
}

static int has_record(const char* file, file_record *rec)
{
  File_t f;
  int prev_pos;
  char line[4*MAX_LEN];
  int found=0;

  if (!file_exists(file)) return -1;
  f = fmap(file);
  
  for (prev_pos=0; !found && read_line(&f,line,4*MAX_LEN); prev_pos=f.pos)
    {
      file_record r;
      if (*line)
	 r = parse_line(line);
      else continue;
      if (!strcmp(r.filename,rec->filename))
	{
	  found=1;
	  rec->todo=r.todo;
	  break;
	}

      for (;f.realfile[f.pos]=='\n';f.pos++);
      for (;f.sb.st_size<f.pos 
	     && ( f.realfile[f.pos]==' ' 
		  || f.realfile[f.pos]=='\t' 
		  || f.realfile[f.pos]=='\n' )
	     ; f.pos++);
    }

  fumap(&f);
  return found?prev_pos:-1;
}

static void change_record(const char* file, file_record rec,int start_pos)
{
  File_t f = fmap(file);
  
  f.pos=start_pos;
  for (;f.sb.st_size<f.pos 
	 && ( f.realfile[f.pos]==' ' || f.realfile[f.pos]=='\t' )
	 ; f.pos++);
  f.realfile[f.pos]=rec.todo;
  fumap(&f);
}

int unset_record(const char* file, int start_pos)
{
  File_t f = fmap(file);
  enum TODO todo ;

  f.pos=start_pos;
  for (;f.sb.st_size<f.pos 
	 && ( f.realfile[f.pos]==' ' 
	      || f.realfile[f.pos]=='\t' 
	      || f.realfile[f.pos]=='\n')
	 ; f.pos++);

  todo = f.realfile[f.pos];
  switch (todo)
    {
    case build_now  :
      //      printf("1");
    case update_now :
      //      printf("2");
    case delete_now :
      //      printf("3");
    case delete_now_recursive :

      f.realfile[f.pos]=noaction_now;
      //      printf(" unset ok\n");
      break;
    default:
      ;
      // printf("bad unset ... realfile '%c' at %d\n",f.realfile[f.pos],f.pos);
    }
  fumap(&f);
  return 0;
}

void add_record(const char* file, file_record rec)
{
  int pos;  
  if ( (pos=has_record(file,&rec))==-1 )
    {
      rec.todo=build_now;
      append_record(file,rec);
    }
  else
    {
      switch (rec.todo)
	{
	case not_set :
	  fprintf(stderr,"should always be set --- error?     rec.todo='%c'\n",rec.todo);
	case noaction_now :
	  rec.todo=update_now;
	  change_record(file,rec,pos);
	  break;
	default:
	  ;
	}
    }
}

int get_record(const char* file, file_record* rec, int start_pos)
{
  File_t f = fmap(file);
  char line[4*MAX_LEN];
  
  f.pos = start_pos;
  if (!read_line(&f,line,4*MAX_LEN)) return -1;
  if (f.sb.st_size<=start_pos) return -1;

  if (*line)
    *rec = parse_line(line);
  else 
    {
      printf("oprav to %d\n",f.pos);
      exit(1);
    }
  
  for (;f.sb.st_size<f.pos 
	 && ( f.realfile[f.pos]==' ' 
	      || f.realfile[f.pos]=='\t' 
	      || f.realfile[f.pos]=='\n')
	 ; f.pos++);

  fumap(&f); 
  return f.pos+1; //+1 ... preskoci konec radku
}
