/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */


#ifndef MACROS_H
#define MACROS_H


#define MAX_LEN PAGE_SIZE
#define MAX_PATH_LEN 2048

#define MASK_AUTO	(1<<10)
#define MASK_REGISTER	(1<<11)
#define MASK_STATIC	(1<<12)
#define MASK_EXTERN	(1<<13)

#define MASK_CONST	(1<<14)
#define MASK_VOLATILE	(1<<15)
#define MASK_SIGNED	(1<<16)
#define MASK_UNSIGNED	(1<<17)

#define MASK_NOCAST	(1<<18)
#define MASK_NODEREF	(1<<19)
#define MASK_ACCESSED	(1<<20)
#define MASK_TOPLEVEL	(1<<21)

#define MASK_TYPEDEF	(1<<22)
#define MASK_STRUCTOF	(1<<23)
#define MASK_UNIONOF	(1<<24)
#define MASK_ENUMOF	(1<<25)

#define MASK_TYPEOF	(1<<26)
#define MASK_ATTRIBUTE	(1<<27)
#define MASK_INLINE	(1<<28)
#define MASK_ADDRESSABLE	(1<<29)

#define MASK_DECLARATION        (1<<30)
#define MASK_CAST               (1<<31)
#define MASK_SYSTEMINCLUDE      (1<<31)

/* 
   pos  je ve starse deklarovana pos  : 10
   line je ve starse deklarovana line : 31
   -> zbyvajici bity muzu vyuzit 
*/

#define REAL_POSITION ( ~((1<<31)>>21) ) /* tedy dolnich 10 bitu */
#define REAL_LINE ( ~(1<<31) ) /* tedy dolnich 31 bitu */

#define DEFAULT_CONFIG           "cspot.cnf"
#define DEFAULT_COMPILER         "/usr/bin/gcc"
#define DEFAULT_OUTPUT_DIRECTORY "cspot.dir"
#define DEFAULT_FILELIST         "cspot.files"

//#define MY_WARNS

#include <stdio.h>

#define WRITE2(id,file)\
 { printf("%s\t at `%s'"\
	, id\
	, file ); \
   printf("\n");\
 }

#define WRITE4(id,file,line,pos)\
 { printf("%s\t at `%s' %d:%d"\
	, id\
	, file\
	, line & REAL_LINE \
	, pos  & REAL_POSITION ); \
   WRITE_INFO(pos); WRITE_LINE_INFO(line);\
   printf("\n");\
 }

#define WRITE7(id,file_1,line_1,pos_1,file_2,line_2,pos_2)\
 { printf("%s\t at `%s' %d:%d (dec %s %d:%d)"\
	, id\
	, file_1\
	, line_1  & REAL_LINE \
	, pos_1 & REAL_POSITION \
	, file_2\
	, line_2 & REAL_LINE \
	, pos_2 & REAL_POSITION); \
   WRITE_INFO(pos_2); WRITE_LINE_INFO(line_1); WRITE_LINE_INFO(line_2);\
   printf("\n");\
 }

int print_bad_result;
#define NO_IDENT(ident)\
  if (print_bad_result) printf("no such ident %s\n" ,ident)
#define NO_RESULT(ident)\
  if (print_bad_result) printf("no results for %s\n",ident)

#define WRITE_IDENT(ident)\
  printf("%s\n",ident)

#define INFO_ONE(pos,pos_mask,write)\
  (pos & pos_mask ? printf("%s ",write) : 0)

#define WRITE_INFO(pos)\
{\
  if (pos & ( MASK_REGISTER | MASK_STATIC | MASK_EXTERN | MASK_CONST | MASK_VOLATILE             \
	      | MASK_SIGNED | MASK_ADDRESSABLE | MASK_INLINE | MASK_ATTRIBUTE | MASK_UNSIGNED    \
	      | MASK_NOCAST | MASK_NODEREF | MASK_ACCESSED | MASK_TOPLEVEL | MASK_TYPEDEF        \
	      | MASK_STRUCTOF | MASK_UNIONOF | MASK_ENUMOF | MASK_TYPEOF | MASK_AUTO))           \
      printf("\tflags: ");\
  INFO_ONE(pos, MASK_REGISTER    , "register");    \
  INFO_ONE(pos, MASK_STATIC      , "static");      \
  INFO_ONE(pos, MASK_EXTERN      , "extern");      \
  INFO_ONE(pos, MASK_CONST       , "const");       \
  INFO_ONE(pos, MASK_VOLATILE    , "volatile");    \
  INFO_ONE(pos, MASK_SIGNED      , "signed");      \
/*   INFO_ONE(pos, MASK_ADDRESSABLE , "addressable"); */ \
  INFO_ONE(pos, MASK_INLINE      , "inline");      \
  INFO_ONE(pos, MASK_ATTRIBUTE   , "attribute");   \
  INFO_ONE(pos, MASK_UNSIGNED    , "unsigned");    \
  INFO_ONE(pos, MASK_NOCAST      , "nocast");      \
  INFO_ONE(pos, MASK_NODEREF     , "noderef");     \
  INFO_ONE(pos, MASK_ACCESSED    , "accessed");    \
/*   INFO_ONE(pos, MASK_TOPLEVEL    , "toplevel");    */ \
  INFO_ONE(pos, MASK_TYPEDEF     , "typedef");     \
  INFO_ONE(pos, MASK_STRUCTOF    , "structof");    \
  INFO_ONE(pos, MASK_UNIONOF     , "unionof");     \
  INFO_ONE(pos, MASK_ENUMOF      , "enumof");      \
  INFO_ONE(pos, MASK_TYPEOF      , "typeof");      \
  INFO_ONE(pos, MASK_AUTO        , "auto");        \
}


#define WRITE_LINE_INFO(line)\
{\
  INFO_ONE(line, MASK_SYSTEMINCLUDE , " sys-include"); \
}

#define FPRINTF_ERROR                                                                                             \
	switch (errno)                                                                                            \
	  {                                                                                                       \
	  case EFAULT:  fprintf(stderr,"path points outside your accessible address space."); break;              \
	  case EIO:     fprintf(stderr,"An I/O error occurred."); break;                                          \
	  case ELOOP:   fprintf(stderr,"Too many symbolic links were encountered in resolving path."); break;     \
	  case ENOENT:  fprintf(stderr,"The file does not exist."); break;                                        \
	  case ENOMEM:  fprintf(stderr,"Insufficient kernel memory was available."); break;                       \
	  case ENOTDIR: fprintf(stderr,"A component of path is not a directory."); break;                         \
	  case EACCES:  fprintf(stderr,"Search permission was denied on the directory open on fd."); break;       \
	  case EBADF:   fprintf(stderr,"fd is not a valid file descriptor."); break;                              \
	  case ENAMETOOLONG:      fprintf(stderr,"path is too long."); break;                                     \
	  case EMFILE:  fprintf(stderr,"too many open files"); break;                                             \
	  case ENFILE:  fprintf(stderr,"file table overflow in system"); break;                                   \
	  default: break;                                                                                         \
	  }                                                                                                       \
	fprintf(stderr,"\n");


#define ABORT \
{ \
  fprintf(stderr,"application error!\n"); \
   abort();  exit(1); \
}\

#endif

