/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <limits.h>

#define DEFAULT_SPOT_CONFIG_NAME "spot_config"

/* inside project */
#include "database/database.h"
#include "interface/build.h"

extern int remember_strings;
extern int optimize_struct;
extern int optimize_string;
extern int test_already_inserted;


#define STRING_LENGTH (1<<20)
#define STRUCT_LENGTH (1<<22)

#include "options.h"
#include "config.h"
#include "macros.h"
#include "xalloc.h"
#include "cmd.h"

/* sparse-interface */
extern const char* additional_includepath[MAX_ADDITIONAL_INCLUDE]; 
extern const char* additional_system_includepath[MAX_ADDITIONAL_INCLUDE]; 
extern int   verbose_spot;


static xalloc_context options_context;

static char* database_filename;

char* spot_config_file;
int spot_config_file_uk;
const char* spot_config_filename;


extern int fce_inline_deactivate;
extern int fce_use_deactivate;
extern int fce_def_deactivate;
extern int fce_dec_deactivate;
extern int var_dec_deactivate;
extern int var_use_deactivate;
extern int incl_deactivate;
extern int mac_def_deactivate;
extern int mac_use_deactivate;
extern int mac_defd_deactivate;
extern int mac_udef_deactivate;
extern int str_dec_deactivate;
extern int str_def_deactivate;
extern int str_use_deactivate;
extern int tdef_deactivate;
extern int tdef_use_deactivate;
extern int cast_deactivate;


void set_includes_file(const char* i);
void set_macros_file(const char* m);
/*
  Vrac cestu k aktuln databzi
 */
const char* get_db_file()
{
  return database_filename;
}


void set_db_file(const char* str)
{
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&options_context);
  if (database_filename) xfree(database_filename);
  database_filename=xalloc(strlen(str)+1);
  strcpy(database_filename,str);
  xalloc_set_context(context);
}

char* read_line (FILE* f, char* buffer)
{
 int pos=0;
 char c;
 while((c=fgetc(f)))
   {
     if (c=='\n')
       {
	 buffer[pos]=0;
	 return buffer;
       }
     else if (c==-1)
       {
	 buffer[pos]=0;
	 return buffer;
       }
     else
       {
	 buffer[pos]=c;
	 buffer[pos+1]=0;
       }
     pos++;
     if(pos==MAX_LEN)
       {
	 fprintf(stderr,"buffer too short\n");
	 exit(1);
       }
   }
 return 0;
}


const char* get_spot_config()
{
  return spot_config_filename?spot_config_filename:DEFAULT_SPOT_CONFIG_NAME;
}



void print_help()
{
  run_command ("help");
  exit(0);
}


/*
  prijima argumenty predane programu a podle nich spousti prislusne commandy
  pokud je sousten nejaky aktivni command (build databaze nebo dotaz) 
  pak po dokonceni vyhodnoceni vsech argumentu skonci
 */
int use_args(int argc, char **argv)
{
/*   optimize_struct=0; */
/*   optimize_string=0; */
  test_already_inserted=1;

  memset(&additional_includepath,0,sizeof(void*)*MAX_ADDITIONAL_INCLUDE);
  memset(&additional_system_includepath,0,sizeof(void*)*MAX_ADDITIONAL_INCLUDE);

  verbose_spot=0;
  no_system_includes=0;

#define OPT(x) char* x=0;
#define OPT2(x)  char* x=0; int x ## _line=0; int x ## _pos=0;
  char* executable=*(argv-1);
  OPT(fl)
    OPT(conf)
    OPT(dir)
    OPT(sym)
    OPT(db)
    OPT(db_ro)
    OPT(db_rw)
    OPT(ip)
    OPT(rf)
    OPT(regexp)
    OPT(def)
    OPT(uf)
    OPT(si)
    OPT(ff)
    OPT(fu)
    OPT(fc)
    OPT(mf)
    OPT(mu)
    OPT(vc)
    OPT(vu)
    OPT(td)
    OPT(tu)
    OPT(in)
    OPT(bf)
    OPT(sc)
    OPT(su)
    OPT(sf)
    OPT(so)
    OPT2(lvd)
    OPT2(lvu)
    OPT2(va)
    OPT2(vla)
    OPT2(vfa)
    OPT2(vlfa)
    OPT2(vma)
    OPT2(vlma)
    OPT2(vva)
    OPT2(vlva)
    OPT2(cfa)
    OPT2(ufa)
  int build_all=0;
  int ns = 0;
  int process = 0;
  int build_mem=0;
  int update_all=0;
  int var_unused=0;
  int include_unused=0;
  int ls = 0;
  int stat=0;
  int system_includes_only=0;
  char **act=0;
#undef OPT
#undef OPT2

#define OPT(x)\
      else if (!strcmp(arg, #x ))\
	{\
	  act++;\
          if (!*act) { print_help(); exit(1); }\
	  x=*act;\
	  continue;\
	}
#define OPT2(x)\
      else if (!strcmp(arg, #x ))\
      {\
	act++;\
	x=*act;\
	act++;\
        if (!*act) { print_help(); exit(1); }\
	x ## _line=atoi(*act);\
	act++;\
        if (!*act) { print_help(); exit(1); }\
	x ## _pos=atoi(*act);\
      }
#define DEACTIVATE(x)\
      else if (!strcmp(arg, #x))  { x++; continue; }


  for (act=argv;*act;act++)
    {
      if (!**act) continue; // ""

      char *arg=*act;
      if (*arg!='-') 
	{
	  fprintf(stderr,"unknown option '%s'\n",arg);
	  fprintf(stderr,"try %s -h\n",executable );
	  exit(1);
	}
      arg++;

      if (!strcmp(arg,"h") || !strcmp(arg,"-help") || !strcmp(arg,"help"))
	{
	  print_help();
	  continue;	  
	}
      else if (!strcmp(arg,"v"))
	{
	  verbose_spot++;
	  continue;	  
	}
      else if (!strcmp(arg,"nsi"))
	{
	  no_system_includes++;
	  continue;	  
	}
      else if (!strcmp(arg,"ns"))
	{
	  ns++;
	  continue;	  
	}
      else if(!strcmp(arg,"ba"))
	{
	  {      set_allow_write(1);      set_allow_create(1);    }
	  build_all=1;
	  continue;
	}
      else if(!strcmp(arg,"p"))
	{
	  {      set_allow_write(1);      set_allow_create(1); }
	  process=1;
	  continue;
	}
      else if(!strcmp(arg,"bm"))
	{
	  build_mem=1;
	  continue;
	}
      else if(!strcmp(arg,"uv"))
	{
	  var_unused=1;
	  continue;
	}
      else if(!strcmp(arg,"stat"))
	{
	  stat=1;
	  continue;
	}
      else if(!strcmp(arg,"ui"))
	{
	  include_unused=1;
	  continue;
	}
      else if(!strcmp(arg,"ls"))
	{
	  ls=1;
	  continue;
	}
      else if(!strcmp(arg,"sys-in"))
	{
	  system_includes_only=1;
	}
      else if(!strcmp(arg,"ua"))
	{
	  {      set_allow_write(1);      set_allow_create(1);}
	  update_all=1;
	  continue;
	}
      else if(!strcmp(arg,"db-ro")) { arg[2]='_'; act--; continue; }
      else if(!strcmp(arg,"db-rw")) { arg[2]='_'; act--; continue; }

/*       else if(!strcmp(arg,"O1")) */
/* 	{ */
/* 	  optimize_struct=1; */
/* 	  continue; */
/* 	} */
/*       else if(!strcmp(arg,"O2")) */
/* 	{ */
/* 	  optimize_string=1; */
/* 	  continue; */
/* 	} */
/*       else if(!strcmp(arg,"O")) */
/* 	{ */
/* 	  optimize_struct=1; */
/* 	  optimize_string=1; */
/* 	  continue; */
/* 	} */
      OPT(fl)
	OPT(conf)
	OPT(dir)
	OPT(sym)
	OPT(db)
	OPT(db_ro)
	OPT(db_rw)
	OPT(ip)
	OPT(rf)
	OPT(regexp)
	OPT(def)
	OPT(fc)
	OPT(uf)
	OPT(si)
	OPT(ff)
	OPT(fu)
	OPT(fc)
	OPT(mf)
	OPT(mu)
	OPT(vc)
	OPT(vu)
	OPT(su)
	OPT(td)
	OPT(tu)
	OPT(in)
	OPT(bf)
	OPT(sc)
	OPT(sf)
	OPT(so)
	OPT2(lvd)
	OPT2(lvu)
	OPT2(va)
	OPT2(vla)
	OPT2(vfa)
	OPT2(vlfa)
	OPT2(vma)
	OPT2(vlma)
	OPT2(vva)
	OPT2(vlva)
	OPT2(cfa)
	OPT2(ufa)
	DEACTIVATE(fce_inline_deactivate)
	DEACTIVATE(fce_use_deactivate)
	DEACTIVATE(fce_def_deactivate)
	DEACTIVATE(fce_dec_deactivate)
	DEACTIVATE(var_dec_deactivate)
	DEACTIVATE(var_use_deactivate)
	DEACTIVATE(incl_deactivate)
	DEACTIVATE(mac_def_deactivate)
	DEACTIVATE(mac_use_deactivate)
	DEACTIVATE(mac_defd_deactivate)
	DEACTIVATE(mac_udef_deactivate)
	DEACTIVATE(str_dec_deactivate)
	DEACTIVATE(str_def_deactivate)
	DEACTIVATE(str_use_deactivate)
	DEACTIVATE(tdef_deactivate)
	DEACTIVATE(tdef_use_deactivate)
      else 
	{
	  printf("unknown option %s\n",arg);
	  fprintf(stderr,"try %s -h\n",executable );
	  exit(1);
	}
    }
    
#undef OPT
#undef OPT2

  if (dir) {
    if (chdir(dir))
      {
	FPRINTF_ERROR;
      }
  }
  if (!fl && conf)
    {
      /* kod zcasti zkopirovan z cspot_cc.c */
#define OPTION(buffer,opt,dst)           \
  {                                      \
    int l = strlen(buffer);              \
    int m = strlen(opt);                 \
    if (l>m && !memcmp(buffer,opt,m))    \
      {                                  \
	char* t = buffer+m;              \
	while(*t=='\t' || *t==' ') t++;  \
	if (*t=='=') t++;                \
	while(*t=='\t' || *t==' ') t++;  \
	strcpy(dst,t);                   \
      }                                  \
  }

      FILE* f = fopen(conf,"r");
      if (!f) 
	{
	  fprintf(stderr, "error occured no config file '%s' found\n",conf);
	  exit(1);
	}
      char buffer[MAX_LEN];
      char dr    [2*MAX_LEN];
      char flst  [MAX_LEN];
      while (!feof(f))
	{
	  read_line(f,buffer);
	  OPTION(buffer, "FILE_LIST", flst  );
	  OPTION(buffer, "DIRECTORY", dr );
	}
      char * fls = dr + strlen(dr);
      strcpy(fls,"/"); fls++;
      strcpy(fls,flst);
      fl = dr;
      fclose(f);
#undef OPTION

    }
#define n2s(x) #x

#define o(x)\
  if ( x )\
    {\
      char* tmp;\
      xalloc_context* context = xalloc_get_context();\
      xalloc_set_context(&options_context);\
      /* int size=strlen(y)+1+strlen( x )+10; */\
      int size=strlen(x)+1+strlen( x )+10;\
      tmp = xalloc(size);\
      snprintf(tmp,size,"%s %s",n2s(x),x);\
      /* snprintf(tmp,size,"%s %s",x,x); */\
     /*  printf("running_command:'%s'\n",tmp); */\
      run_command(tmp);\
      xfree(tmp);\
      xalloc_set_context(context);\
    }
#define o2(x)\
  if ( x )\
    {\
      char* tmp;\
      xalloc_context* context = xalloc_get_context();\
      xalloc_set_context(&options_context);\
      /* int size=strlen(y)+1+strlen( x )+2*sizeof(uint32_t)+2+10; */\
      int size=strlen(x)+1+strlen( x )+2*sizeof(uint32_t)+2+10;\
      tmp = xalloc(size);\
      /* snprintf(tmp,size,"%s %s %d %d",y,x,x ## _line ,x ## _pos); */\
      snprintf(tmp,size,"%s %s %d %d",n2s(x),x,x ## _line ,x ## _pos);\
      /* printf("running_command:'%s'\n",tmp);  */\
      run_command(tmp);\
      xfree(tmp);\
      xalloc_set_context(context);\
    }

  /* open sparse output */
  o(so);

  if (  system_includes_only )
    run_command ("sys-in");
  if (ns)
    {
      if (rf || uf || bf || process || build_all || build_mem || update_all || db || ls || stat || regexp || sym || rf || uf || fc || ff || fu || vc || vu || td || tu || mf || mu || in || sc || su || sf || si || bf || lvu || lvd || va || vla || vfa || vlfa || vma || vlma || vva || vlva || cfa || ufa)
	{
	  fprintf(stderr,"command 'ns' implies only sparsing will be performed \n");
	  exit(1);
	}
    }
  if (ip)  set_includes_file(ip);
  if (fl)  spot_config_filename=fl;
  if (def) set_macros_file(def);

  /* pro 'ba' 'ua' 'p' already set */
  if (rf || uf || bf ) 	  {      set_allow_write(1);      set_allow_create(1);}

  if (db)  { set_db_file(db); database_init(get_db_file()); }
  if ( db_ro )
    {
      char* tmp;
      xalloc_context* context = xalloc_get_context();
      xalloc_set_context(&options_context);
      int size=strlen("db-ro")+1+strlen( "db-ro" )+10;
      tmp = xalloc(size);
      snprintf(tmp,size,"db-ro %s",db_ro);
      run_command(tmp);
      xfree(tmp);
      xalloc_set_context(context);
    }
  if ( db_rw )
    {
      printf("Asad\n");
      char* tmp;
      xalloc_context* context = xalloc_get_context();
      xalloc_set_context(&options_context);
      int size=strlen("db-rw")+1+strlen( "db-rw" )+10;
      tmp = xalloc(size);
      snprintf(tmp,size,"db-rw %s",db_rw);
      run_command(tmp);
      printf("%s\n", tmp);
      xfree(tmp);
      xalloc_set_context(context);
    }

  if (process)         run_command ("p");
  else if (build_all)  run_command ("ba");
  else if (update_all) run_command ("ua");
  else if (build_mem)  run_command ("bm");

  if (ls)  run_command ("ls");
  if (stat) run_command ("stat");
  if (ns)  run_command ("ns"); 


  o(regexp);
  o(sym);
  o(rf);
  o(uf);
  o(fc);
  o(ff);
  o(fu);
  o(vc);
  o(vu);
  o(td);
  o(tu);
  o(mf);
  o(mu);
  o(in);
  o(sc);
  o(su);
  o(sf);
  o(si);
  o(bf);
  o2(lvu);
  o2(lvd);
  o2(va);
  o2(vla);
  o2(vfa);
  o2(vlfa);
  o2(vma);
  o2(vlma);
  o2(vva);
  o2(vlva);
  o2(cfa);
  o2(ufa);
#undef o
#undef o2

 
  if (var_unused)       run_command ("var_unused");
  if (include_unused)   run_command ("include_unused");

  if (build_all || update_all || fc || ff || fu || vc || vu || mf || mu 
      || in || sc || sf || si || uf || bf || rf
      || va || vla || vfa || vlfa ||vma || vlma || vva || vlva || cfa || ufa
      || td || tu || su || ls || sym || lvd || lvu
      || var_unused || include_unused || regexp || stat  || process
      || build_mem || ns
      )
      run_command ("bye");
  return 0;
}


