/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include "base/macros.h"

#ifndef MAP_ANONYMOUS
#define MAP_ANONYMOUS MAP_ANON
#endif

#include "xalloc.h"
struct head_struct {
  struct head_struct* next;
  struct head_struct* prev;
  int size;
};

typedef struct head_struct head;

static unsigned int allocated_all;
static unsigned int allocated_all_size;

//static xalloc_context global_context;
#if 0
static xalloc_context* context=&global_context;
#else
static xalloc_context* context;
#endif

void xalloc_set_context(xalloc_context* c)
{
  context=c;
}

xalloc_context* xalloc_get_context(){
  return context;
}



static void append(void* ptr)
{
#ifdef XALLOC_DEBUG
  if (!context) { printf("nebyl nastaven context --- chyba !!! \n"); exit(1); }
#endif  
  allocated_all++;
  context->allocated++;

  head* h =(head*) ptr;
  h->next=h->prev=0;
  if (context->last)
    {
      ((head*)context->last)->next=h;
      h->prev=((head*)context->last);
      context->last=h;
    }
  else
    {
      context->last=context->first=h;
    }
}

static void delete(void* ptr)
{
  allocated_all--;
  context->allocated--;

  head* h=(head*) ptr;

  if (h==context->first)
    {
      if(context->first==context->last)
	{
	  context->first=context->last=0;
	}
      else
	{
	  context->first=h->next;
	  context->first->prev=0;
	}
      return;
    }
  if (ptr==context->last)
    {
      if(context->first==context->last)
	context->first=context->last=0;
      else
	{
	  context->last=h->prev;
	  context->last->next=0;
	}
      return;
    }
#ifdef TEST_CONTEXT
  {
    head* tmp=h;
    while (tmp && tmp!=context->last)
      {
	tmp=tmp->next;
      }
    if (!tmp)
      {
	printf("dealokuji z jineho kontextu '%s'\n",context->name);
	ABORT
      }
  }
#endif
  if (h->next)
    {
      h->next->prev=h->prev;
    }
  if (h->prev)
    {
      h->prev->next=h->next;
    }
}

void* myalloc(int s)
{
  if (s<=0) return 0;
  int size=s+sizeof(head);
  void *ptr;

  ptr = malloc(size);

  if (ptr)
    {
      memset(ptr, 0, size);
      append(ptr);
      ((head*) ptr)->size=size;
      context->size+=size;
      allocated_all_size+=size;
    }
  else
    {
      ABORT
    }
  //  printf("allocated %p\n",ptr+sizeof(head)); fflush(stdout);
  return ptr+sizeof(head);
}

void  myfree(void* ptr)
{
  void* addr=ptr-sizeof(head);
  allocated_all_size-=((head*) addr)->size;
  delete(addr);

  int size=((head*) addr)->size;
  context->size-=size;

  if (  context->size<0 )
    { 
      printf("odalokoval jsem vice nez jsem v kontextu\
  '%s' naalokoval\n",context->name);      
      ABORT
    }

  //  printf("freed %p\n",addr); fflush(stdout);
  free(addr);
}



void  myfree_all()
{
#ifdef SHOW_XALLOC_INFO
  if (!context->size) return;
  printf("allocated mine free_all start = %d\n",context->size);
#endif
  int all_size=0;
  head* act=context->first;
  head* next;
  if (act)
  for (;;)
    {
      allocated_all--;
      context->allocated--;
      allocated_all_size-=act->size;
      next=act->next;

      void* addr=act;
      all_size += act->size;
      //      printf("freed_all %p\n",addr); fflush(stdout);
      free(addr);

      if (next)
	act=next;
      else break;
    }
  if( (!context->first && context->last) || (context->first && !context->last) )
    fprintf(stderr,"nekonsistence contex->first=%p  context->last=%p\n",context->first,context->last);
  if (context->allocated)    
    fprintf(stderr,"context->allocated=%d\n",context->allocated);
  context->first=context->last=0;
  context->allocated=0;
  context->size-=all_size;
#ifdef SHOW_XALLOC_INFO
  printf("freed size=%d\n", all_size);
  mystat();
#endif
}

void mystat()
{
  printf("allocated all = %d\n" ,allocated_all_size);
  printf("allocated mine = %d\n",context->size);
}



void* myrealloc(void* old, int s)
{
  if (s<=0)
    {
      myfree(old);
      return 0;
    }
  if(!old)
    {
      return myalloc(s);
    }


#if 0
  head* o=(head*) (old-sizeof(head));
  {
    int rozdil = (s+sizeof(head)) - o->size;  
    allocated_all_size+=rozdil; 
    context->size+=rozdil;
  }
  head* new = realloc(o, s+sizeof(head));
  if (new->next) new->next->prev = new;
  if (new->prev) new->prev->next = new;
  new->size=s+sizeof(head);
  return ((void*) new)+sizeof(head);
#else
  void *ptr;
  ptr = myalloc(s);
  if (ptr) 
    {
      head* o=(head*) (old-sizeof(head));
      int sz = 
	s > o->size - sizeof(head)
	? o->size - sizeof(head)
	: s;
      //      printf("old=%p ptr=%p s=%d  o->size=%d  sz=%d\n",old,ptr, s, o->size, sz);
      memcpy(ptr, old, sz);
      myfree(old);
    }
  else 
    {
      fprintf(stderr,"cannot allocate memory");
      exit(1);
    }
  return ptr;
#endif
}
