/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

/*
  Implementace Faginova heovn
  piem tabulka obsahuje vechny funkce potebn pro manipulaci se zznamy
 */
#include <sys/mman.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include "file.h"
#include "fagin.h"
#include "base/xalloc.h"
#include "base/macros.h"

#include "database.h"

#if 0
#ifdef BOUND
#if BOUND < 1
#error "macro BOUND znamena pocet procent"
#elif BOUND > 99
#error "macro BOUND znamena pocet procent"
#endif
#endif
#endif

static int check_if_inserted ;
void   set_check_inserted() { check_if_inserted=1; }
void unset_check_inserted() { check_if_inserted=0; }


/*
  vrac adresu posledn strnky koliznho etzku
 */
uint32_t last_collision_page(file_t* file, uint32_t adr);

#if 0
void print_uint32(uint32_t i) {
  uint32_t mask=1<<31;
  int num;

  for (num=0; num<32;num++,i<<=1)
    putchar("01"[!!(i&mask)]);
}
#endif

#if 0
static uint32_t print_collision_chain(file_t* file, uint32_t adr)
{
  while (adr)
    {
      fprintf(stderr,"%X ",adr);
      uint32_t t=get_next_page_address(file,adr);
      if (!t) break; 
      else adr=t;
    }
  fprintf(stderr,"\n");
  if (!adr) abort(); /*mohlo vzniknout jedine tak, ze prvni stranka nebyla alokovana ale ona byla*/
  return adr;
}
#endif

static int enlarge_folder(hash_table* table);
/*
  Pro dan poet byt a he vrt index v adresi
 */
static inline unsigned int __get_index(uint32_t hash, int bites)
{
  unsigned int shift=(CHAR_BIT*(sizeof(uint32_t))-bites);
  unsigned int r; 
  if (!bites) r=0;
  else r=hash>>shift;
  return r;
}
/*
  Vrac index v heovac tabulce pro pslunou he a tabulku
 */
inline unsigned int get_index(hash_table* table, uint32_t hash)
{
  return __get_index(hash,get_hash_bites_folder(table->file,get_folder_address(table)));
}
/*
  Vrac index v heovac tabulce pro pslunou he po zvten adrese
 */
static inline unsigned int get_new_index(hash_table* table, uint32_t hash)
{
  return __get_index(hash,get_hash_bites_folder(table->file,get_folder_address(table))+1);
}

/*
  Vrac 1 pokud strnka na pslun adrese pouv mn bit ne j umouje adres
  Jinak 0 
 */
static int less_used_bites(hash_table* table,file_t* file, uint32_t adr)
{
  unsigned char adresar ;
  unsigned char stranka ;
  if (!file ||! table || !adr ) ABORT
  adresar = get_hash_bites_folder(table->file,get_folder_address(table));
  stranka = get_page_bites(file,adr);
  return adresar > stranka;
}

/*
  K etzci, kter je uren hodnotou he pid dal strnku 
  a vrt jej adresu
 */
static uint32_t next_collision_page(hash_table* table, unsigned int hash  )
{
  file_t* file;
  unsigned int index; /* poloha kam se vlozi dana hesh */
  file  = table->file;
  index = get_index( table,
		     hash);
  uint32_t prev_adr =  get_address_at_folder_index(  file,
						     get_folder_address(table),
						     index); /* nastavi adresu stranky podle adresare */
  while (prev_adr) /* posune ukazatel az na konec retizku */
    {
      uint32_t t=get_next_page_address(file,prev_adr);
      if (!t) break; 
      else prev_adr=t;
    }

  uint32_t adr     = page_alloc(file);     
  set_page_bites(file,adr,  get_page_bites( file, 
					    prev_adr));
  set_next_page_address(file,prev_adr,adr); /* zaradi stranku do retizku */
  return adr;
}

/*
  vrati nenulove cislo pokud je nutne vytvorit kolizi
  vrati 0 pokud ma smysl prehesovat
 */
static int count_collisions(hash_table* table, unsigned int hash  )
{
  if (!table) ABORT
  file_t* file;
  unsigned char x;
  unsigned char y;
  unsigned int index; /* poloha kam se vlozi dana hash */
  uint32_t first_page;
  file  = table->file;
  index = get_index( table,
		     hash);
  x=get_hash_bites_folder(  table->file,
			    get_folder_address(table));
  first_page= get_address_at_folder_index(  table->file,
					    get_folder_address(table),
					    index);
  y=get_page_bites( file, 
		    first_page);
  if (x<y) /* je nesmysl aby mel adresar mene bajtu nezli stranka */
    {
      fprintf(stderr,"x (%d)  < y (%d)\n",x,y);
      ABORT
    }

  unsigned int count=0, i1=0, i2=0;
  uint32_t pg_adr=first_page;
  do { /* pro vechny strnky koliznho etzku a pro kad jejich zznam
	  spotm kam by se peheovali
       */
    uint32_t adr=skip_page_head(pg_adr);
    for (  ;
	   table->get_len(0)+adr-pg_adr<PAGE_SIZE /*jeste jsem uvnitr stranky*/
	     && table->has_item(file,adr,pg_adr) /* a je tam dalsi polozka */
	     ; adr=table->next_item(file,adr)) /* posunu se na dalsi */
      {
	int hash           = table->get_hash(file,adr);
	uint32_t act_index = get_index(table,hash);
	uint32_t new_index = get_new_index(table,hash);
	int act_start  = act_index-act_index%(1<<(x-y));
	int act_length = 1<<(x-y);
	int new_start  = 2*act_start;
	int new_length = 2*act_length;
	int middle     = act_length;
	/* nyni urcim kam padne po prehesovani */
	if (new_index < new_start+middle) i1++;
	else if ( new_index<new_start+new_length ) i2++;
	else { printf("spatne urceno kam prehesovat "); exit(1); }
#if !defined BOUND
	if (i1 && i2) goto rehesh;
#endif

	count++;
      }
  } while ((pg_adr=get_next_page_address(table->file,pg_adr))); /* posunu na dalsi stranku retizku */

#if defined BOUND
  double d1 = 100 * ((double) i1 / (double) count); 
  double d2 = 100 * ((double) i2 / (double) count);
#endif

#if defined BOUND
  if ( d1<BOUND || d2<BOUND) 
  /* jedna ze vzniklych stranek ma mene procent nez BOUND -> pripojit dalsi */
#else
 rehesh:
  if ( i1==0 || i2==0) 
#endif
    {
      return 1; /* pripoji dalsi stranku ke koliznimu retezci */
    }
  return 0; /* bude prehasovavat */
}


/*
  funkce pozaduje:
     stranka pouziva mene bitu nez adresar povoluje
  funkce nepozaduje:
     stranky mohou mit kolizniho naslednika
  funkce provede:
     alokuje stranku do niz lze vlozit item
     prehesuje relevantni stranky do vice stranek
     preznaci vsechny relevantni stranky na vice bitu
  funkce vraci:
     adresu stranky do ktere lze item zahesovat
 */
#define debug_chain(x,y)           \
     if (1) {                      \
       uint32_t nadr,adr=x;        \
       nadr=adr;                   \
       printf("%s %X",y,nadr);     \
       while (nadr)                \
	 {                         \
	   nadr=get_chain_page_address(file,nadr); \
	   printf("->%X",nadr);                    \
	 }                                         \
       printf("\n");                               \
     }

static void rehash_chains(hash_table* table, uint32_t first_page)
{
  return ;

  file_t* file;
  file  = table->file;

  uint32_t next=0;
  uint32_t act;
  if (!first_page) goto skip_chain;
  if (first_page && !get_chain_page_address(file,first_page)) goto skip_chain;
  
/*   printf("BUDU PREHAZOVAT RETIZKY\n"); */
/*   debug_chain(first_page,"stary retizek:"); */

  act=get_chain_page_address(file,first_page);
  do {
    const void* act_key=table->get_chain_key(table->file,skip_page_head(act));
    {
      unsigned int hash    = table->hash_fce(act_key);
      unsigned int index   = get_index(table,hash);
      uint32_t     lpg_adr = get_address_at_folder_index(table->file,get_folder_address(table),index);
      if (!lpg_adr) /* neexistuje stranka hlavniho retezce */
	{
	  lpg_adr = page_alloc(file);
	  set_address_at_folder_index(table->file, get_folder_address(table), index, lpg_adr);
	  printf("!!!!!!!!!!!!!!!! nova stranka %X\n",lpg_adr);
	  ABORT
	}
      
      { /* append this chain page to this index */
#if 0
	unsigned char x=get_hash_bites_folder(  table->file,
				  get_folder_address(table));
	unsigned char y=get_page_bites( file, 
					first_page);
#endif

/* 	printf("\tact page = %X (ta ktera ma byt zarazena do retizku)\n", act); */
/* 	printf("\t\t key = %X\n",*(uint32_t*) act_key); */
/* 	printf("\t\t hash = "); print_uint32(hash); printf("\n"); */
/* 	printf("\t\t index = %d bits folder=%d page=%d\n",index,x,y); */

#if 0 // 
	if (act)  next = get_chain_page_address(table->file,act);
	else 	  next = 0;
#endif

/* 	debug_chain(lpg_adr,"\told 2nd retizek - k nemu se pripojuje:\n"); */	
/* 	printf("\tsetting act = %X points to %X\n" */
/* 	       , act */
/* 	       , get_chain_page_address(file,lpg_adr)); */
	
	set_chain_page_address(file
			       , act
			       , get_chain_page_address(file,lpg_adr));
	
/* 	printf("\tsetting first = %X points to %X\n" */
/* 	       , lpg_adr */
/* 	       , act); */
	set_chain_page_address(file
			       , lpg_adr
			       , act);
/* 	debug_chain(lpg_adr,"\tvyrobeny retizek: "); */
      
      /* 	   if (act) set_chain_page_address(table->file,act,0); */
	act=next;
      }
    }
  } while (act);
 skip_chain:
  ;
}

static uint32_t create_more_bit_page(hash_table* table, unsigned int hash)
{
  file_t* file;
  unsigned char x;
  unsigned char y;
  unsigned int index; /* poloha kam se vlozi dana hesh */
  uint32_t first_page;
  file  = table->file;
  index = get_index( table, hash);
  first_page = get_address_at_folder_index( table->file,
					    get_folder_address(table),
					    index);
  x=get_hash_bites_folder(  table->file,
			    get_folder_address(table));
  y=get_page_bites( file, 
		    first_page);
  if (x<=y) /* zakazano pozadavky funkce */
    {
      fprintf(stderr,"x (%d)  < y (%d)\n",x,y);
      ABORT
    }
  int length=1<<(x-y); /* soucastne rozsah v adresari pro dany retizek stranek */
  int start=(index/length)*length; /* misto kde v adresari odkazy na retizekk zacinaji */
  int middle=length>>1; /* pred middle bude ukazovat na retizek1 za middle na retizek2*/

  /* vim ze obe stranky budou vyuzite, nebot jinak bych pridaval do retizku...*/
  uint32_t adr1=page_alloc(file);
  uint32_t adr2=page_alloc(file);
  set_page_bites(file,adr1,y+1);
  set_page_bites(file,adr2,y+1);
  { /* zde budu prehesovavat obsah jednoho retizku do dvou retizku */

/*     printf("REhASH of %X\n",first_page); */

   { /* nyni nastavim adresar aby ukazoval na nove alokovane stranky */
      if (length == 1 || middle==0 ) ABORT /* pokud je middle==0 pak by doslo k chybe stejne tak length 1 rika ze x==y coz je zakazane */
      uint32_t fa=get_folder_address(table);
      int i;
      for (i=0;i<middle;i++)
	set_address_at_folder_index(file,fa,start+i,adr1);
      for (i=middle;i<length;i++)
	set_address_at_folder_index(file,fa,start+i,adr2);
    }
#ifdef INSERT_TEST
   int find_failed=2;
#endif

/*    printf("###### INDEX = %d   bits folder=%d page=%d\n",index,x,y); */
/*    printf("$$$$$$ HASH = "); print_uint32(hash); printf("\n"); */
   rehash_chains(table,first_page);

   { /* nyni projdu cely retizek a zaradim kazdy jeho prvek do jednoho z retizku */
     /* rovnou dealokuji stranky stare stranky */
     uint32_t adr=0;
     uint32_t pg_adr=first_page;
     do {
       adr=skip_page_head(pg_adr);
       for ( 
	    ; table->get_len(0)+adr-pg_adr<PAGE_SIZE  && table->has_item(file, adr , pg_adr)
	      ; adr=table->next_item(file,adr))
	 {
	   char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
	   void* item = (void*) tmp;
	   table->get_item(file,adr,item);
	   
#ifdef SHOW_INSERT_TEST
	   fprintf(stderr,"@@@@@@@@@@@@@@@ INSERT @@@@@@@@@@@@@@@@@\n");
#endif
#ifndef INSERT_TEST
	   int ai = test_already_inserted;
	   test_already_inserted=0;
	   insert(table, item);
	   test_already_inserted=ai;
#else
	   int ai = test_already_inserted;
	   test_already_inserted=0;
	   int i = insert(table,item);
	   test_already_inserted=ai;
	   if (find_failed==2) find_failed=0;
	   if ( table->__inserted?
		table->__inserted(table,item,i)
		:0)
	     {
	       table->__print?table->__print(file,adr):0;
	       find_failed=1;
	       fprintf(stderr,"create more bits page ... rehashing failed adr=%X  insert returns %X\n###################################################\n",adr,i);
	       exit(0);
	     }
#endif
	 }
       uint32_t t=get_next_page_address(file,pg_adr);
#ifdef INSERT_TEST
       //#define INSERT_TEST_REMEMBER_REHASH
#endif
#ifndef INSERT_TEST_REMEMBER_REHASH
       page_free(file,pg_adr);
#endif
       pg_adr=t;
     } while (pg_adr);
   }
#ifdef INSERT_TEST_REMEMBER_REHASH
   if (0&&find_failed) { /* nyni projdu cely retizek a kazdy prvek vyhledam v databazi */
     fprintf(stderr,"jednotlive checky uspesne -> kontroluju globalne\t");
     uint32_t adr=0;
      uint32_t pg_adr=first_page;
      uint32_t *all = list(table);
      do {
	adr=skip_page_head(pg_adr);
	for ( 
	     ; table->get_len(0)+adr-pg_adr<PAGE_SIZE  && table->has_item(file, adr , pg_adr)
	       ; adr=table->next_item(file,adr))
	  {
	    char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
	    void* item = (void*) tmp;
	    table->get_item(file,adr,item);
	    if ( table->__inserted?
		 table->__inserted(table,item,adr)
		 :0)
	      {
		table->__print?table->__print(file,adr):0;
		fprintf(stderr,"rehash check failed adr=%X\n",adr);
		if (1)
		  {
		    uint32_t *a;
		    int f=0;
		    for (a=all;*a;a++)
		    {
		      char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
		      void  *act=(void*) tmp;
			table->get_item(file,*a,act);
		      if (!table->compar_item(act,item)) 
			{
			  f++;
			  if (!find_failed)
			    printf("item found at %X\n",*a);
			}
		    }
		  if (!f)
		    fprintf(stderr,"not found in entire database count\n");
		  }
		printf("try\n");
	      }
	    //	    xfree(item);
	  }
	uint32_t t=get_next_page_address(file,pg_adr);
	page_free(file,pg_adr);
	pg_adr=t;
      } while (pg_adr);
      xfree(all);
      fprintf(stderr,"konec kontroly globalne\n\n");
    }
#endif
  }

  { /* jeste zbyva zajistit aby se vesel dalsi zaznam do daneho retizku a vratit adresu stranky do ktere vlozit */
    uint32_t adr; /* adresa stranky kam hash pujde */
    adr=get_address_at_folder_index(  table->file,
				      get_folder_address(table),
				      get_index( table, hash));
    /* nyni uz rovnou najdu posledni straknu retezce a vlozim do ni */
    /* tim mirne porusim pripadny BOUND, ale ten se bere jen jako doporuceni pro prehesovavani a pri dalsim vlozeni by se to opravilo */
    
    while (adr)
      {
	uint32_t t=get_next_page_address(file,adr);
	if (!t) break; 
	else adr=t;
      }
    if (!adr) ABORT /*mohlo vzniknout jedine tak, ze prvni stranka nebyla alokovana ale ona byla*/
    /* pokud neni v posledni strance mist, tak alokuj dalsi */
    if (!table->enough_space(file,adr,0))
      // opravit enough space aby brala jen 2 parametry
      {
	uint32_t next_adr  = page_alloc(file);     
	set_page_bites(file,next_adr,get_page_bites(file,adr));
	set_next_page_address(file,adr,next_adr);
#ifdef SHOW_INSERT_TEST
	fprintf(stderr,"added last page %X prev %X\n",adr,next_adr);
#endif
 	adr=next_adr;
     }

#if 0
    fprintf(stderr, "*********************************************************\n");
    fprintf(stderr, "add page suggests page %X  first in collision chain is %X\n",adr,get_address_at_folder_index(  table->file,
				      get_folder_address(table),
				      get_index( table, hash)));
#endif
    return adr;
  }
  return 0;
}

/*
  pid strnku do kter se bude moci pslun he vloit
 */
static uint32_t add_page(hash_table* table, uint32_t hash)
{
  if (!table) ABORT

  uint32_t adr_sug=0;
  uint32_t adr = get_address_at_folder_index(  table->file,
				  get_folder_address(table),
				  get_index( table,
					     hash));
  int less=less_used_bites(table,table->file,adr);
  //  char a=0;
  if (count_collisions(table,hash)) 
    /* zjisti zda je nutne vytvaret kolizni retezec*/
    {
      /* provede pripojeni dalsi stranky ke koliznimu retezci */
      //      a='A';
      adr_sug = next_collision_page(table,hash);
    }
  else if (less) /* udava zda je adresar vicebitovy nezli stranka */
    {
      /* prida stranku a pro druhou stranku(retezec stranek) preznaci pocty bitu */
      //      a='B';
      adr_sug=create_more_bit_page(table,hash);
    }
  else
    {
      if (!enlarge_folder(table)) 
	{
	  fprintf(stderr,"cannot enlarge folder\n");
	  return 0;
	}
      else 
	{
	  //	  a='C';
	  adr_sug=create_more_bit_page(table,hash);
	}
    }

  uint32_t adr_col =
    last_collision_page(table->file
			, get_address_at_folder_index(table->file
						      , get_folder_address(table)
						      , get_index(table,hash) ));
  adr=adr_col;
#if 0
  if (1) fprintf(stderr,"did: %c\n",a);
  if (adr_sug!=adr_col)
    {
      fprintf(stderr,"did: %c\n",a);
      fprintf(stderr,"address_suggested %X\t",adr_sug);
      fprintf(stderr,"address_collision %X\n",adr_col);
    }

  if (!table->enough_space(table->file,adr,0))
    // opravit enough space aby brala jen 2 parametry
    {
      uint32_t next_adr  = page_alloc(table->file);     
      set_page_bites(table->file,next_adr,get_page_bites(table->file,adr));
      set_next_page_address(table->file,adr,next_adr);
      adr=next_adr;
      fprintf(stderr, "pridavam stranky tells %X\n",adr);
    }
#endif

  return adr;
}

/*
  funkce zvt adres dan tabulky
 */
static int enlarge_folder(hash_table* table)
{

  if (!table) ABORT
  file_t* file=table->file;
  if (get_hash_bites_folder(file,get_folder_address(table))==MAX_BITS) 
    {
      printf("max bites used  cannot enlarge\n");
      ABORT
      return 0;
    }

  uint32_t old_folder_adr=get_folder_address(table);

  //  printf("fagin folder too small _enlarge_folder\n");
  uint32_t adr=folder_page_alloc(file ,get_hash_bites_folder(file,old_folder_adr)+1);

  int i,j,k;
  k=get_hash_bites_folder(file,old_folder_adr);
  j=1<<k;
  for(i=0;i<j;i++)
    {
      set_address_at_folder_index(file, adr, 2*i,   get_address_at_folder_index(file, old_folder_adr,i));
      set_address_at_folder_index(file, adr, 2*i+1, get_address_at_folder_index(file, old_folder_adr,i));
    }
  
  if (old_folder_adr)
    folder_page_free(file,old_folder_adr);
  set_folder_address(table,adr);
 return 1;
}

/*
  vrac adresu posledn strnky koliznho etzku
 */
uint32_t last_collision_page(file_t* file, uint32_t adr)
{
  while (adr)
    {
      uint32_t t=get_next_page_address(file,adr);
      if (!t) break; 
      else adr=t;
    }
  if (!adr) ABORT /*mohlo vzniknout jedine tak, ze prvni stranka nebyla alokovana ale ona byla*/
  return adr;
}

uint32_t prev_collision_page(file_t* file, uint32_t adr, uint32_t drag)
{
  while (adr  )
    {
      uint32_t t=get_next_page_address(file,adr);
      if (!t || t==drag ) break; 
      else adr=t;
    }
  if (!adr) ABORT /*mohlo vzniknout jedine tak, ze prvni stranka nebyla alokovana ale ona byla*/
  return adr;
}



/*
  vlo zznam item do dan tabulky
 */
int insert(hash_table* table,  const void* item  )
{

  file_t* file;
  unsigned int hash;
  unsigned int index;
  uint32_t adr;
  if (!table || !item ) ABORT;

  if (!**table->i_folder_address) 
    create_folder(&files,0);

  table->__protect?table->__protect(item):0;
  file  = table->file;
  hash = table->hash_fce(table->get_key(item));
  index= get_index(table,hash);
  adr  = get_address_at_folder_index(table->file,get_folder_address(table),index);
  if (!adr)
    {
      adr = page_alloc(file);
      set_address_at_folder_index(table->file, get_folder_address(table), index, adr);
    }

   {
     uint32_t r;
     uint32_t aia=0;

     if ( test_already_inserted 
	  && table->already_inserted 
	  && ((aia=table->already_inserted(file, adr, item)))) /* pokud jiz zaznam existuje, neprida ho */
       {
	 return 1; /* zaznam byl 'uspesne' pridan */
       }

     if (check_if_inserted && table->check_item_inserted(file,adr,item)) return 1;

     r=table->add_item(file, last_collision_page(file,adr), item);

     if (!r)
       {
	 adr=add_page(table,hash); /* nepodarilo se tedy potrebuji novou stranku */
	 if (!adr)
	   {
	     fprintf(stderr,"not added page\n");
	     ABORT
	     return -1;
	   }

	 r=table->add_item(file,adr,item);
	 return r;
       }
     return r?1:0;
   }
}

/*
 */
int find(hash_table* table
	       , const void* key
	       , int (*found_item_callback) (uint32_t adr)
	       , int (*chain_found_item_callback) (uint32_t adr)
	       )
{
#ifdef FAST_BUILD_STRING_ONLY  
  { printf("find called\n"); ABORT }
#endif
  int ret=0;

  if (!table||!key) { printf("find spatne zavolan\n"); ABORT }
  if (!**table->i_folder_address) 
      return 0; 
  unsigned int hash    = table->hash_fce(key);
  unsigned int index   = get_index(table,hash);
  unsigned int pg_adr  
    = get_address_at_folder_index(table->file
				  , get_folder_address(table),index);
#ifdef SHOW_INSERT_TEST
  fprintf(stderr, "find at index %d  first chain page_address %X\n", index, pg_adr);
#endif
  if (!pg_adr) { 
    return 0; 
  }

  do {
    uint32_t adr=skip_page_head(pg_adr);
    for (;table->has_item(table->file,adr,pg_adr);adr=table->next_item(table->file,adr))
      {
	char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
	void* act=(void*) tmp;
	table->get_item(table->file,adr,act);
	const void* act_key=table->get_key(act);
	if (!table->compar_key(act_key
			       , key
			       , table->item_hash(act)
			       , hash))
	  {
	    ret++;
	    if ( table->allow_chain 
		&& table->is_item_chain_start(act))
	      {
		uint32_t aadr = table->get_item_chain_start(act);
		uint32_t adr  = skip_page_head(aadr);
		
		do {
		  for (
		       ; table->has_chain_item(table->file,adr,aadr)
			 ; adr=table->next_chain_item(table->file,adr))
		    {
		      char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
		      void* act=(void*) tmp;
		      table->get_chain_item(table->file
					    ,adr
					    ,act
					    ,act_key);
		      ret++;
		      chain_found_item_callback?
			chain_found_item_callback( table->find_return(table->file,adr) )
			:0;
		    }
		  aadr=get_next_page_address(table->file, aadr);
		  adr = skip_page_head(aadr);
		} while (aadr);
	      }
	    else
	      {
		found_item_callback?
		  found_item_callback(
				      table->find_return(table->file,adr))
		  :0;
	      }
	  }

      }
  } while ((pg_adr=get_next_page_address(table->file,pg_adr)));

  return ret;
}

/*

  smae vechny item, kter odpovdaj danmu kli  
 */
uint32_t delete(hash_table* table , const void* key)
{
  unsigned int hash = table->hash_fce(key);
  unsigned int index= get_index(table,hash);
  unsigned int pg_adr  = get_address_at_folder_index(table->file,get_folder_address(table),index);
  /*  do { */  /* funkce remove se sama postara o to, aby prosla vsechny stranky retizku */
  if(pg_adr) 
    table->remove_item(table,pg_adr,key);
  /*   } while ((pg_adr=get_next_page_address(table->file,pg_adr))); */
  return 0;
}
/*
  nsledujc funkce delete jsou uren k vymazn vech zznam ktermi pispl urit soubor
 */

/*
  vrt seznam vech item v dan tabulce
 */
void list(hash_table* table, void (*action) (uint32_t adr))
{
  if (!table) { printf("list spatne zavolan\n"); ABORT }
  int i=0;
  int bites=get_hash_bites_folder(table->file,get_folder_address(table));
  int length=1<<bites;
  uint32_t old_adr=0;
  uint32_t pg_adr=0;
  for (;i<length;i++)
    {
      unsigned int index= i;
      uint32_t a = 
	get_address_at_folder_index(table->file,
				    get_folder_address(table), index);

      if ( a != old_adr && (old_adr=a) )
	{
	  pg_adr=a;
	  do {
	    uint32_t adr=skip_page_head(pg_adr);
	    for (;table->has_item(table->file,adr,pg_adr);
		 adr=table->next_item(table->file,adr))
	      {
		action(adr);
	      }
	    
	  } while ((pg_adr=get_next_page_address(table->file,pg_adr)));
	}
    }
}


#ifdef CHAIN_LENGTH
void chain_length_counter(hash_table* table)
{
  if (!table) { printf("list spatne zavolan\n"); ABORT }
  int i=0;
  int bites=get_hash_bites_folder(table->file,get_folder_address(table));
  int length=1<<bites;
  uint32_t last_adr=0;
  for (;i<length;i++)
    {
      unsigned int index= i;
      uint32_t first_adr = 
	get_address_at_folder_index(table->file,
				    get_folder_address(table),index);
      if (!first_adr) continue;
      if (first_adr==last_adr) continue;

      uint32_t pg_adr=first_adr;
      int length=0;
      do {
	length++;	  
      } while (pg_adr && (pg_adr=get_next_page_address(table->file,pg_adr)));
      last_adr=first_adr;
      printf("index %d  first_page %X  list length %d\n",i,first_adr,length);

      if (get_chain_page_address(table->file,first_adr))
	{
	  int i;
	  uint32_t adr;
	  for (i=0, adr=get_chain_page_address(table->file,first_adr)
		 ; adr && adr != 0xffffffff
		 ; i++, adr=get_chain_page_address(table->file,adr));
	  if (i>1) 
	    {
	      printf("\t chain length=%d   list of lengths: ",i);
	      
	      for ( adr=get_chain_page_address(table->file,first_adr)
		      ; adr
		      ; adr=get_chain_page_address(table->file,adr))
		{
		  int j=0;
		  uint32_t adr2=adr;
		  for (  ; adr2
			   ; j++, adr2=get_next_page_address(table->file,adr2));
		  printf(" %d",j);
		}
	      printf("\n");
	    }
	}
    }
}
#endif










int delete_all_FILE_IDENT_chain(hash_table* table , const void* fi_key)
{

  if (!table||!fi_key) { printf("delete spatne zavolan\n"); ABORT }
  //  int bites=get_hash_bites_folder(table->file,get_folder_address(table));

  file_item** fi;
  int key_length=0;


  for (fi=(file_item**) fi_key;*fi;fi++,key_length++)
    {
      uint32_t identifier_adr=
	get_file_item_identifier( (*fi) );
      unsigned int hash   = table->hash_fce(&identifier_adr);
      unsigned int index  = get_index(table,hash);
      uint32_t pg_adr     = get_address_at_folder_index(table->file,get_folder_address(table),index);

      void* key = &identifier_adr;

      do {
	uint32_t adr=skip_page_head(pg_adr);
	for (;table->has_item(table->file,adr,pg_adr)
	       ;adr=table->next_item(table->file,adr)) { 
	  char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);
	  void* act=(void*) tmp;
	  table->get_item(table->file,adr,act);
	  const void* act_key=table->get_key(act);
	  if (!table->compar_key(act_key
				 , key
				 , 0
				 , 0)) {
	    if (table->allow_chain 
		&& table->is_item_chain_start(act)) {
	      uint32_t aadr = table->get_item_chain_start(act);
	      table->chain_file_remove(table,aadr,fi_key);
	    }
	  }
	}
      } while ((pg_adr=get_next_page_address(table->file,pg_adr)));
    } 
  return 0;
}
/*
  smae vechny item, kter maj filename a ident_name stejn 
  jako njak file_item v seznamu
*/
int delete_all_FILE_IDENT(hash_table* table , const void* fi_key)
{
  /*
    to do
    
    naalokovat velikost adresare kazda polozka 1 bit
    projit vsechny fi a nastavovat na true pokud se nektera sem zahesuju

    prochazet dany adresar a pokud true, pak vytvorit seznam tech ktere se sem zahesuji
    predat tento seznam remove_item

  */ 

  if (!table||!fi_key) { printf("delete spatne zavolan\n"); ABORT }
  int bites=get_hash_bites_folder(table->file,get_folder_address(table));
  int length=1<<bites;
  uint32_t* pages=xalloc(length*sizeof(uint32_t));


  // memset(files,0,sizeof(uint32_t)*length);

  file_item** fi;
  int key_length=0;

  /*
    nejprve nastavim ty adresy stranek, v nichz je nejaky identifikator k odstraneni
   */
  for (fi=(file_item**) fi_key;*fi;fi++,key_length++)
    {
      uint32_t identifier_adr=
	get_file_item_identifier( (*fi) );
      unsigned int hash   = table->hash_fce(&identifier_adr);
      unsigned int index  = get_index(table,hash);
      uint32_t pg_adr     = get_address_at_folder_index(table->file,get_folder_address(table),index);
      pages[index]=pg_adr;
    } 

  uint32_t last=0; /* pokud se z adresare ukazuje nekolikrat na tentyz retizek staci projit jen jednou */

  { /* nyni budu mazat z nalezenych stranek */
    int j;  uint32_t* p;  
    for (j=0,p=pages;j<length;j++,p++)
      {
	if (last!=*p && *p!=0)  /* pokud == last, pak jiz jednou delete probehl, pokud == 0 pak zadny identifikator neni v dane strance*/
	  {
	    table->file_remove(table,*p,fi_key);
#if 0
#ifdef CHAIN_BOUND
	    uint32_t chpg  = get_chain_page_address(table->file,*p);
	    if (chpg && chpg!=0xffffffff)
	      table->chain_file_remove(table,chpg,fi_key);
#endif
#endif
	    last=*p;
	  }
      }
  }
  xfree(pages);

  /*   for (fi=(file_item**) fi_key;*fi;fi++,key_length++) { }  */

  return 0;
}


#if 0
int chain_find(hash_table* table
	       , const void* key
	       , int (*chain_found_item_callback) (uint32_t adr)
	       )
{
  int ret=0;
  if (!table||!key) { printf("find spatne zavolan\n"); ABORT }
  unsigned int hash    = table->hash_fce(key);
  unsigned int index   = get_index(table,hash);
  uint32_t     pg_adr  = get_address_at_folder_index(table->file,get_folder_address(table),index);

  if (!pg_adr) {  return 0;  } 
  pg_adr = get_chain_page_address(table->file,pg_adr);

  if (pg_adr)
    do {
      uint32_t aadr = pg_adr;
      uint32_t adr  = skip_page_head(aadr);
      const void* act_key=table->get_chain_key(table->file,adr);
      adr+=sizeof(uint32_t); /* posunuti o identifikator */
      
      if (!table->compar_key(act_key,key
#warning !!!!!
			       , 0
			       , hash))
	{
	  do {
	    //	    printf("vypisuju stranku...\n");
	    for (; table->has_chain_item(table->file,adr,aadr)
		   ; adr=table->next_chain_item(table->file,adr))
	      {
		char tmp[MAX_ITEM_SIZE]; memset(tmp,0,MAX_ITEM_SIZE);

		void* act=(void*) tmp;
		table->get_chain_item(table->file,adr,act,act_key);
		ret++;
		found_item_callback?
		  found_item_callback( table->find_return(table->file,adr) )
		  :0;
	      }
	    aadr=get_next_page_address(table->file, aadr);
	    adr = skip_page_head(aadr);
	  } while (aadr);
	  break; // vse je v jednom retizku...
	}
    } while ((pg_adr=get_chain_page_address(table->file,pg_adr)));
  return ret;
}
#endif

#if 1
int chain_insert(hash_table* table
		 , void* item
		 , void*    (* NEXT_CHAIN_ITEM) (void *chain)
		 , int      (* WRITE_CHAIN_ITEM) (file_t* file, uint32_t kam, void* co)
		 , uint32_t (* GET_IDENTIFIER) (void* item)
		 , unsigned int ITEM_LENGTH
		 )
{
  file_t* file;
  unsigned int hash;
  unsigned int index;
  uint32_t adr;

  if (!table->allow_chain) 
    {
      fprintf(stderr,"SHOULD NOT be called!\n");
      ABORT;
    }
  if (!table || !item ) ABORT

  uint32_t ident=GET_IDENTIFIER(item);
  file  = table->file;
  
  hash  = table->hash_fce(&ident);
  index = get_index(table,hash);
  adr   = get_address_at_folder_index(table->file
				      , get_folder_address(table),index);
  if (!adr) /* neexistuje stranka hlavniho retezce */
    {
      adr = page_alloc(file);
      set_address_at_folder_index(table->file, get_folder_address(table), index, adr);
    }

  uint32_t chain_adr;

#ifdef OLD_CHAIN
  { /* napojeni retizku ... musim ale overit zda nekde neexistuje*/
#ifdef MY_WARNINGS
#warning tady je rest...
#endif
    uint32_t nadr = page_alloc(file);
    chain_adr     = get_chain_page_address(file,adr);
    set_chain_page_address (file, nadr, chain_adr);
    set_chain_page_address (file, adr , nadr);
    chain_adr     = nadr;
    
    if (0) {
      nadr=adr;
      printf("novy retizek: %X",nadr);
      while (nadr)
	{
	  nadr=get_chain_page_address(file,nadr);
	  printf("->%X",nadr);
	}
      printf("\n");
    }
  }
#else
  chain_adr = page_alloc(file);
  const void* fake_item = table->create_chain_item(ident,chain_adr);
#ifdef MY_WARNS
#warning READ THIS - POSSIBLE UPDATE ERROR
#endif
  /* 
     je nutne zkontrolovat, zda jiz zaznam neexistuje, 
     nebot jinak by nebyl novy nemusel byt zapsan...
     nyni se ale tyto kontroly neprovadeji proto je korektni, 
     casem ale byt nemusi
  */
  insert(table,fake_item);

#endif
  /* mam naalokovanou stranku do niz budu vkladat retizky -> vkladam */
  
  {
    uint32_t act_adr = skip_page_head(chain_adr);
    int free_space = PAGE_SIZE - (act_adr-chain_adr);
#ifdef OLD_CHAIN
    uint32_t ident = GET_IDENTIFIER(item);
    write_int(file,act_adr,ident);
    act_adr+=sizeof(uint32_t);   free_space-=sizeof(uint32_t);
#endif
    do
      {
	/* int r = */ WRITE_CHAIN_ITEM ( file, act_adr, item);
	act_adr+=ITEM_LENGTH;   free_space-=ITEM_LENGTH;
	if (free_space <=ITEM_LENGTH)
	  {
	    //	    printf("naalokovavam dalsi stranku\n");
	    act_adr = page_alloc(file); /* alokuji */
	    set_next_page_address (file, chain_adr, act_adr); /* napojim na starou */
	    chain_adr=act_adr; /* naalokovana je koncova */

	    act_adr = skip_page_head(chain_adr); /* nastavim pozici pro item */
	    free_space = PAGE_SIZE - (act_adr-chain_adr);
	  }

	item = NEXT_CHAIN_ITEM(item);
      } while (item);
  }
  return 0;
}
#endif

void create_folder(hash_table* table, unsigned int bits)
{
  if (!table) ABORT
  file_t* file=table->file;
  if (bits>=MAX_BITS) 
    {
      printf("cannot allocate space for folder of size 2^%d\n",bits);
      exit(1);
    }

  if (get_folder_address(table)) return;
  uint32_t adr=folder_page_alloc(file , bits);
  
  set_folder_address(table,adr);
}

