/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

//#define DEBUG_FILE_HEAD
/*
  Implementace databze
  pamov mapovan databze + hlavika v n jsou adresy adres pslunch heovacch tabulek 
 */
#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>

#include "base/macros.h"
#include "database/fagin.h"
#include "database/file.h"

#define init_file_size PAGE_SIZE

typedef struct {
  /* ve strance je vsak alokovan - kvuli zarovnani cely uint32_t */ 
  unsigned char* hash_bites;
  unsigned char* usr;

#ifdef MY_WARNINGS
#warning urcuje maximalni velikost stranky na 2^16 bitu
#endif
  uint16_t* last_free; 

  /* udava adresu dalsi stranky v koliznim retezci (popr. ve free_space) */
  /* je li vsak stranka zaplnovana a neni plna
     , muze byt polozka vyuzita jako adresa prvniho volneho mista
     pokud je pak optimalizovany build a vim, ze se polozky neopakuji
     nemusim prochazet cely retezec a zrychli se build
  */
  uint32_t* next_page_address__be; // next page (considered free_space)
  uint32_t* chain_page_address__be;
  uint32_t page_address;
} file_page;
static file_page read_page(file_t* file, uint32_t page_address);

typedef struct {
  unsigned char* hash_bites__; /* kvuli zarovnani ale bude pouzivat sizeof(int)*/
  uint32_t* next_page_address__be__; // next semantic same page
  uint32_t* content_data__;          // content of the folder
} folder_page;
static folder_page read_folder_page(file_t* file, uint32_t folder_address);

static file_head read_file_head(file_t* file);


static inline uint16_t be16int(uint16_t i)
{
#ifdef ENDIAN
  return ntohs(i);
#else
  return i;
#endif
}

static inline uint16_t int16be(uint16_t i)
{
#ifdef ENDIAN
  return htons(i);
#else
  return i;
#endif
}

//static 
inline uint32_t be2int(uint32_t i)
{
#ifdef ENDIAN
  return ntohl(i);
#else
  return i;
#endif
}

static inline uint32_t int2be(uint32_t i)
{
#ifdef ENDIAN
  return htonl(i);
#else
  return i;
#endif
}

static uint32_t get_next_folder_address(file_t* file, uint32_t adr);
#ifdef FOLDER_NOT_INSIDE_PAGE
static void set_next_folder_address(file_t* file, uint32_t adr, uint32_t data);
#endif

/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
#ifdef ENABLE_DATABASE_PROTECT
/*
  funkce uren k zamezen pstupu k databzi jinak ne pes pstupov funkce
  pro normln bh programu by mly bt vypnut, nebo mnostv chrnn pamti je pli velik 
  a ovlivnilo by to rychlost operac nad databz
 */
static void protect_on(file_t* file)
{
  if (mprotect(file->realfile,file->sb.st_size,PROT_NONE)) abort();
}

static void inline protect_off(file_t* file)
{
  if (mprotect(file->realfile,file->sb.st_size,PROT_WRITE|PROT_READ)) abort();

}

static void inline protect_read(file_t* file)
{
  if (mprotect(file->realfile,file->sb.st_size,PROT_READ)) abort();
}
#endif
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////

/*
  namapuje databzi a vytvo hlaviku pokud neexistovala
 */

static int allow_write ;
static int allow_create;

void set_allow_write (int i) {allow_write =i;}
void set_allow_create(int i) {allow_create=i;}
int  get_allow_write () {return allow_write ;}
int  get_allow_create() {return allow_create;}

/* nastavi databazi jako nikym nepouzivanou */
static void file_reset_access(file_t* file)
{

}
static int file_set_access(file_t* file, int read_only)
{
  return 0;
}

static int file_get_access(file_t* file, int read_only)
{
  return 0;
}

static void file_unset_access(file_t* file, int read_only)
{
  file_reset_access(file);
}

file_t open_file(const char* filename)
{
  //    printf("open file\n");
  file_t ret;
  int new_file=0;
  ret.filename=filename;
  if (allow_write)      
    {
      ret.fd=open(filename,O_RDWR);
      //         printf("O_RDWR\n");
    }
  else
    {
      ret.fd=open(filename,O_RDONLY);
      //          printf("O_RDONLY\n");
    }
  ret.realfile=0;

  if (ret.fd<0 && allow_create && allow_write)
    {
      //           printf("O_RDWR|O_CREAT,S_IRUSR|S_IWUSR\n");
      ret.fd=open(filename,O_RDWR|O_CREAT,S_IRUSR|S_IWUSR);
      if (ret.fd<0)
	{
	  fprintf(stderr, "couldn't open nor create file %s\n",filename);
	  FPRINTF_ERROR;
	  exit(1);
	}
      if ( ftruncate(ret.fd,init_file_size) <0)
	{
	  fprintf(stderr, "couldn't allocate space for file %s\n",filename);
	  FPRINTF_ERROR;
	  exit(1);
	}
      new_file=1;
    }
  
  if (fstat(ret.fd,&ret.sb)<0)
    {
      char buffer[MAX_LEN];
      fprintf(stderr, "couldn't stat file %s in directory %s\n",filename,getcwd(buffer,MAX_LEN));
      FPRINTF_ERROR;
      exit(1);
    }
  if (((ret.realfile) = (char*) mmap(ret.realfile,
				     ret.sb.st_size,
				     allow_write?PROT_WRITE|PROT_READ:PROT_READ,
				     MAP_SHARED,
				     ret.fd,
				     0))==(void*)-1)
    {
      fprintf(stderr, "coulnd't map file %s\n",filename);
      exit(1);
    }
  if (!ret.realfile) 
    {
      fprintf(stderr,"file not mapped !!!\n");
      exit(1);
    }

  if (new_file)
    {
      strcpy(ret.realfile, CSPOT_DB_NAME);
      *(uint32_t*) (ret.realfile+SIZEOF_CSPOT_DB_NAME)=int2be(PAGE_SIZE);
    }
  else
    {
      if (strcmp(ret.realfile,CSPOT_DB_NAME))
	{
	  fprintf(stderr, "not a database file %s\n",filename);
	  exit(1);
	}
      if (PAGE_SIZE!=be2int(*(uint32_t*) (ret.realfile+SIZEOF_CSPOT_DB_NAME)))
	{
	  fprintf(stderr,
		  "database has different page size than expected\t expect %d but has %d\n",
		  PAGE_SIZE,be2int(*(uint32_t*) (ret.realfile+SIZEOF_CSPOT_DB_NAME)));
	  exit(1);
	}
      
    }
  ret.head=read_file_head(&ret);

  if ( file_get_access(&ret,0) )
  {
      fprintf(stderr,"file is being used read_write by another process !!!\n");
      exit(1);    
  }

  if (allow_write)     file_set_access(&ret,0);
  else                 file_set_access(&ret,1);
  return ret;
}

/*
  len>0 zvt databzi file o velikost len 
  len<0 zmn velikost file na len 
      (primrn ureno k smazn databze a vytoen nov)
      avsak takto jiz se nepouziva
 */
int enlarge_file(file_t* file, int len)
{
#ifdef PRINTF_ENLARGE
      printf("enlarge file\n");
#endif
  int size;
  if (len>0)    size=len+file->sb.st_size;
  else          size = -len;
  if ( munmap(file->realfile,file->sb.st_size) < 0 )
    {
      fprintf(stderr, "coulnd't unmap file %s\n",file->filename);
      exit(1);
      return 1;
    }
 if ( ftruncate(file->fd,size) <0)
    {
      fprintf(stderr, "coulnd't allocate space for file %s\n",file->filename);
      exit(2);
      return 2;
    }

  if (fstat(file->fd,&file->sb)<0)
    {
      fprintf(stderr, "coulnd't stat file %s\n",file->filename);
      exit(3);
      return 3;
    }

/*   printf("size=%d\n", (int)file->sb.st_size); */
  int fsize=(int)file->sb.st_size;
  if (fsize != size ) printf("divne %d ~ %d\n",size,fsize);
  if (((file->realfile) = (char*) mmap(NULL,
				       size,
				       allow_write?PROT_WRITE|PROT_READ:PROT_READ,
				       MAP_SHARED ,
				       file->fd,
				       0))==(void*)-1)
    {
      fprintf(stderr, "coulnd't map file %s\n",file->filename);
      exit(4);
      return 4;
    }
  file->head=read_file_head(file);
  return 0;
}

int shrink_file(file_t* file, int len)
{
  return enlarge_file(file,-len);
}

void close_file(file_t* file)
{
  if (!file->realfile) return;
  if (allow_write)    
    file_unset_access(file,0);
  else
    file_unset_access(file,1);

  if ( munmap(file->realfile,file->sb.st_size) < 0 )
    {
      fprintf(stderr, "coulnd't unmap file %s\n",file->filename);
    }
  close(file->fd);
}

/*
  vrt strnku na adrese page_address
 */
static file_page read_page(file_t* file, uint32_t page_address)
{
  if (!page_address) ABORT
  if (!file->realfile) ABORT
  char* act=file->realfile+page_address;
  file_page pg;
  /* promenne zjednodusujici pristup k nekterym polozkam stranky */
  pg.hash_bites     = (unsigned char*) (act  );
  pg.usr            = (unsigned char*) (act+1); /* zatim nepouzita */
  pg.last_free      = (uint16_t*) (act+2);
  pg.next_page_address__be  = (uint32_t*)(act +   sizeof(uint32_t));
  pg.chain_page_address__be = (uint32_t*)(act + 2*sizeof(uint32_t));
  pg.page_address=page_address;
  return pg;
}

/* funkce volana po namapovani databaze */
file_head read_file_head(file_t* file)
{
  file_head ret;
  uint32_t* act=(uint32_t*) (file->realfile + SIZEOF_CSPOT_DB_NAME);
  if (!file) ABORT;
  if (!file->realfile) ABORT;
#define page_free(pol) ret.free_page__be=(uint32_t*) (uint32_t*) (file->realfile+pol) ; act++;
#define string_free_page(pol) ret.string_free_page=(uint32_t*) (file->realfile+pol) ; act++;
#define hash_free(x) ret.free_folder_page_ ## x ## __be=act;act++;
#define table(x,polX,unused) ret. x ## _table_address__be = (uint32_t*)(file->realfile + polX) ; act++;
  /*act++ nutne protoze hash_table je az za temato tabulkama */
  FILE_HEAD;
#undef hash_free
#undef table
#undef page_free
#undef string_free_page

#ifdef DEBUG_FILE_HEAD /* vypise polohy adresaru */
  printf("\n\n");
#define page_free(pol) printf("free_page = %X\n",be2int(*ret.free_page__be));
#define string_free_page(pol) printf("string_free_page = %X\n",be2int(*ret.string_free_page));
#define hash_free(x) printf("hash_free(%d) = %X\n",x,be2int(*ret.free_folder_page_ ## x ## __be));
#define table(x,polX,unused) printf("table %s \t%X -> %X\n",#x,polX,be2int(*ret. x ## _table_address__be));
  /*act++ nutne protoze hash_table je az za temato tabulkama */
  FILE_HEAD;
#undef hash_free
#undef table
#undef page_free
#undef string_free_page
#endif

  return ret;
}
/*
  vrt adresu nsledujc nepopuit ji naalokovan strnky
 */
uint32_t next_free_page(file_t* file,uint32_t adr)
{
  if (!adr) ABORT
  file_page pg = read_page(file, adr);
  uint32_t a=pg.next_page_address__be?be2int(*pg.next_page_address__be):0;
  return a;
}
/*
  pokud existuje nepouit naalokovan strnka vrt jej adresu
  jinak zvt databzi a vrt adresu nov naalokovan strnky
 */
uint32_t page_alloc(file_t* file)
{
  if (!file || !file->realfile) ABORT
  uint32_t adr=0;
  if (*file->head.free_page__be && *file->head.free_page__be!=-1)
    {
      adr=be2int(*file->head.free_page__be);
      uint32_t next=next_free_page(file,adr);
      *file->head.free_page__be=int2be(next);
      memset(file->realfile+adr,0,PAGE_SIZE); 
      return adr;
    }
  else
    {
      adr=file->sb.st_size;  // new page will be allocated at the end of file
#ifdef DATABASE_ADDITIONAL_INFO
#define AA "page_start"
#define AAA sizeof(AA)
#define BB "page_end"
#define BBB sizeof(BB)
      enlarge_file(file,PAGE_SIZE+AAA+BBB);
      strcpy(file->realfile+adr,AA);
      strcpy(file->realfile+adr+AAA+PAGE_SIZE,BB);
      return adr+AAA;
#undef AA
#undef AAA
#undef BB
#undef BBB	      
#else 
      enlarge_file(file,PAGE_SIZE);
      return adr;
#endif
    }
}
/*
  vymae obsah strnky a zaad ji mezi nepouit naalokovan
 */
void page_free(file_t* file, uint32_t page_address)
{
  if (!page_address) { fprintf(stderr,"freeing address 0 !\n"); ABORT}
  memset(file->realfile+page_address,255,PAGE_SIZE);
  file_page pg = read_page(file, page_address);    
  *pg.next_page_address__be=*file->head.free_page__be;
  *file->head.free_page__be=int2be(page_address);
}



extern int updating;
void page_prealloc(file_t* file, int number)
{
  if(updating) return ;
#ifdef PRINT_PREALLOC
  fprintf(stderr,"preallocating %d\n",number);
#endif
  if (!file || !file->realfile) ABORT;
  if (number <= 0) return ;
  
  uint32_t adr = file->sb.st_size;  // new page will be allocated at the end of file
  enlarge_file(file,PAGE_SIZE*number);
  int i;
  for ( i = 0; i<number; i++)
    {
      page_free(file,adr+PAGE_SIZE*i);
    }
}


uint32_t next_free_folder(file_t* file, uint32_t adr)
{
  uint32_t a=get_next_folder_address(file,adr);
  return a;
}


folder_page read_folder_page(file_t* file, uint32_t folder_address)
{
  if (!file || !folder_address) ABORT
  folder_page f;
  char* act=file->realfile+folder_address;
  f.hash_bites__ = (unsigned char*) act  + sizeof(uint32_t) ; ;
  f.next_page_address__be__=(uint32_t*) act; // + sizeof(uint32_t) ;
  f.content_data__ =(uint32_t*) (act+ sizeof(uint32_t) + sizeof(uint32_t) ) ;
  return f;
}

unsigned char get_hash_bites_folder(file_t* file, uint32_t adr)
{
  if (!adr || !file) ABORT
  folder_page pg  = read_folder_page(file,adr);
  unsigned char c = *pg.hash_bites__;
  return c;
}

void set_hash_bites_folder(file_t* file, uint32_t adr, unsigned char c)
{
  if (!adr) ABORT
  folder_page pg   = read_folder_page(file,adr);
  *pg.hash_bites__ = c;
}

uint32_t get_next_folder_address(file_t* file, uint32_t adr)
{
  folder_page pg = read_folder_page(file,adr);
  uint32_t a     = be2int(*pg.next_page_address__be__);
  return a;
}

#ifdef FOLDER_NOT_INSIDE_PAGE
void set_next_folder_address(file_t* file, uint32_t adr, uint32_t data)
{
#if 0
  if (!adr) return;
#else
  if (!adr) { printf("nill address as a parametr of set_next_folder_address\n"); ABORT }
#endif
  folder_page pg              = read_folder_page(file,adr);
  *pg.next_page_address__be__ = int2be(data);
}
#endif

uint32_t get_address_at_folder_index(file_t* file, uint32_t adr, int index)
{
  if (!adr || !file) ABORT
  folder_page pg=read_folder_page(file,adr);
  return be2int(*(pg.content_data__+index));
}

void set_address_at_folder_index(file_t* file
				 , uint32_t adr
				 , int index
				 , uint32_t data)
{
  if (!adr)  { printf("nill address (adr) as a parametr of set_address_at_folder_address\n"); ABORT }
  //  if (!data) { printf("nill address (data) as a parametr of set_next_folder_address\n"); ABORT }

  folder_page pg             = read_folder_page(file,adr);
  *(pg.content_data__+index) = int2be(data);
}

uint32_t get_folder_address(hash_table* t)
{
  if (!t) ABORT 
  return be2int(**t->i_folder_address);
}

void set_folder_address(hash_table* t,uint32_t adr)
{
  if (!adr) ABORT
   **t->i_folder_address=int2be(adr);
}

void read_table(hash_table* tb, file_t* file, uint32_t** i_folder_address)
{
  if (!tb || !file) ABORT
  //  if (!i_folder_address || *i_folder_address) ABORT
  tb->file=file;
  tb->i_folder_address=i_folder_address;
  /*
    i_folder_address is a pointer to mapped file
    writing to **i_folder_address is writing into file ... 
    reading from **i_folder_address is reading from file ...
    THEREFORE WE MUST NOT FORGET TO REINITIALIZE THIS POINTER EVERYTIME AFTER MAPPING FILE
   */
#if 1      

#ifdef MY_WARNS
#warning tady zmena ale create_folder by to mel obstarat...
#warning promyslet poradne jestli neni jeste nekde potreba nejaky kus kodu
#endif

      return;
#else
  if (!**i_folder_address)
    {
      printf("**i_folder_address==0 -> allocate\n");
      uint32_t adr=folder_page_alloc(file, 0);
      set_folder_address(tb,adr);

      // budou to nil pages... analokuji se az budou potreba
      //      uint32_t page_adr=page_alloc(file);
      //      set_address_at_folder_index(file,adr,0,page_adr);
    }
  else
    {
      printf("folder jiz znam ... %X   ",be2int(**i_folder_address));
      printf("jeho adresa byla predana v parametru\n");
    }
#endif

}

void write_string(file_t* file, uint32_t adr, const char* str)
{
#ifdef CHECK_BOUNDS
  if (adr+strlen(str)>=file->sb.st_size) { fprintf(stderr,"zapis za koncem souboru adr=%X size=%X\n",adr,(uint32_t)file->sb.st_size); ABORT }
#endif
  strcpy(file->realfile+adr,str);
}

const char* read_string(file_t* file, uint32_t adr)
{
#ifdef CHECK_BOUNDS
  if (adr+strlen(file->realfile+adr)>=file->sb.st_size) 
    { 
      fprintf(stderr,"cteni za koncem souboru adr=%X size=%X\n",adr,(uint32_t)file->sb.st_size); 
      ABORT ;
    }
#endif
/*   printf("read stirng at %X\n",adr); */
  return file->realfile+adr;
}

void write_int(file_t* file, uint32_t adr, uint32_t w)
{
#ifdef CHECK_BOUNDS
  if (adr+sizeof(uint32_t)>file->sb.st_size) { fprintf(stderr,"zapis za koncem souboru adr=%X size=%X\n",adr,(uint32_t)file->sb.st_size); ABORT }
#endif
  *(uint32_t*) (file->realfile+adr)=int2be(w);
}

uint32_t read_int(file_t* file, uint32_t adr)
{
#ifdef CHECK_BOUNDS
  if (adr+sizeof(uint32_t)>file->sb.st_size) { fprintf(stderr,"cteni za koncem souboru adr=%X size=%X\n",adr,(uint32_t)file->sb.st_size); ABORT }
#endif
  uint32_t ret= be2int(*(uint32_t*) (file->realfile+adr));
  return ret;
}

unsigned char get_page_bites(file_t* file, uint32_t adr)
{
  file_page pg=read_page(file,adr);
  unsigned char a = *pg.hash_bites;
  return a;
}

unsigned char set_page_bites(file_t* file, uint32_t adr, unsigned char a)
{
  file_page pg=read_page(file,adr);
  *pg.hash_bites=a;
  return a;
}


void delete_file(file_t* file)
{
  const char* filename=file->filename;
  close_file(file);
  unlink(filename);
}

void file_clean(file_t* file)
{
  return delete_file(file);

  /* neco spatne zrejme ve shrink -enlarge nema rado zmensovani!*/
  shrink_file(file,PAGE_SIZE);
  printf("predchozi radek\n");
  printf("file size=%d\n",(int)file->sb.st_size);
  printf("erase from %d, lenght  %d\n",DATABASE_HEAD,  PAGE_SIZE-1-DATABASE_HEAD );
  memset(file->realfile+DATABASE_HEAD,0,PAGE_SIZE-1-DATABASE_HEAD);
  close_file(file);
}







#ifdef FOLDER_NOT_INSIDE_PAGE
uint32_t folder_page_alloc(file_t* file, unsigned char hash_bites)
{
  if (hash_bites>MAX_BITS) ABORT
#ifdef DATABASE_ADDITIONAL_INFO
#define AA "FOLDER_START"
#define AAA sizeof(AA)
#define BB "FOLDER_END"
#define BBB sizeof(BB)
#define hash_free(x) case x:\
if (*file->head.free_folder_page_ ## x ## __be && *file->head.free_folder_page_ ## x ## __be!=-1)  \
{\
adr=be2int(*file->head.free_folder_page_ ## x ## __be);\
int i=next_free_folder(file,adr);\
*file->head.free_folder_page_ ## x ## __be=i;\
*((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
i=new_folder_size(hash_bites);\
memset(file->realfile+adr,0,i); \
*((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
return adr;\
}\
else\
{\
adr=file->sb.st_size;\
int len=new_folder_size(hash_bites);\
 enlarge_file(file,len+AAA+BBB);\
  strcpy(file->realfile+adr, AA );\
  strcpy(file->realfile+adr+AAA+len,BB);\
*((unsigned char*)file->realfile+adr+AAA+sizeof(uint32_t))=hash_bites;\
 return adr+AAA;\
}\
break
#else
#define hash_free(x) case x:\
if (*file->head.free_folder_page_ ## x ## __be && *file->head.free_folder_page_ ## x ## __be!=-1)  \
{\
adr=be2int(*file->head.free_folder_page_ ## x ## __be);\
int i=next_free_folder(file,adr);\
*file->head.free_folder_page_ ## x ## __be=i;\
*((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
i=new_folder_size(hash_bites);\
memset(file->realfile+adr,0,i); \
*((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
return adr;\
}\
else\
{\
adr=file->sb.st_size;\
int len=new_folder_size(hash_bites);\
 enlarge_file(file,len);\
*((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
 return adr;\
}\
break 
#endif


  uint32_t adr;
  switch ((int)hash_bites)
    {
      HASH_HEAD;
    default:
      fprintf(stderr,"alloc default option %d\n",hash_bites);
      return 0;
    }
  protect_on(file);
#undef hash_free
#undef AA
#undef AAA
#undef BB
#undef BBB	   
}

void folder_page_free(file_t* file, uint32_t page_address)
{
  if (!page_address) ABORT

  unsigned char prev_hashed_bites=get_hash_bites_folder(file,page_address);
#define hash_free(x) case x :\
set_next_folder_address(file,page_address,be2int(*file->head.free_folder_page_ ## x ## __be));\
*file->head.free_folder_page_ ## x ## __be=int2be(page_address);\
break


  switch (get_hash_bites_folder(file,page_address))
    {
      HASH_HEAD;
    default:
      fprintf(stderr,"free default option address=%X   pg.hash_bites=%d\n"
	      ,page_address
	      ,get_hash_bites_folder(file,page_address));
      break ;
    }

  memset(file->realfile+page_address
	 	 ,0xFF
	 //	 ,0
	 ,new_folder_size(prev_hashed_bites));

#undef hash_free
}


#else


uint32_t folder_page_alloc(file_t* file, unsigned char hash_bites)
{
  if (hash_bites>MAX_BITS) ABORT

#define hash_free(x) case x:\
{\
  adr=file->sb.st_size;\
  int len=new_folder_size(hash_bites);\
  int size=(1+len/PAGE_SIZE)*PAGE_SIZE;\
/* printf("allocating folder at %X size =%d\n",adr,size); */\
  enlarge_file(file,size);\
 *((unsigned char*)file->realfile+adr+sizeof(uint32_t))=hash_bites;\
  return adr;\
}\
break 


  uint32_t adr;
  switch ((int)hash_bites)
    {
      HASH_HEAD;
    default:
      fprintf(stderr,"alloc default option %d\n",hash_bites);
      return 0;
    }
#undef hash_free
}

void folder_page_free(file_t* file, uint32_t page_address)
{
  if (!page_address) ABORT

#define hash_free(x) case x :\
  {\
    memset(file->realfile+page_address\
	 	 ,0xFF\
	 	 ,new_folder_size(get_hash_bites_folder(file,page_address)));\
    int len=new_folder_size(x);\
    int i; for (i=0;i<(1+len/PAGE_SIZE);i++)\
              {\
                page_free(file,page_address+i*PAGE_SIZE);\
              }\
  }\
  break;

  switch (get_hash_bites_folder(file,page_address))
    {
      HASH_HEAD;
    default:
      fprintf(stderr,"free default option address=%X   pg.hash_bites=%d\n"
	      ,page_address
	      ,get_hash_bites_folder(file,page_address));
      break ;
    }
#undef hash_free
}

void set_free_pos_in_page(file_t* file, uint32_t page_adr, uint32_t adr)
{
  file_page page = read_page(file, page_adr);
  if (!adr)
    {
      *page.last_free=0;
    }
  else
    {
      *page.last_free= int16be(adr-page_adr);
    }
}

uint32_t get_free_pos_in_page(file_t* file, uint32_t adr)
{
  file_page page = read_page(file, adr);
  if (*page.last_free)
    {
      uint32_t r =  adr + be16int( *page.last_free );
      return r;
    }
  return 0;
}

#ifdef MY_WARNS
#warning ODEBRAT FUNKCI
#endif
uint32_t get_chain_page_address(file_t* file, uint32_t adr)
{
#ifdef CHAIN_OFSET
  if (!adr) ABORT
#if 0
  file_page pg=read_page(file,adr);
  uint32_t a= be2int(*pg.chain_page_address__be);
  return a;
#else
  return be2int(*(uint32_t*) (file->realfile+adr+CHAIN_OFSET));
#endif
#else 
  return 0;
#endif
}

#ifdef MY_WARNS
#warning ODEBRAT FUNKCI
#endif
void set_chain_page_address(file_t* file, uint32_t adr, uint32_t data)
{
#ifdef CHAIN_OFSET
  if (!adr) ABORT
  if (!adr) return;
  //  printf("set_chain_page_address adr=%X, data=%X\n",adr,data);
  *(uint32_t*) (file->realfile+adr+CHAIN_OFSET)=int2be(data);
#else
  return;
#endif
}

uint32_t get_next_page_address(file_t* file, uint32_t adr)
{
  file_page pg=read_page(file,adr);
  uint32_t a= be2int(*pg.next_page_address__be);
  return a;
}

void set_next_page_address(file_t* file, uint32_t adr, uint32_t data)
{
#if 1
  if (!adr) return;
#else
  if (!adr) { printf("nill address as a parametr of set_next_page_address\n"); ABORT }
#endif
  file_page pg=read_page(file,adr);
/*   if (*pg.has_free_space)  *pg.has_free_space = 0; */
  *pg.next_page_address__be=int2be(data);
}


/* void switch_add_item_check(hash_table* t, int b) */
/* { */
/*   if (b) t->add_item  = t->add_item_check; */
/*   else   t->add_item  = t->add_item_nocheck; */
/* } */





#endif
