/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#ifndef FILE_H
#define FILE_H
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/stat.h>
void set_allow_write (int i) ;
void set_allow_create(int i) ;
int  get_allow_write () ;
int  get_allow_create() ;


typedef struct {
  /* 
     makro table 
     table( jmeno tabulky, poloha tabulky, typ tabulky )
  */
#define MAX_BITS 20

#define HASH_HEAD\
  hash_free(0);\
  hash_free(1);\
  hash_free(2);\
  hash_free(3);\
  hash_free(4);\
  hash_free(5);\
  hash_free(6);\
  hash_free(7);\
  hash_free(8);\
  hash_free(9);\
  hash_free(10);\
  hash_free(11);\
  hash_free(12);\
  hash_free(13);\
  hash_free(14);\
  hash_free(15);\
  hash_free(16);\
  hash_free(17);\
  hash_free(18);\
  hash_free(19);\
  hash_free(20);\
  hash_free(21);\
  hash_free(22);\
  hash_free(23);\
  hash_free(24);\
  hash_free(25);\
  hash_free(26);\
  hash_free(27);\
  hash_free(28);\
  hash_free(29);\
  hash_free(30);\
  hash_free(31);

  /* string oznacujici soubor za cspot databazi */

  /*
   * DATABASE HEAD
   */
#define CSPOT_DB_NAME "cspot-db"
#define SIZEOF_CSPOT_DB_NAME 3*sizeof(uint32_t)
#define DATABASE_HEAD SIZEOF_CSPOT_DB_NAME+sizeof(uint32_t) /* */

#define SIZE  sizeof(uint32_t)

#define FREE_PAGE        DATABASE_HEAD + SIZE*0
#define STRING_FREE_PAGE DATABASE_HEAD + SIZE*1
#define STRING_FRST_FREE DATABASE_HEAD + SIZE*2
#define STRING_FRST_PAGE DATABASE_HEAD + SIZE*3

#define ACCESS_TYPE      DATABASE_HEAD + SIZE*4
#define UPDATE_COUNT     DATABASE_HEAD + SIZE*5

#define IDENTIFIER       DATABASE_HEAD + SIZE*6 
#define FILES            DATABASE_HEAD + SIZE*7
#define INCLUDE          DATABASE_HEAD + SIZE*8
#define MACRO_DEF        DATABASE_HEAD + SIZE*9
#define MACRO_USE        DATABASE_HEAD + SIZE*10
#define FCE_DEC          DATABASE_HEAD + SIZE*11
#define FCE_DEF          DATABASE_HEAD + SIZE*12
#define FCE_USE          DATABASE_HEAD + SIZE*13
#define VAR_DEC          DATABASE_HEAD + SIZE*14
#define VAR_USE          DATABASE_HEAD + SIZE*15
#define TYPEDEF_         DATABASE_HEAD + SIZE*16
#define TYPEDEF_USE      DATABASE_HEAD + SIZE*17
#define STUCT_USE        DATABASE_HEAD + SIZE*18
#define STUCT_DEC        DATABASE_HEAD + SIZE*19
#define STUCT_DEF        DATABASE_HEAD + SIZE*20

#define INCLUDED         DATABASE_HEAD + SIZE*21

#define TABLES\
  table(identifier,   IDENTIFIER  , string);\
  table(files,        FILES       , struct2);\
  table(included,     INCLUDED    , struct2);\
  table(include ,     INCLUDE     , struct4);\
  table(macro_def,    MACRO_DEF   , struct4);\
  /* table(macro_cnd,    MACRO_CND   , struct4); */\
  table(macro_use,    MACRO_USE   , struct4);\
  table(function_dec, FCE_DEC     , struct4);\
  table(function_def, FCE_DEF     , struct4);\
  table(function_use ,FCE_USE     , struct4);\
  table(variable_dec, VAR_DEC     , struct4);\
  table(variable_use, VAR_USE     , struct7);\
  table(typedef_,     TYPEDEF_    , struct4);\
  table(typedef_use,  TYPEDEF_USE , struct4);\
  table(struct_use,   STUCT_USE   , struct4);\
  table(struct_dec,   STUCT_DEC   , struct4);\
  table(struct_def,   STUCT_DEF   , struct4);\

#define FH\
  page_free (FREE_PAGE); \
  string_free_page(STRING_FREE_PAGE); \
  TABLES

#ifdef FOLDER_NOT_INSIDE_PAGE
   #define FILE_HEAD FH HASH_HEAD 
#else
   #define FILE_HEAD FH
#endif
 
#define hash_free(x) uint32_t* free_folder_page_ ## x ## __be
#define table(x,unused1,unused2) uint32_t* x ## _table_address__be
#define page_free(unused) uint32_t* free_page__be
#define string_free_page(unused) uint32_t* string_free_page

  FILE_HEAD

#undef hash_free
#undef table
#undef page_free
#undef string_free_page
} file_head;

typedef struct {
  const char* filename;
  char* realfile;
  int fd;
  struct stat sb;
  file_head head;
} file_t;




/* soubor funkci a adres pro hesovaci tabulku */
/* kazda tabulka si nese vsechny odkazy na funkce ktere potrebuje */
typedef struct hash_table_struct  {
  uint32_t** i_folder_address;  // indirect address of hash folder
  file_t* file; /* soubor databaze */
  int  allow_chain:1;

  uint32_t (*hash_fce)         ( const void* key); /* funkce pouzita na zahesovani klice */
  const void* (*get_key)       ( const void* item);
  int      (*enough_space)     ( file_t* file, uint32_t adr, const void* item);
  int      (*has_item)         ( file_t* file, uint32_t adr, uint32_t start_page);
  uint32_t (*add_item)         ( file_t* file, uint32_t fpg_adr, const void* item);
  uint32_t (*add_item_nocheck) ( file_t* file, uint32_t fpg_adr, const void* item);
  int (*check_item_inserted)   ( file_t* file, uint32_t fpg_adr, const void* item);

  int      (*is_item_chain_start ) (const void* item);
  uint32_t (*get_item_chain_start) (const void* item);
   void*   (*create_chain_item)    (uint32_t ident, uint32_t address);

  int      (*compar_item)       ( const void* item1, const void* item2);
  uint32_t (*next_item)         ( file_t* file, uint32_t adr);
  void     (*get_item)          ( file_t* file, uint32_t adr, void* item);
  int      (*get_len)           ( const void* item);
  int      (*already_inserted)  ( file_t* file, uint32_t adr, const void* item);
  int      (*compar_key)        ( const void* key1, const void* key2, int hash1, int hash2);
  uint32_t (*find_return)       ( file_t*, uint32_t );
  uint32_t (*remove_item)       ( struct hash_table_struct* table, uint32_t adr, const void* key );
  uint32_t (*chain_remove_item) ( struct hash_table_struct* table, uint32_t adr, const void* key );
  uint32_t (*file_remove)       ( struct hash_table_struct* table, uint32_t adr, const void* key );
  uint32_t (*chain_file_remove) ( struct hash_table_struct* table, uint32_t adr, const void* key );
  uint32_t (*get_hash)          ( file_t* file, uint32_t adr);
  int      (*__inserted)        ( void* , const void* , uint32_t   );
  void     (*__print)           ( file_t* file, uint32_t adr);
  void     (*__protect)         ( const void*);

  uint32_t (*item_hash)         ( const void* item);

  const void* (*get_chain_key)       ( file_t* file, uint32_t adr);
  int      (*has_chain_item)         ( file_t* file, uint32_t adr, uint32_t start_page);
  uint32_t (*next_chain_item)        ( file_t* file, uint32_t adr);
  void     (*get_chain_item)         ( file_t* file, uint32_t adr, void* item, const void* key);

#ifdef MY_WARNINGS
#warning check if functions are implemented
#endif
  uint32_t (*add_item_into_chain) ( file_t* file, uint32_t fpg_adr, const void* item);
  int      (*is_item_in_this_address_of_chain) (file_t* file, uint32_t chain_adr, const void* item);
#define MAX_ITEM_SIZE PAGE_SIZE+50 /* potreba ve faginovi pro alokovani polozek na zasobniku (je jich malo) */
} hash_table;

void     file_clean(file_t* file);
void     read_table(hash_table* tb, file_t* file, uint32_t** table_address);

file_t   open_file    ( const char* filename  );
void     close_file   ( file_t* file          );
int      enlarge_file ( file_t* file, int len );

/* folder access functions */
uint32_t folder_page_alloc(file_t* file, unsigned char used_bites);
void     folder_page_free(file_t* file, uint32_t page_address);

unsigned char get_hash_bites_folder(file_t*, uint32_t adr);
void          set_hash_bites_folder(file_t*, uint32_t adr, unsigned char c);

uint32_t get_address_at_folder_index(file_t* file, uint32_t adr, int index);
void     set_address_at_folder_index(file_t* file, uint32_t adr, int index, uint32_t data);

uint32_t get_folder_address(hash_table* t);
void     set_folder_address(hash_table* t,uint32_t);
/* END */


/* page access functions */
uint32_t page_alloc(file_t* file); /* vraci adresu stranky */
void     page_free(file_t* file, uint32_t page_address);
void     page_prealloc(file_t* file, int number);

uint32_t get_next_page_address(file_t* file, uint32_t adr);
void     set_next_page_address(file_t* file, uint32_t adr, uint32_t data);

uint32_t get_chain_page_address(file_t* file, uint32_t adr);
void     set_chain_page_address(file_t* file, uint32_t adr, uint32_t data);

unsigned char get_page_bites(file_t* file, uint32_t adr);
unsigned char set_page_bites(file_t* file, uint32_t adr, unsigned char a);
/* END */


void        write_string(file_t* file, uint32_t adr, const char* str);
const char* read_string(file_t* file, uint32_t adr);
void     write_int(file_t* file, uint32_t adr, uint32_t w);
uint32_t read_int(file_t* file, uint32_t adr);


//#define new_folder_size(hash_bites) (((( 1<<hash_bites ))*sizeof(uint32_t))+ sizeof(uint32_t) + 1)
#define new_folder_size(hash_bites) (((( 1<<hash_bites ))*sizeof(uint32_t))+ sizeof(uint32_t) + sizeof(uint32_t))

/* po ad */
/* pocet bitu dane stranky (zarovnano na 4 bity) + adresa nasledujici stranky + adresa prvni stranky s dlouhymi retizky */


/* 
   strnka fagina
 */
#define INFO_OFFSET sizeof(uint32_t)*0
#define INFO_HASH_BITES_OFFSET 0 /* byte */
#define INFO_USR_OFFSET 1 /* byte */
#define INFO_LAST_FREE_OFFSET 2 /* uint16_t */
#define NEXT_OFSET  sizeof(uint32_t)*1
//#define CHAIN_OFSET sizeof(uint32_t)*2

#ifdef CHAIN_OFSET
#define skip_page_head(x) (x+3*sizeof(uint32_t))
#else
#define skip_page_head(x) (x+2*sizeof(uint32_t))
#endif
#define skip_string_page_head(x) (x+sizeof(uint32_t))

uint32_t get_free_pos_in_page(file_t* file, uint32_t adr);
void     set_free_pos_in_page(file_t* file, uint32_t page_adr, uint32_t adr);


//#ifdef  ALLIGN_STRINGS  
#if 1
#define ALLIGN_STRING_ADR(x,free_space)       \
      {                                           \
	int a = sizeof(int) - (x % sizeof(int));  \
	if ( a != sizeof(int))                    \
	  {                                       \
	    x+=a;                                 \
	    free_space-=a;                        \
	    if (free_space<0) free_space=0;       \
	  }                                       \
      }
#else
#define ALLIGN_STRING_ADDRESS(x) x
#endif
/* 
   pri vklavadni do hesovaci tabulky se bude provadet kontrola 
   (pozor to neni tak docela pravda-jen v posledni kolizni strance)
   zda bylo vlozeno
*/
void switch_add_item_check(hash_table* t, int b);
#endif
