/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include "base/xalloc.h"
#include "database/database.h"
/*
  predpoklada nadefinovane makro STRUCT_NAME 
  poskytuje funkce
  get_${STRUCT_NAME}
  add_${STRUCT_NAME}
  save_${STRUCT_NAME}
  init_${STRUCT_NAME}
  destroy_${STRUCT_NAME}

  ostatni funkce jsou pro rb-stromy a nejsou zamyslene k jinemu pouzivani 
 */

/* poskytuje funkce */

extern const char** readd_deleted_files; 
extern int not_add_include_into_db; 
void create_folder(hash_table* table, unsigned int bits);

#define INIT_MACROS
#include "struct_macros.h"
static int NUMBER_ITEMS;
static int NUMBER_COLLISIONS;

#define STD_ALLOC

#define STRUCT_SIZE (sizeof(uint32_t)*2)
static xalloc_context ALLOC_CONTEXT;
/* ///////////////////////////////////////////////////////// */

typedef struct {
  string_t* ident;
  string_t* file;
  void *    next_same;
  int       no_add;
} STRUCT ;

int  GET (const char* ident);
int  ADD (const char* file
	  , const char* ident);
static void SAVE ();
static void INIT ();
static void DESTROY ();



static int   MYCMP (string_t* k1, string_t* k2 ) ;
static void  MYINITKEY (STRUCT * s, string_t* i) ;
static void  INIT_DATA (STRUCT * s) { s->next_same=0; s->file=0; }
static void* ALLOC(unsigned int size) ;
static void  FREE(void*);

#define TREE_NODE STRUCT
#define TREE_PREFIX(x) TREE_PREF(x) 

#define TREE_KEY_COMPLEX(x) x ident
#define TREE_KEY_DECL string_t* ident
#define TREE_GIVE_CMP MYCMP
#define TREE_GIVE_INIT_KEY MYINITKEY
#define TREE_GIVE_INIT_DATA

#define TREE_WANT_CLEANUP
#define TREE_WANT_LOOKUP
#define TREE_WANT_FIND

#define TREE_WANT_ITERATOR
#define TREE_STATIC
#define TREE_CONSERVE_SPACE

#define TREE_GIVE_ALLOC
#include "rb-tree.h"

struct TREE_PREF(tree) TREE_NAME ;

static int MYCMP (string_t* k1, string_t* k2 ) 
{
  int i;
  const char *str1, *str2;
  if (k1->typ==DB) str1=read_string(&file,k1->adr);
  else str1=k1->str;

  if (k2->typ==DB) str2=read_string(&file,k2->adr);
  else str2=k2->str;

  if ((i=strcmp(str1 , str2))) return i;
  return 0;
}


int  ADD (const char* file
	  , const char* ident)
{
  string_t* id_s = make_string(ident);
  string_t* fl_s = make_string(file);

  STRUCT *item = TREE_PREF(lookup) (& TREE_NAME , id_s);
  if (!item->file) /* lookup tento node vytvoril */
    {
      item->file=fl_s; item->ident=id_s;  item->next_same=0;
      NUMBER_ITEMS++;
      return 0;
    }
  else /* je nutne vytvorit dalsi a navazat v seznamu */
    {
#ifdef MY_WARNINGS
#warning linearni seznam a musim ho cely projit abych vedel zda neni zaznam vlozen -> pomale!!!
#endif
      STRUCT *prev=0, *next=item;
      while (next)
	{
	  if (next->file==fl_s)	    return 1;
	  else
	    {
	      prev=next;
	      next=(STRUCT*) next->next_same;
	    }
	}
	{/* pridani dalsiho prvku do spojoveho seznamu */
	  NUMBER_COLLISIONS++;
	  xalloc_context* context = xalloc_get_context();
	  xalloc_set_context(& ALLOC_CONTEXT);
	  STRUCT *it = xalloc(sizeof( STRUCT ));
	  it->file=fl_s; it->ident=id_s; it->next_same=0;
	  prev->next_same=it;
	  xalloc_set_context(context);
	}
      return 0;
    }
}

int GET (const char* ident)
{
  int ret=0;
  string_t* id_s = seek_string(ident);
  if (!id_s) return ret;
  STRUCT *item = TREE_PREF(find) (& TREE_NAME , id_s);
  STRUCT *next=item;
  while (next)
    {
      ret=1;
      WRITE2( next->ident->str
	     , next->file->str);

      next=(STRUCT*) next->next_same;
    }  
  return ret;
}

static void MYINITKEY (STRUCT * s, string_t* i) 
{
  s->ident=i;
  s->next_same=0;
}

static void * NEXT_CHAIN_ITEM (void *item)
{
  STRUCT *tmp = (STRUCT *) item;
  return tmp->next_same;
}
static int WRITE_CHAIN_ITEM (file_t* file, uint32_t kam, void* co)
{
  if (!file || !co) {printf("addstruct WRITE_CHAIN_ITEM abort();");ABORT}
  STRUCT *tmp = (STRUCT *) co;
  uint32_t fa   = (tmp->file->adr);
  write_int(file,kam+0*sizeof(uint32_t),fa   );
  return 0;
}
static uint32_t GET_IDENTIFIER (void* item)
{
  STRUCT *tmp = (STRUCT *) item;
  return (tmp->ident->adr);
}
static unsigned int ITEM_LENGTH = sizeof(uint32_t) * 3;

static void SAVE () 
{
#ifdef CREATE_FOLDER
  unsigned int records_in_page = (PAGE_SIZE-skip_page_head(0))/STRUCT_SIZE;
  unsigned int required_folder_size = 
    ( NUMBER_ITEMS
#ifdef INCLUDE_COLLISIONS_IN_COUNTING_FOLDER_SIZE
      + NUMBER_COLLISIONS
#endif
      )
    /records_in_page + 1;
#ifdef PREALLOC_PAGES
  page_prealloc(&file,required_folder_size);
#endif
  unsigned int required_folder_bits = 0;
  unsigned int c =1;
  for (required_folder_bits = 0, c=1
	 ; c < required_folder_size
	 ; required_folder_bits++, c*=2 );
  create_folder(&STRUCT_NAME,required_folder_bits);
#endif
  STRUCT *tmp;
  TREE_FOR_ALL( STRUCT_NAME , & TREE_NAME , add)
    {
      tmp=add;
#ifdef CHAIN_BOUND
      int count=0;
      for (;tmp ;count++) tmp=tmp->next_same;
      if ((double) count <= (double) (PAGE_SIZE/STRUCT_SIZE) * CHAIN_BOUND )
#else 
      if (1)
#endif
	{   
	  tmp=add;
	  while (tmp)
	    {
	      uint32_t fa   = (tmp->file->adr);
	      uint32_t ia   = (tmp->ident->adr);
	      /* tady ma byt pridani do file_t */

	      /* not adding if not allowed to add includes and file is na include */
	      if (!tmp->no_add)
		{
		  add_file_ident(tmp->file,tmp->ident);
		  SAVE_DATABASE (ia,fa);
		}
	      else
		{
		  //		  printf("not saving\n");
		}
	      tmp=tmp->next_same;
	    }
	}
      else
	{
#ifdef MY_WARNINGS
#warning neresil jsem no_add	  
#endif
	  //	  printf("saving chain for %X ~ %s\n",add->ident->adr,add->ident->str);
	  SAVE_CHAIN_DATABASE(add,NEXT_CHAIN_ITEM,WRITE_CHAIN_ITEM, GET_IDENTIFIER, ITEM_LENGTH);
	}
    }
  TREE_END_FOR;
}

static void INIT ()
{
  ALLOC_CONTEXT .name=  "addstruct4" ;
#ifndef STD_ALLOC
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
#endif
  TREE_PREF(init) ( & TREE_NAME  );
#ifndef STD_ALLOC
  xalloc_set_context(context);
#endif
}

static void DESTROY ()
{
  TREE_PREF(cleanup) ( & TREE_NAME  );
  memset(& TREE_NAME,0,sizeof(TREE_NAME));

  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  xfree_all();
  xalloc_set_context(context);
}

static void* ALLOC(unsigned int size) {
#ifdef STD_ALLOC
  void *s = malloc(size);
  memset(s,0,size);
  return s;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  void* r = xalloc(size);
  xalloc_set_context(context);
  return r;
#endif
}

static void FREE(void* s)
{
#ifdef STD_ALLOC
  free(s);
  return;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  xfree(s);
  xalloc_set_context(context);
#endif
}

static void CHECK () 
{
  STRUCT *tmp;
  TREE_FOR_ALL( STRUCT_NAME , & TREE_NAME , add)
    {
      tmp=add;
      {   
	tmp=add;
	while (tmp)
	  {
	    uint32_t fa   = (tmp->file->adr);
	    uint32_t ia   = (tmp->ident->adr);
	    
	    /* not adding if not allowed to add includes and file is na include */
	    if (HAS_DATABASE (ia,fa))
	      tmp->no_add=1;
	    tmp=tmp->next_same;
	  }
      }
    }
  TREE_END_FOR;
}


#undef INIT_MACROS
#include "struct_macros.h"
#undef STRUCT_SIZE
