/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */


/* #define "string.h" */
/*
  nadefinovane makro STRUCT_NAME
 */

/* poskytuje funkce */

#define INIT_MACROS /* definuje nutna makra */
#include "struct_macros.h"
static int NUMBER_ITEMS;
static int NUMBER_COLLISIONS;

static xalloc_context ALLOC_CONTEXT ;
// automaticke alokace budou standardni malloc, alokace shodnych nodu (linearniho retizku) 

#define STD_ALLOC
#define STRUCT_SIZE (sizeof(uint32_t)*7)

void create_folder(hash_table* table, unsigned int bits);


/* ///////////////////////////////////////////////////////// */

typedef struct {
  string_t* ident;
  string_t* file_1;
  uint32_t  line_1;
  uint32_t   pos_1;
  string_t* file_2;
  uint32_t  line_2;
  uint32_t   pos_2;

  void * next_same;

  int no_add;
} STRUCT ;

int  GET (const char* ident);
int  ADD (const char* file
	  , const char* ident
	  , uint32_t line
	  , uint32_t pos
	  , const char* file_d
	  , uint32_t line_d
	  , uint32_t pos_d);
static void SAVE ();
static void INIT ();
static void DESTROY ();



static int MYCMP (string_t* k1, string_t* k2 ) ;
static void MYINITKEY (STRUCT * s, string_t* i) ;
static void INIT_DATA (STRUCT * s) { s->next_same=0; s->file_1=0; }
static void* ALLOC(unsigned int size) ;
static void FREE(void*);

#define TREE_NODE STRUCT
#define TREE_PREFIX(x) TREE_PREF(x) 

#define TREE_KEY_COMPLEX(x) x ident
#define TREE_KEY_DECL string_t* ident
#define TREE_GIVE_CMP MYCMP
#define TREE_GIVE_INIT_KEY MYINITKEY
#define TREE_GIVE_INIT_DATA

#define TREE_WANT_CLEANUP
#define TREE_WANT_LOOKUP
#define TREE_WANT_FIND

#define TREE_WANT_ITERATOR
#define TREE_STATIC
#define TREE_CONSERVE_SPACE

#define TREE_GIVE_ALLOC
#include "rb-tree.h"

struct TREE_PREF(tree) TREE_NAME ;

static int MYCMP (string_t* k1, string_t* k2 ) 
{
  int i;
  const char *str1, *str2;
  if (k1->typ==DB) str1=read_string(&file,k1->adr);
  else str1=k1->str;

  if (k2->typ==DB) str2=read_string(&file,k2->adr);
  else str2=k2->str;

  if ((i=strcmp(str1 , str2))) return i;
  return 0;
}

/* vice parametru + spravne inicializovat */
int  ADD (const char* file
	  , const char* ident
	  , uint32_t line
	  , uint32_t pos
	  , const char* file_d
	  , uint32_t line_d
	  , uint32_t pos_d)
{

  string_t* id_s = make_string(ident);
  string_t* fl_s = make_string(file);
  string_t* fld_s = make_string(file_d);


  STRUCT *item = TREE_PREF(lookup) (& TREE_NAME , id_s);
  if (!item->file_1) /* lookup tento node vytvoril */
    {
      item->file_1=fl_s;  item->ident=id_s; item->line_1=line; item->pos_1=pos; 
      item->file_2=fld_s; item->ident=id_s; item->line_2=line_d; item->pos_2=pos_d; 
      item->next_same=0;
      NUMBER_ITEMS++;
      return 0;
    }
  else /* je nutne vytvorit dalsi a navazat v seznamu */
    {
#ifdef MY_WARNINGS
#warning linearni seznam a musim ho cely projit abych vedel zda neni zaznam vlozen -> pomale!!!
#endif
      STRUCT *prev=0, *next=item;
      while (next)
	{
	  if (next->file_1    == fl_s 
	      && next->file_2 == fld_s 
	      /* dulezita je poloha jako bylo rozpoznano pomoci sparse je neduezite */
	      && ( next->line_1 & REAL_LINE     ) == ( line & REAL_LINE      )
	      && ( next->pos_1  & REAL_POSITION ) == ( pos  & REAL_POSITION  )
	      && ( next->line_2 & REAL_LINE     ) == ( line_d & REAL_LINE    )
	      && ( next->pos_2  & REAL_POSITION ) == ( pos_d & REAL_POSITION )
	      )
	    return 1;
	  else
	    {
	      prev=next;
	      next=(STRUCT*) next->next_same;
	    }
	}

      { /* pridani dalsiho prvku do spojoveho seznamu */
	NUMBER_COLLISIONS++;
	xalloc_context* context = xalloc_get_context();
	xalloc_set_context(& ALLOC_CONTEXT);
	STRUCT *item = xalloc(sizeof( STRUCT ));
	item->file_1=fl_s;  item->ident=id_s; item->line_1=line; item->pos_1=pos; 
	item->file_2=fld_s; item->ident=id_s; item->line_2=line_d; item->pos_2=pos_d; 
	item->next_same=0;
	prev->next_same=item;
	xalloc_set_context(context);
      }
      return 0;
    }
}

int GET (const char* ident)
{
  int ret=0;
  string_t* id_s = seek_string(ident);
  if (!id_s) return ret;
  STRUCT *item = TREE_PREF(find) (& TREE_NAME , id_s);
  STRUCT *next=item;
  while (next)
    {
      ret=1;
      WRITE7( next->ident->str
	     , next->file_1->str
	     , next->line_1
	     , next->pos_1
	     , next->file_2->str
	     , next->line_2
	     , next->pos_2);

      next=(STRUCT*) next->next_same;
    }  
  return ret;
}

static void MYINITKEY (STRUCT * s, string_t* i) 
{
  s->ident=i;
}

static void * NEXT_CHAIN_ITEM (void *item)
{
  STRUCT *tmp = (STRUCT *) item;
  return tmp->next_same;
}
static int WRITE_CHAIN_ITEM (file_t* file, uint32_t kam, void* co)
{
  if (!file || !co) {printf("addstruct WRITE_CHAIN_ITEM abort();");ABORT}
  STRUCT *tmp = (STRUCT *) co;
  uint32_t fa1   = (tmp->file_1->adr);
  uint32_t line1 = (tmp->line_1);
  uint32_t pos1  = (tmp->pos_1);
  uint32_t fa2   = (tmp->file_2->adr);
  uint32_t line2 = (tmp->line_2);
  uint32_t pos2  = (tmp->pos_2);
  write_int(file,kam+0*sizeof(uint32_t),fa1  );
  write_int(file,kam+1*sizeof(uint32_t),line1);
  write_int(file,kam+2*sizeof(uint32_t),pos1 );
  write_int(file,kam+3*sizeof(uint32_t),fa2  );
  write_int(file,kam+4*sizeof(uint32_t),line2);
  write_int(file,kam+5*sizeof(uint32_t),pos2 );
  return 0;
}
static uint32_t GET_IDENTIFIER (void* item)
{
  STRUCT *tmp = (STRUCT *) item;
  return (tmp->ident->adr);
}
static unsigned int ITEM_LENGTH = sizeof(uint32_t) * 6;

static void SAVE () 
{
#ifdef CREATE_FOLDER
  unsigned int records_in_page = (PAGE_SIZE-skip_page_head(0))/STRUCT_SIZE;
  unsigned int required_folder_size = 
    ( NUMBER_ITEMS
#ifdef INCLUDE_COLLISIONS_IN_COUNTING_FOLDER_SIZE
      + NUMBER_COLLISIONS
#endif
      )
    /records_in_page + 1;
#ifdef PREALLOC_PAGES
  page_prealloc(&file,required_folder_size);
#endif
  unsigned int required_folder_bits = 0;
  unsigned int c =1;
  for (required_folder_bits = 0, c=1
	 ; c < required_folder_size
	 ; required_folder_bits++, c*=2 );
  create_folder(&STRUCT_NAME,required_folder_bits);
#endif

  STRUCT* tmp;
  TREE_FOR_ALL( STRUCT_NAME , & TREE_NAME , add)
    {
      tmp=add;
#ifdef CHAIN_BOUND
      int count=0;
      for (;tmp ;count++) tmp=tmp->next_same;
      if ((double) count <= (double) (PAGE_SIZE/STRUCT_SIZE) * CHAIN_BOUND )
#else 
      if (1)
#endif
	{   
	  tmp=add;
	  while(tmp)
	    {
	      uint32_t ia    = (tmp->ident->adr);
	      uint32_t fa1   = (tmp->file_1->adr);
	      uint32_t line1 = (tmp->line_1);
	      uint32_t pos1  = (tmp->pos_1);
	      uint32_t fa2   = (tmp->file_2->adr);
	      uint32_t line2 = (tmp->line_2);
	      uint32_t pos2  = (tmp->pos_2);
	      /* tady ma byt pridani do file_t */
	      if (!tmp->no_add)
		{
		  SAVE_DATABASE (ia,fa1,line1,pos1,fa2,line2,pos2);
		  add_file_ident(tmp->file_1,tmp->ident);
		}
	      tmp=tmp->next_same;
	    }
	}
      else
	{
	  //	  printf("saving chain for %X\n",add->ident->adr);
	  SAVE_CHAIN_DATABASE(add,NEXT_CHAIN_ITEM,WRITE_CHAIN_ITEM, GET_IDENTIFIER, ITEM_LENGTH);
	}
    }
  TREE_END_FOR;
}

static void INIT ()
{
  ALLOC_CONTEXT .name=  "addstruct7" ;
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  TREE_PREF(init) ( & TREE_NAME  );
  xalloc_set_context(context);
}

static void DESTROY ()
{
  TREE_PREF(cleanup) ( & TREE_NAME  );
  memset(& TREE_NAME,0,sizeof(TREE_NAME));
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT);
  xfree_all();
  xalloc_set_context(context);
}

static void* ALLOC(unsigned int size) {
#ifdef STD_ALLOC
  void *s = malloc(size);
  memset(s,0,size);
  return s;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  void* r = xalloc(size);
  xalloc_set_context(context);
  return r;
#endif
}

static void FREE(void* s)
{
#ifdef STD_ALLOC
  free(s);
  return;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& ALLOC_CONTEXT );
  xfree(s);
  xalloc_set_context(context);
#endif
}

static void CHECK () 
{
  STRUCT *tmp;
  TREE_FOR_ALL( STRUCT_NAME , & TREE_NAME , add)
    {
      tmp=add;
      {   
	tmp=add;
	while (tmp)
	  {
	    uint32_t ia    = (tmp->ident->adr);
	    uint32_t fa1   = (tmp->file_1->adr);
	    uint32_t line1 = (tmp->line_1);
	    uint32_t pos1  = (tmp->pos_1);
	    uint32_t fa2   = (tmp->file_2->adr);
	    uint32_t line2 = (tmp->line_2);
	    uint32_t pos2  = (tmp->pos_2);
	    
	    if (HAS_DATABASE (ia,fa1,line1,pos1,fa2,line2,pos2))
	      tmp->no_add=1;
	    tmp=tmp->next_same;
	  }
      }
    }
  TREE_END_FOR;
}

#undef INIT_MACROS /* undefinuje potrebna makra */
#include "struct_macros.h"
#undef STRUCT_SIZE
