/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */


static xalloc_context file_context;
typedef struct {
  string_t* file;
  string_t* ident;
  void*     next_same;
  int       no_add;
} sfile_t;

#define STRUCT_SIZE (sizeof(uint32_t)*2)
static int NUMBER_ITEMS;
static int NUMBER_COLLISIONS;

static void add_file_ident(string_t* file_s, string_t* ident_s);


static int   sfile_cmp       (string_t* k1, string_t* k2 ) ;
static void  sfile_init_key  (sfile_t * s, string_t* i) ;
static void  sfile_init_data (sfile_t * s) { s->next_same=0; s->file=0; }
static void* sfile_alloc     (unsigned int size) ;
static void  sfile_free      (void*);

#define TREE_NODE sfile_t
#define TREE_PREFIX(x) sfile_ ## x

#define TREE_KEY_COMPLEX(x) x ident
#define TREE_KEY_DECL string_t* ident
#define TREE_GIVE_CMP sfile_cmp
#define TREE_GIVE_INIT_KEY sfile_init_key
#define TREE_GIVE_INIT_DATA

#define TREE_WANT_CLEANUP
#define TREE_WANT_LOOKUP
#define TREE_WANT_FIND

#define TREE_WANT_ITERATOR
#define TREE_STATIC
#define TREE_CONSERVE_SPACE

#define TREE_GIVE_ALLOC 

#include "rb-tree.h"

static struct sfile_tree sf_tree;
static void sfile_init_key (sfile_t * s, string_t* i) 
{
  s->ident=i;
  s->next_same=0;
}
static int sfile_cmp (string_t* k1, string_t* k2 ) 
{
  int i;
  const char *str1, *str2;
  if (k1->typ==DB) str1=read_string(&file,k1->adr);
  else str1=k1->str;

  if (k2->typ==DB) str2=read_string(&file,k2->adr);
  else str2=k2->str;

  if ((i=strcmp(str1 , str2))) return i;
  return 0;
}


static void* sfile_alloc(unsigned int size) {
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& file_context );
  void* r = xalloc(size);
  xalloc_set_context(context);
  return r;
}

static void sfile_free(void* s)
{
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(& file_context );
  xfree(s);
  xalloc_set_context(context);
}


static void init_file()
{ 
  sfile_init(&sf_tree); 
}

static void file_destroy() 
{ 
  sfile_cleanup(&sf_tree); 
  memset(& sf_tree,0,sizeof(sf_tree));
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&file_context);
  xfree_all();
  xalloc_set_context(context);
}



static void add_file_ident(string_t* file_s, string_t* ident_s)
{
  string_t* id_s = ident_s;
  string_t* fl_s = file_s;

  sfile_t *item = sfile_lookup (& sf_tree , id_s);
  if (!item->file) /* lookup tento node vytvoril */
    {
      item->file=fl_s; item->next_same=0;
      NUMBER_ITEMS++;
      return ;
    }
  else /* je nutne vytvorit dalsi a navazat v seznamu */
    {
      sfile_t *prev=0, *next=item;
      while (next)
	{
	  if (next->file==fl_s )
	    return ;
	  else
	    {
	      prev=next;
	      next=(sfile_t*) next->next_same;
	    }
	}
	{
	  NUMBER_COLLISIONS++;
      
	  xalloc_context* context = xalloc_get_context();
	  xalloc_set_context(& file_context );
	  sfile_t* it = xalloc(sizeof( sfile_t ));
	  it->file=fl_s; it->ident=id_s; it->next_same=0;
	  prev->next_same=it;
	  xalloc_set_context(context);
	}
      return ;
    }
}

static void save_sfile_t()
{

#ifdef CREATE_FOLDER
  unsigned int records_in_page = (PAGE_SIZE-skip_page_head(0))/STRUCT_SIZE;
  unsigned int required_folder_size = 
    ( NUMBER_ITEMS
#ifdef INCLUDE_COLLISIONS_IN_COUNTING_FOLDER_SIZE
      + NUMBER_COLLISIONS
#endif
      )
    /records_in_page + 1;
#ifdef PREALLOC_PAGES
  page_prealloc(&file,required_folder_size);
#endif
  unsigned int required_folder_bits = 0;
  unsigned int c =1;
  for (required_folder_bits = 0, c=1
	 ; c < required_folder_size
	 ; required_folder_bits++, c*=2 );
  create_folder(&files,required_folder_bits);
#endif

      sfile_t *tmp;
      TREE_FOR_ALL( sfile , & sf_tree , add)
	{
	  tmp=add;
	  while (tmp)
	    {
	      if (!tmp->no_add)
		db_add_file_ident(tmp->file->adr,tmp->ident->adr);
	      tmp=tmp->next_same;
	    }
	}
      TREE_END_FOR;
}

static void check_sfile_t()
{
      sfile_t *tmp;
      TREE_FOR_ALL( sfile , & sf_tree , add)
	{
	  tmp=add;
	  while (tmp)
	    {
	      if (db_has_file_ident(tmp->file->adr,tmp->ident->adr))
		tmp->no_add=1;
	      
	      tmp=tmp->next_same;
	    }
	}
      TREE_END_FOR;
}
#undef STRUCT_SIZE
