/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */
#include <stdio.h>
#include <string.h>
#ifndef STRING_H
#define STRING_H
#include "database/templates/hash_algo.h"

#include "database/fagin.h"
#define CREATE_FOLDER
#define PREALLOC_PAGES

#define STRING_ALLOC
#define STD_ALLOC

#ifdef STRING_ALLOC
#ifndef STD_ALLOC
#include "essential/xalloc.h"
static xalloc_context string_context   ;
#endif
#endif

int remember_strings;
int optimize_string;

#define STRUCT_SIZE (sizeof(uint32_t)*2)

static int STRINGS_LENGTH;
static int NUMBER_ITEMS;
static int NUMBER_COLLISIONS;


enum string_t_typ {
  MEM,
  DB,
};

typedef struct {
  enum string_t_typ typ;
  uint32_t   adr;
  char       str[1];
} string_t;

static void  string_init_data(string_t* s);
#ifdef STRING_ALLOC
static void* string_alloc (unsigned int size);
static void  string_free  (void* s);
#define TREE_GIVE_ALLOC 
#endif

#define TREE_NODE string_t
#define TREE_PREFIX(x) string_##x
#define TREE_KEY_ENDSTRING str
#define TREE_WANT_CLEANUP
#define TREE_WANT_LOOKUP
#define TREE_WANT_FIND
#define TREE_GIVE_INIT_DATA

#define TREE_WANT_ITERATOR
#define TREE_STATIC
#define TREE_CONSERVE_SPACE

#include "rb-tree.h"

struct string_tree tree;

typedef struct {
  string_t*  s;
  int        str_hash;
} optstring_t;

#ifdef OPTIMIZE
#define TREE_NODE optstring_t
#define TREE_PREFIX(x) optstring_##x

#define TREE_DUPLICATES
#define TREE_KEY_ATOMIC str_hash
#define TREE_WANT_CLEANUP
#define TREE_WANT_NEW

#define TREE_WANT_ITERATOR
#define TREE_STATIC
#define TREE_CONSERVE_SPACE
#include "rb-tree.h"

struct optstring_tree opt_tree;
#endif

///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

static string_t* make_string(const char* string)
{
  string_t *item;
  item=string_lookup(&tree,(char*)string);
  return item;
}

static string_t* seek_string(const char* string)
{
  return string_find(&tree,(char*)string);
}

/* static void check_strings() */
/* { */
/*   remember_strings=0; */
/*       TREE_FOR_ALL(string, &tree, act) */
/* 	{ */
/*             uint32_t adr; */
/* 	    int ai=test_already_inserted; */
/* 	    test_already_inserted=0; */
/* 	    if (remember_strings)      adr=db_add_string(/\* &identifier, *\/act->str); */
/* 	    else                       adr=db_get_string(/\* &identifier, *\/act->str); */
/* 	    test_already_inserted=ai; */
/* 	    act->adr=adr; */
/* 	    act->typ=DB; */
/* 	} */
/*       TREE_END_FOR; */
/* } */

static void save_strings()
{
#ifdef CREATE_FOLDER
  unsigned int records_in_page = (PAGE_SIZE-skip_page_head(0))
    /STRUCT_SIZE;
  unsigned int required_folder_size = 
    ( NUMBER_ITEMS
#if 1
      + NUMBER_COLLISIONS
#endif
      )
    /records_in_page + 1;
#ifdef PREALLOC_PAGES 
  page_prealloc(&file,required_folder_size);
#endif
#ifdef PREALLOC_PAGES
  page_prealloc(&file,STRINGS_LENGTH/PAGE_SIZE+1);
#endif
  
  unsigned int required_folder_bits = 0;
  unsigned int c =1;
  for (required_folder_bits = 0, c=1
	 ; c < required_folder_size
	 ; required_folder_bits++, c*=2 );
  create_folder(&identifier,required_folder_bits);
#endif


#ifdef OPTIMIZE
  if (!optimize_string)
    {
#endif
      TREE_FOR_ALL(string, &tree, act)
	{
            uint32_t adr;
	    int ai=test_already_inserted;
	    test_already_inserted=0;
	    if (remember_strings)      adr=db_add_string(/* &identifier, */act->str);
	    else                       adr=db_get_string(/* &identifier, */act->str);
	    test_already_inserted=ai;
	    act->adr=adr;
	    act->typ=DB;
	}
      TREE_END_FOR;
#ifdef OPTIMIZE
    }
  else
    {
      TREE_FOR_ALL(string, &tree, act)
	{
	  optstring_t* o = optstring_new(&opt_tree,stringhash((unsigned char*)act->str));
	  o->s=act;
	}
      TREE_END_FOR
	;
      TREE_FOR_ALL(optstring, &opt_tree, act)
	{
            uint32_t adr;
	    int ai=test_already_inserted;
	    test_already_inserted=0;
	    if (remember_strings)
	      adr=db_add_string(/* &identifier, */act->s->str);
	    else
	      adr=db_get_string(/* &identifier, */act->s->str);
	    test_already_inserted=ai;
	    act->s->adr=adr;
	    act->s->typ=DB;
	}
      TREE_END_FOR;
    }
#endif
}

void list_strings()
{
  TREE_FOR_ALL(string, &tree, act)
    {
      printf("%s\n",act->str);
    }
  TREE_END_FOR;
}

static void  string_init_data(string_t* s)    
{
  s->typ=MEM; 
  s->adr=0;

  int len = strlen(s->str);
  //  STRINGS_LENGTH+=len; // append string 

  int length = (len/sizeof(uint32_t) + 1)*sizeof(uint32_t); // padd string with zero
  int prev = STRINGS_LENGTH/PAGE_SIZE;
  int act = (STRINGS_LENGTH+length)/PAGE_SIZE;
  if (act!=prev)    STRINGS_LENGTH=act*PAGE_SIZE; // string doesn't fit to page pad with zero...
  STRINGS_LENGTH+=length; // append string 

  NUMBER_ITEMS++;

} 

static void string_destroy()
{
  string_cleanup(&tree);
  memset(&tree,0,sizeof(tree));
}

#ifdef STRING_ALLOC
static void* string_alloc(unsigned int size)
{
#ifdef STD_ALLOC
  void *s = malloc(size);
  memset(s,0,size);
  return s;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&string_context);
  void* r = xalloc(size);
  xalloc_set_context(context);
  return r;
#endif
}
static void string_free(void* s)
{
#ifdef STD_ALLOC
  free(s);
  return;
#else
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&string_context);
  xfree(s);
  xalloc_set_context(context);
#endif
}
#endif


char*    mem_generator_next_string(int reset)
{
  static int running;
  static int end;
  
  if (reset) { 
    running=0;
    end=0;
    return 0;
  } else {
    if (end) return 0;
    if (running) goto cycle_next;
    running=1;
  }

  static string_t* old;
  TREE_FOR_ALL(string, &tree, act)
    {
      old=act;
      return act->str;
      
    cycle_next: 
      act=old;
    }
  TREE_END_FOR;
  end=1;
  return 0;
}




///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

static void init_string()
{
  STRINGS_LENGTH=0;
  NUMBER_ITEMS=0;
  NUMBER_COLLISIONS=0;
  string_init(&tree);
}
#endif
#undef STRUCT_SIZE
