/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */


#include <sys/mman.h>
#include "database/file.h"
#include "hash_algo.h"
#include "base/macros.h"

STRUCT_DEF


#define   REMOVE_FROM_FOLDER                                             \
	  int x=get_hash_bites_folder(  table->file,                     \
					get_folder_address(table));      \
	  int y=get_page_bites( table->file,                             \
				adr );                                   \
	  int length=1<<(x-y);                                           \
	  int start_index = index-(index%length);                        \
	  int stop_index = start_index+length;                           \
	  int i;                                                         \
	  for (i=start_index;i<stop_index;i++)                           \
	    if ( get_address_at_folder_index(table->file                 \
					     ,get_folder_address(table)  \
					     ,i) == adr )                \
	      set_address_at_folder_index(table->file                    \
					  , get_folder_address(table)    \
					  , i                            \
					  , 0);                          \
	    /* else printf("neocekavane!!!\n"); */                       \
	  set_address_at_folder_index(table->file                        \
				      , get_folder_address(table)        \
				      , index                            \
				      , 0);


//#define VERBOSE_CHECK
extern unsigned int get_index(hash_table* table, uint32_t hash); /* fagin.c */

static uint32_t HASH_FUNCTION      (const void* key)
{
#ifdef STR_HASH_FCE
  return (uint32_t) stringhash((unsigned char*)key) ;
#else
#ifdef NUM_HASH_FCE
  return inthash(*(uint32_t*)key) ;
#endif
#endif
}
static const void*    GET_KEY_FUNCTION       (const void* item)
{
  if (!item) ABORT
#ifdef SHOW_GET_KEY
#ifdef STR_HASH_FCE
  printf("key='%s'\n",((ITEM_STRUCT *) item)->key);
#else
  printf("key='%d'\n",*((ITEM_STRUCT *) item)->key);
#endif
#endif
  return ((ITEM_STRUCT *) item)->key;
}

static uint32_t    GET_ITEMHASH_FUNCTION       (const void* item)
{
  if (!item) ABORT
  return ((ITEM_STRUCT *) item)->hash;
}

static const void*    GET_FILEADR_FUNCTION       (const void* item)
{
  if (!item) ABORT
  return &((ITEM_STRUCT *) item)->filename_adr;
}

static int  IS_ITEM_CHAIN_START  (const void* item)
{
#if 0
  if (((ITEM_STRUCT*) item)->is_chain) 
    {
      printf("je to chain item\n");
      __PRINT_ITEM(((ITEM_STRUCT *) item));
    }
#endif
  return ((ITEM_STRUCT*) item)->is_chain;
}
static uint32_t GET_ITEM_CHAIN_START (const void* item)
{
  return ((ITEM_STRUCT*) item)->chain_adr;
}
static void* CREATE_CHAIN_ITEM (uint32_t ident, uint32_t address)
{
  static ITEM_STRUCT item;
  memset(&item, sizeof(ITEM_STRUCT), 0);
  item.key_adr   = ident;
  item.chain_adr = address;
  item.is_chain  = 1;
  item.key = & item.key_adr; 
  if (&item.key_adr == &item.chain_adr)
  { printf("chyba:chain a key na stejne adrese!!!\n"); exit(1);}

  return &item;
}

static uint32_t GET_ITEM_HASH (file_t* file, uint32_t adr)
{
  ITEM_STRUCT a; ITEM_STRUCT* t=&a;
  READ(t);
  return HASH_FUNCTION(t->key);
}

static void    GET_ITEM_FUNCTION     (file_t* file, uint32_t adr, void* item)
{
  if (!file||!adr) ABORT

  ITEM_STRUCT *i = (ITEM_STRUCT*) item;
  READ(i);
}


static int    GET_LEN_FUNCTION       (const void* item)
{
  return LENGTH_ITEM;
}

static uint32_t GET_NEXT_ITEM     (file_t* file, uint32_t adr)
{
  if (!file || !adr) ABORT
  return adr+LENGTH_ITEM;
}

/* pokud se lisi vrati 1 pokud se nelisi vrati 0 */
static int      GET_COMPAR_ITEM    (const void* item1, const void* item2)
{
  if (!item1 || !item2) ABORT
  ITEM_STRUCT * i1 = ( ITEM_STRUCT *) item1;
  ITEM_STRUCT * i2 = ( ITEM_STRUCT *) item2;
  if ( COMPAR_ITEM(i1,i2) ) return 1;
  return 0;
}


static int      GET_COMPAR_KEY   (const void* key1
				  , const void* key2
				  , int hash1
				  , int hash2)
{
  if (! key1 || !key2) ABORT
  COMPAR_KEY (key1,key2,hash1,hash2);
}

	   /* return ==  0 pokud nebylo volne misto pro item */
	   /* return == -1 pokud byla item jiz vlozena do databaze */
	   /* return == adresa vlozeneho zaznamu pokud byl zapsan */
static int    CHECK_ADD_ITEM       (file_t* file, uint32_t adr, const void* item)
{
  if (!file || !adr || !item ) ABORT
  ITEM_STRUCT * i=( ITEM_STRUCT *) item;

  do {
    uint32_t act_adr=skip_page_head(adr);
    int free_space=PAGE_SIZE-(act_adr-adr);
    for ( ; free_space>=LENGTH_ITEM ; 
	  free_space-=LENGTH_ITEM, 
	    act_adr+=LENGTH_ITEM)
      {
	char tmp[MAX_ITEM_SIZE];
	ITEM_STRUCT* it=(ITEM_STRUCT*) tmp;
	GET_ITEM_FUNCTION(file,act_adr,it);
	if ( COMPAR_ITEM(i,it) ) return 1;   // nasel zaznam

      }
  } while ((adr=get_next_page_address(file,adr)));
  return 0;
}
#ifdef MY_WARNINGS
#warning nefunguje zatim stranka typu identifikator -> zaznamy
#endif
static uint32_t    FAST_ADD_ITEM       (file_t* file, uint32_t adr, const void* item)
{
/*   return CHECK_ADD_ITEM(file,adr,item); */

  if (adr<PAGE_SIZE) { fprintf(stderr,"chyba adresa=%X < 1024\n",adr); ABORT }
  uint32_t pos_for_item;
  pos_for_item = (uint32_t) skip_page_head(adr);
  if ( (*(uint32_t*) (file->realfile + pos_for_item)) )
    /* prvni je obsazen -> musim nacist volnou pozici */
    pos_for_item = get_free_pos_in_page(file,adr);

  if ( pos_for_item-adr + LENGTH_ITEM >= PAGE_SIZE )  /* stranka byla plna */  return 0;

  uint32_t records=pos_for_item;
  ITEM_STRUCT * i=( ITEM_STRUCT *) item;
  WRITE;

  set_free_pos_in_page(file,adr,records+LENGTH_ITEM);

  return records;
}



static int GET_HAS_ITEM (file_t* file, uint32_t adr, uint32_t start)
{
  if (!file|| ! adr || !start) ABORT
  if (adr-start>PAGE_SIZE-LENGTH_ITEM) return 0;
  if (adr-start<0) return 0;
  int has=read_int(file,adr)!=0;
  return has;
}

static int GET_ENOUGH_SPACE (file_t*   file,
		     uint32_t  adr,
		     const void*     item)
{
  uint32_t records=(uint32_t) skip_page_head(adr);
  int free_space=PAGE_SIZE-(records-adr);
  for (;free_space>=LENGTH_ITEM && read_int(file,records);records+=LENGTH_ITEM,free_space-=LENGTH_ITEM);
  if (free_space>=LENGTH_ITEM)
    return 1;
  else return 0;
}


static uint32_t GET_FIND_RETURN  (file_t* file, uint32_t adr )
{
  if (!file || !adr) ABORT
  KEY_RETURN
}


uint32_t last_collision_page(file_t* file, uint32_t adr);
uint32_t prev_collision_page(file_t* file, uint32_t adr, uint32_t drag);



#if 0
static int  __INSERTED  ( void* t
			  , const void* orig
			  , uint32_t old_adr)
     //#define  VERY_VERBOSE_CHECK
{
#ifdef __CHECK_INSERTED
  if (!old_adr) ABORT
  uint32_t i;
  hash_table* table =(hash_table*) t;

  i=find(table,table->get_key(orig),0);
  if ( !i ) 
    {
      fprintf(stderr,"\nfail ... not inserted 1\tsupposed_adr = %X\n",old_adr);
      __PRINT_ITEM(((ITEM_STRUCT *) orig));
      return 1;
    }
  if(!old_adr) goto ok;

  uint32_t* j;
#ifdef VERY_VERBOSE_CHECK
  fprintf(stderr,"\n\nnove overovani suggested_adr = %X\n",old_adr);
#endif
  for(j=i;*j;j++)
    {
      uint32_t adr=*j;
      file_t* file=table->file;

      ITEM_STRUCT item2;
      ITEM_STRUCT *i2=&item2;
      READ(i2);

      ITEM_STRUCT *i1 = (ITEM_STRUCT *) orig;
      if (! COMPAR_ITEM(i1,i2) )
	{
#ifdef __PRINT_ITEM
#ifdef VERY_VERBOSE_CHECK
	  fprintf(stderr,"%X\twant ",old_adr);
	  __PRINT_ITEM(i1);
	  
	  fprintf(stderr,"%X\thave ",adr);
	  __PRINT_ITEM(i2);
	  fprintf(stderr,"\n");
#endif
#endif
#ifdef VERY_VERBOSE_CHECK
	  fprintf(stderr,"certifikat... suggested_adr of item %X ~ %X new_adr of item \n",old_adr,adr);
#endif
	  goto ok;
	}
#ifdef VERY_VERBOSE_CHECK
      else
	{
	  fprintf(stderr,"lisi se %X\twant ",old_adr);
	  __PRINT_ITEM(i1);
	  
	  fprintf(stderr,"lisi se %X\thave ",adr);
	  __PRINT_ITEM(i2);
	  fprintf(stderr,"\n");
	}
#endif
      if (*j == old_adr)
	fprintf(stderr,"adresy se shoduji, ale porovnani selhalo\n");
     }
  fprintf(stderr,"\nfail ... not inserted 2\tsupposed_adr = %X\n", old_adr);
  __PRINT_ITEM(((ITEM_STRUCT *) orig));
   if (i) xfree(i);
   // exit(1);
   return 1;
 ok:
/* #ifdef VERY_VERBOSE_CHECK */
/*   fprintf(stderr,"tisknu vse\n"); */
/*   for(j=i;*j;j++) */
/*     { */
/*       uint32_t adr=*j; */
/*       file_t* file=table->file; */
/*       READ; */
/*       ITEM_STRUCT i2 = GVAR_NAME; */
/* #ifdef __PRINT_ITEM */
/* 	  fprintf(stderr,"%X\told ",old_adr); */
/* 	  ITEM_STRUCT * item=i1; */
/* 	  __PRINT_ITEM(i1); */
/* 	  item=&i2; */
/* 	  fprintf(stderr,"%X\tnew ",adr); */
/* 	  __PRINT_ITEM(item); */
/* 	  printf("\n"); */
/* #endif */
/*     } */
/*   fprintf(stderr,"hotovo\n\n"); */
/* #endif */
  if(i) xfree(i);
  ;
#endif
  return 0;
}
#endif /* __INSERTED */

static int    GET_ALREADY_INSERTED       (file_t* file, uint32_t adr, const void* item)
{
  if (!file || !adr || !item ) ABORT
  ITEM_STRUCT * i=( ITEM_STRUCT *) item;

  if (adr<PAGE_SIZE) 
    { fprintf(stderr,"chyba adresa=%X < PAGE_SIZE %X\n",adr,PAGE_SIZE); ABORT }
  do {
    uint32_t records=(uint32_t) skip_page_head(adr);
    int pos = records-adr;
    int free_space=PAGE_SIZE-pos;
    for (;free_space>=LENGTH_ITEM && read_int(file,records)
	   ;records+=LENGTH_ITEM,free_space-=LENGTH_ITEM)
      {
	if ( COMPAR_INSERTED ) 
	  {
	    //	    __PRINT_ITEM(i);
	    /* adresa stejneho zaznamu */
	    return records;  
	  }
      }
  } while ((adr=get_next_page_address(file,adr)));

  return 0;
}

static void __PROTECT(const void* item)
{
#ifdef SHOW_INSERT_TEST
  fprintf(stderr,"protect called  %X\t",HASH_FUNCTION(GET_KEY_FUNCTION(item)));
  __PRINT_ITEM(((ITEM_STRUCT*)item));
#endif
}

static void __PRINT(file_t* file, uint32_t adr)
{
  ITEM_STRUCT a;
  ITEM_STRUCT* b=&a;
  if (!file||!adr) ABORT

  READ(b);
  __PRINT_ITEM(b);
}

static const void* GET_CHAIN_KEY          ( file_t* file, uint32_t adr)
{
  return file->realfile+adr;
}
static int         HAS_CHAIN_ITEM         ( file_t* file, uint32_t adr, uint32_t start)
{
  if (!file|| ! adr || !start) ABORT
  if (adr-start>PAGE_SIZE-(LENGTH_ITEM-sizeof(uint32_t))) return 0;
  if (adr-start<0) return 0;
  int has=read_int(file,adr)!=0;
  return has;
}
static uint32_t    NEXT_CHAIN_ITEM        ( file_t* file, uint32_t adr)
{
  if (!file || !adr) ABORT
  return adr+LENGTH_ITEM-sizeof(uint32_t);
}
static void        GET_CHAIN_ITEM         ( file_t* file, uint32_t adr, void* item, const void *key)
{
  if (!file||!adr) ABORT
		     ;
/*   printf("read address %X\n",adr); */

  ITEM_STRUCT *i = (ITEM_STRUCT*) item;
  SREAD(i);
  i->key = (void*) key;

/*   __PRINT_ITEM(i); */

}


///////////////////////////////////////////////////
/////////////// remove functions //////////////////
///////////////////////////////////////////////////
static int MOVE_LAST_ITEM_TO_ADDRESS (file_t* file, uint32_t page_adr)
{
  uint32_t end_item       = get_free_pos_in_page(file, page_adr) ; /* kam zapsat ... */
  set_free_pos_in_page(file,page_adr,end_item+LENGTH_ITEM); /* posunu zaznamy o zapsanou item */

  //  printf("chain_bg=%X~%d end_item=%X~%d\n",page_adr,page_adr,end_item,end_item);
  uint32_t last_chain_adr = last_collision_page( file, page_adr);
  //  printf("last_chain_adr %X\n",last_chain_adr);
  uint32_t free_pos       = get_free_pos_in_page(file, last_chain_adr) - LENGTH_ITEM;
  //  printf("free_pos = %X\n",free_pos);

  memmove(file->realfile+end_item, file->realfile+free_pos, LENGTH_ITEM);
  memset (file->realfile+free_pos,0,LENGTH_ITEM);
  set_free_pos_in_page(file,last_chain_adr,free_pos);
  
  if (skip_page_head(last_chain_adr) == free_pos) /* stranka se vyprazdnila... */
    {
      if (page_adr == last_chain_adr) 
	return 1; /* delete entire page which is the end of chain*/


      uint32_t prev = prev_collision_page(file,page_adr,last_chain_adr);
      set_next_page_address (file, prev, 0);
      page_free(file,last_chain_adr);

/*       printf("deleteng last page - moving data, and no data left in page %X\n",last_chain_adr); */
/*       last_collision_page( file, page_adr); */
    }
  return 0;
}

static int __REMOVE_ (hash_table *table
		       , uint32_t act_adr /* adresa polozky ke smazani*/
		       , uint32_t adr /* adresa stranky */
		       , int hash /* hes libovolne polozy ve strance - abych mohl zjistit index v tabulce */
		       )
{
  file_t* file = table->file;

  { /* chci posunout datovy obsah stranky */
    uint32_t end_item       = get_free_pos_in_page(file, adr) - LENGTH_ITEM;
    if (end_item!=act_adr) 
      memmove(file->realfile   + act_adr
	      , file->realfile + end_item
	      , LENGTH_ITEM);
    memset(file->realfile+end_item,0,LENGTH_ITEM); /* posledni zaznam smazat - chci v obou pripadech */
    set_free_pos_in_page(file,adr,end_item); /* posunu zaznamy o item */
  }
  
#if 1
  if (MOVE_LAST_ITEM_TO_ADDRESS(file,adr))
    { /* je nutno dealokovat stranku adr*/
/*       printf("delete entire page which is the end of chain %X\n",adr); */
      unsigned int index   
	= get_index(table,hash);
      unsigned int pg_adr
	= get_address_at_folder_index(table->file
				      ,get_folder_address(table)
				      ,index);
      uint32_t first_chain_page = pg_adr;
      if (adr==first_chain_page)
	{
	  /* 
	     but i have to remove the address from all indexes in folder, 
	     not just one
	   */
	  REMOVE_FROM_FOLDER
	  page_free(file,adr);

/* 	  printf("removing from folder %X index %d (adr=%X)\n" */
/* 		 ,get_folder_address(table) */
/* 		 ,index,adr); */

	  if (get_address_at_folder_index(table->file
					  ,get_folder_address(table)
					  ,index))
	    printf("TO SNAD NE!!! - spatne smazano v adresari\n");
	}
      else
	{
	  uint32_t prev = 
		      prev_collision_page(file,adr,first_chain_page);
	  /* ukoncit retizek */
	  set_next_page_address (file, prev, 0);
	  page_free(file,adr);

	  printf("deleting page but i'm at end of retizek prev=%X folder=%X index=%d\n"
		 , prev, get_folder_address(table), index);
	  last_collision_page( file, first_chain_page);
	}
      return 1; /* tuto stranku mazu a je posledni */
    }
  return 0;
#endif
}

static uint32_t FILE_REMOVE  (hash_table* table, uint32_t adr, const void* fi_key )
{
/*   printf("file remove %X\n",adr); */
  file_t* file = table->file;
  if (adr<PAGE_SIZE) ABORT /* zda spravne zavolano! */
  do {
    //    int o=0; for (;o<1;o++) {
    uint32_t act_adr=skip_page_head(adr);
    int free_space=PAGE_SIZE-(act_adr-adr);
    for ( ; free_space>=LENGTH_ITEM ; 
	  free_space-=LENGTH_ITEM, 
	    act_adr+=LENGTH_ITEM)
      {
	char tmp[MAX_ITEM_SIZE];
	ITEM_STRUCT* it=(ITEM_STRUCT*) tmp;
	GET_ITEM_FUNCTION(file,act_adr,it);
	
	/* muze se stat ze budu potrebovat vedet jaky index mam v hasovaci tabulce */
	
	int delete_item=0;
#ifndef DELETE_ALL_IN_UPDATE
	{ /* zjisti, zda ma byt tato polozka smazana */
	  file_item** fi = (file_item**) fi_key;
	  for ( fi=(file_item**) fi_key ;*fi ;fi++) 
	    /* mezi polozkami fi_key existuje nejaka 
	       ktera pozaduje smazani aktivni */
	    {
	      uint32_t filename_adr   = get_file_item_filename  ((*fi));
	      uint32_t identifier_adr = get_file_item_identifier((*fi));
	      if  ( filename_adr == *(uint32_t*) GET_FILEADR_FUNCTION(it))
		if  ( identifier_adr == *(uint32_t*) GET_KEY_FUNCTION(it))
		  {
		    delete_item=1;
		    break;
		  }
	    }
	}
#else 
	delete_item=1;
#endif
	//	int hash = HASH_FUNCTION(GET_KEY_FUNCTION(it));
	if(delete_item)
	  {
	    int hash = HASH_FUNCTION(GET_KEY_FUNCTION(it));
	    if (__REMOVE_(table, act_adr, adr, hash))
	      return 0; /* uz neni dalsi zaznam */
	    act_adr-=LENGTH_ITEM;
	    free_space+=LENGTH_ITEM;
	  }
/* 	if (0) */
/* 	last_collision_page( file,  */
/* 			     get_address_at_folder_index(table->file */
/* 							 , get_folder_address(table) */
/* 							 , get_index(table,hash)) */
/* 			     ); */
      }
    //    }
  } while ((adr=get_next_page_address(file,adr)));
  return 0;
}

static uint32_t REMOVE_ITEM  (hash_table* table, uint32_t adr, const void* key )
{
  file_t* file = table->file;
  if (adr<PAGE_SIZE) ABORT;
  if (adr>file->sb.st_size) ABORT;
  do {
    uint32_t act_adr=skip_page_head(adr);
    int free_space=PAGE_SIZE-(act_adr-adr);
    for ( ; free_space>=LENGTH_ITEM ; 
	  free_space-=LENGTH_ITEM, 
	    act_adr+=LENGTH_ITEM)
      {
	char tmp[MAX_ITEM_SIZE];
	ITEM_STRUCT* it=(ITEM_STRUCT*) tmp;
	GET_ITEM_FUNCTION(file,act_adr,it);

	if  ( ! GET_COMPAR_KEY(key
			       , GET_KEY_FUNCTION(it)
			       , 0
			       , 0 ) )
	  {
	    int hash = HASH_FUNCTION (GET_KEY_FUNCTION(it));
	    if (__REMOVE_(table, act_adr, adr, hash)) return 0; /* uz neni dalsi zaznam */
	    act_adr-=LENGTH_ITEM;
	    free_space+=LENGTH_ITEM;
	  }
      }
  } while ((adr=get_next_page_address(file,adr)));
  return 0;
}
///////////////////////////////////////////////////
/////////// end remove functions //////////////////
///////////////////////////////////////////////////



///////////////////////////////////////////////////
//////////chain remove functions //////////////////
///////////////////////////////////////////////////
#define LENGTH_ITEM_DATA (LENGTH_ITEM - sizeof(uint32_t))
static int CHAIN_MOVE_LAST_ITEM_TO_ADDRESS (file_t* file, uint32_t page_adr)
{
  uint32_t end_item       = get_free_pos_in_page(file, page_adr) ; /* kam zapsat ... */
  set_free_pos_in_page(file,page_adr,end_item+LENGTH_ITEM_DATA); /* posunu zaznamy o zapsanou item */

  //  printf("chain_bg=%X~%d end_item=%X~%d\n",page_adr,page_adr,end_item,end_item);
  uint32_t last_chain_adr = last_collision_page( file, page_adr);
  //  printf("last_chain_adr %X\n",last_chain_adr);
  uint32_t free_pos       = get_free_pos_in_page(file, last_chain_adr) - LENGTH_ITEM_DATA;
  //  printf("free_pos = %X\n",free_pos);

  memmove(file->realfile+end_item, file->realfile+free_pos, LENGTH_ITEM_DATA);
  memset (file->realfile+free_pos,0,LENGTH_ITEM_DATA);
  set_free_pos_in_page(file,last_chain_adr,free_pos);
  
  if (skip_page_head(last_chain_adr) == free_pos) /* stranka se vyprazdnila... */
    {
      if (page_adr == last_chain_adr) 
	return 1; /* delete entire page */

      uint32_t prev = prev_collision_page(file,page_adr,last_chain_adr);
      set_next_page_address (file, prev, 0);
      page_free(file,last_chain_adr);

    }
  return 0;
}

static int __CHAIN_REMOVE_ (hash_table *table
		       , uint32_t act_adr /* adresa polozky ke smazani*/
		       , uint32_t adr /* adresa stranky */
		       , int hash /* hes libovolne polozy ve strance - abych mohl zjistit index v tabulce */
		       )
{
  //  fprintf(stderr,"chain remove nebyla implekmentovana"); exit(1);
  file_t* file = table->file;

  { /* chci posunout datovy obsah stranky */
    uint32_t end_item       = get_free_pos_in_page(file, adr) - LENGTH_ITEM_DATA;
    if (end_item!=act_adr) 
      memmove(file->realfile   + act_adr
	      , file->realfile + end_item
	      , LENGTH_ITEM_DATA);
    memset(file->realfile+end_item,0,LENGTH_ITEM_DATA); /* posledni zaznam smazat - chci v obou pripadech */
    set_free_pos_in_page(file,adr,end_item); /* posunu zaznamy o item */
  }
  
#if 1
  if (CHAIN_MOVE_LAST_ITEM_TO_ADDRESS(file,adr))
    { /* je nutno dealokovat stranku adr*/
      unsigned int index   
	= get_index(table,hash);
      unsigned int pg_adr
	= get_address_at_folder_index(table->file
				      ,get_folder_address(table)
				      ,index);
      uint32_t first_chain_page = pg_adr;
      if (adr==first_chain_page)
	{
	  REMOVE_FROM_FOLDER
	  page_free(file,adr);

/* 	  printf("freeing page (1) %X\n",adr); */
	}
      else
	{
	  uint32_t prev = prev_collision_page(file,adr,first_chain_page);
	  set_next_page_address (file, prev, 0);
	  	  page_free(file,adr);

/* 	  printf("freeing page (2) %X\n",adr); */
	}
      return 1; /* tuto stranku mazu a je posledni */
    }
#endif
  return 0;
}

static uint32_t FILE_CHAIN_REMOVE  (hash_table* table, uint32_t chain_adr, const void* fi_key )
{
  //  fprintf(stderr,"FILE chain remove nebyla implekmentovana"); exit(1);

  file_t* file = table->file;
  if (chain_adr<PAGE_SIZE) ABORT /* zda spravne zavolano! */
  uint32_t key_adr = skip_page_head(chain_adr);
  uint32_t *key=(uint32_t* ) (table->file->realfile+key_adr);
  int first = 1;


  if (chain_adr)
  do {
    //    printf("chain %X\n",chain_adr);
    uint32_t adr = chain_adr;
    do {
      uint32_t act_adr=skip_page_head(adr);
      if (first) { first = 0; act_adr+=sizeof(uint32_t); } // preskoci identifikator...
      int free_space=PAGE_SIZE-(act_adr-adr);
      for ( ; free_space>=LENGTH_ITEM_DATA ; 
	    free_space-=LENGTH_ITEM_DATA, 
	    act_adr+=LENGTH_ITEM_DATA)
	{
	  char tmp[MAX_ITEM_SIZE];
	  ITEM_STRUCT* it=(ITEM_STRUCT*) tmp;
	  //	GET_ITEM_FUNCTION(file,act_adr,it);
	  //	  printf("reading at %X\n",act_adr);
	  GET_CHAIN_ITEM(file,act_adr,it,key);
	  
	  /* muze se stat ze budu potrebovat vedet jaky index mam v hasovaci tabulce */
	  
	  int delete_item=0;
	  file_item** fi = (file_item**) fi_key;
	  for (fi=(file_item**) fi_key;*fi;fi++) /* mezi polozkami fi_key existuje nejaka ktera pozaduje smazani aktivni */
	    {
	      uint32_t filename_adr   = get_file_item_filename  ((*fi));
	      uint32_t identifier_adr = get_file_item_identifier((*fi));
	      if  ( identifier_adr == key_adr)
		//	    if  ( identifier_adr == *(uint32_t*) GET_KEY_FUNCTION(it))
		if  ( filename_adr == *(uint32_t*) GET_FILEADR_FUNCTION(it))
		{
		  delete_item=1;
		  break;
		}
	    }
	
	  if(delete_item)
	    {
	      int hash = HASH_FUNCTION(GET_KEY_FUNCTION(it));
	      if (__CHAIN_REMOVE_(table, act_adr, adr, hash)) return 0; /* uz neni dalsi zaznam */
	      act_adr-=LENGTH_ITEM_DATA;
	    }
	}
    } while ((adr=get_next_page_address(file,adr)));
  } while ((chain_adr = get_chain_page_address(file, chain_adr )));
  return 0;
}

static uint32_t CHAIN_REMOVE_ITEM  (hash_table* table, uint32_t adr, const void* key )
{
  fprintf(stderr,"chain remove item nebyla implekmentovana"); exit(1);
  file_t* file = table->file;
  if (adr<PAGE_SIZE) ABORT
  uint32_t prev = 0;
  do {
    uint32_t act_adr=skip_page_head(adr);
    uint32_t *key_act=(uint32_t*) (table->file->realfile+act_adr);
    if  ( ! GET_COMPAR_KEY(key
			   , key_act
			   , 0
			   , 0 ) )
      {
	uint32_t my_adr = adr;
	if  (prev)
	  set_chain_page_address(file,adr,get_chain_page_address(file,adr));
	while (my_adr)
	  {
	    uint32_t del = my_adr;
	    my_adr=get_next_page_address(file,my_adr);
	    page_free(file,del);
	  }
	break;
      }
    prev=adr;
  } while ((adr=get_chain_page_address(file,adr)));
  return 0;
}
#undef LENGTH_ITEM_DATA
///////////////////////////////////////////////////
///////end chain remove functions /////////////////
///////////////////////////////////////////////////




void SET_TABLE (hash_table* t)
{
  t->allow_chain  = 1;
  t->hash_fce     =  HASH_FUNCTION;
  t->get_key      =  GET_KEY_FUNCTION;
  t->get_item     =  GET_ITEM_FUNCTION;
  t->get_len      =  GET_LEN_FUNCTION;
  t->enough_space =  GET_ENOUGH_SPACE ;
  t->has_item     =  GET_HAS_ITEM;

#ifdef MY_WARNINGS
#warning problem optimalizace - nutna oprava
#endif
  t->add_item         =  FAST_ADD_ITEM ;
  t->check_item_inserted  =  CHECK_ADD_ITEM ;
  t->add_item_nocheck =  FAST_ADD_ITEM ;

  t->compar_item   =  GET_COMPAR_ITEM;
  t->next_item     =  GET_NEXT_ITEM;
  t->compar_key    =  GET_COMPAR_KEY;
  t->find_return   =  GET_FIND_RETURN;

  t->already_inserted =  GET_ALREADY_INSERTED;
  t->get_hash         =  GET_ITEM_HASH ;
  t->__print          = __PRINT;
  t->__protect        = __PROTECT;


  t->get_chain_key    = GET_CHAIN_KEY;
  t->has_chain_item   = HAS_CHAIN_ITEM;
  t->next_chain_item  = NEXT_CHAIN_ITEM;
  t->get_chain_item   = GET_CHAIN_ITEM;

  t->item_hash        = GET_ITEMHASH_FUNCTION;

  t->is_item_chain_start  = IS_ITEM_CHAIN_START;
  t->get_item_chain_start = GET_ITEM_CHAIN_START;
  t->create_chain_item    = CREATE_CHAIN_ITEM;

#if 1
  t->remove_item =  REMOVE_ITEM;
  t->file_remove =  FILE_REMOVE;
  t->chain_remove_item =  CHAIN_REMOVE_ITEM;
  t->chain_file_remove =  FILE_CHAIN_REMOVE;
#else
  t->remove_item = 0;
  t->file_remove = 0;
#endif


#if 0
  t->__inserted       =  __INSERTED;
#endif

/*   t->file_remove = 0; */

}
#undef REMOVE_FROM_FOLDER

