/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#ifndef ADD_FCE_H
#define ADD_FCE_H
#include "sparse/symbol.h"
#include "sparse/expression.h"
#include "sparse/token.h"
#include "sparse/lib.h"
#include "sparse/allocate.h"
#include "sparse/token.h"
#include "sparse/parse.h"
#include "sparse/symbol.h"
#include "sparse/expression.h"
#include "sparse/linearize.h"
#include "sparse/scope.h"
extern struct symbol* act_fn;

void add_init();
void add_push();
void add_pop();

void set_symbol_at     (struct symbol* ) ;
void set_token_at      (struct token*  ) ;
void set_expression_at (struct expression* );
void set_scope_at      (struct scope* )  ;
void nothing_at        (struct position );

typedef struct {
  int (*cond_fce_inline    )   (struct symbol* );
  int (*cond_fce_use       )   (struct symbol* , struct expression* ) ;
  int (*cond_fce_def       )   (struct symbol* ) ;
  int (*cond_fce_dec       )   (struct symbol* ) ;
  int (*cond_var_dec       )   (struct symbol*, struct symbol* )     ;
  int (*cond_var_use       )   (struct symbol*, struct expression* ) ;
  int (*cond_incl          )   (const char* , struct token* ) ;
  int (*cond_mac_def     )   (struct symbol* ) ;
  int (*cond_mac_use     )   (struct symbol*, struct token* ) ;
  int (*cond_mac_defd    )   (struct token* )      ;
  int (*cond_mac_undef   )   (struct token* )      ;
  int (*cond_str_dec    )   (struct symbol* ) ;
  int (*cond_str_def    )   (struct symbol* ) ;
  int (*cond_str_use    )   (struct token* , struct position* ) ;
  int (*cond_uni_dec    )   (struct symbol* ) ;
  int (*cond_uni_def    )   (struct symbol* ) ;
  int (*cond_uni_use    )   (struct token* , struct position* ) ;
  int (*cond_tdef       )   (struct symbol* ) ;
  int (*cond_tdef_use   )   (struct symbol* , struct position* ) ;
  int (*cond_cast       )   ( struct expression* ) ;
  int (*cond_struct_member  )   ( struct expression* ) ;

  void (*action_fce_inline    )   (struct symbol* );
  void (*action_fce_use       )   (struct symbol* , struct expression* ) ;
  void (*action_fce_def       )   (struct symbol* ) ;
  void (*action_fce_dec       )   (struct symbol* ) ;
  void (*action_var_dec       )   (struct symbol*, struct symbol* )     ;
  void (*action_var_use       )   (struct symbol*, struct expression* ) ;
  void (*action_incl       )   (const char* , struct token* ) ;
  void (*action_mac_def     )   (struct symbol* ) ;
  void (*action_mac_use     )   (struct symbol*, struct token* ) ;
  void (*action_mac_defd    )   (struct token* )      ;
  void (*action_mac_undef   )   (struct token* )      ;
  void (*action_str_dec    )   (struct symbol* ) ;
  void (*action_str_def    )   (struct symbol* ) ;
  void (*action_str_use    )   (struct token* , struct position* ) ;
  void (*action_uni_dec    )   (struct symbol* ) ;
  void (*action_uni_def    )   (struct symbol* ) ;
  void (*action_uni_use    )   (struct token* , struct position* ) ;
  void (*action_tdef      )   (struct symbol* ) ;
  void (*action_tdef_use   )   (struct symbol* , struct position* ) ;

  int (*cond_symbol_at     )   (struct symbol* ) ;
  int (*cond_token_at      )   (struct token*  ) ;
  int (*cond_expression_at )   (struct expression* );
  int (*cond_scope_at      )   (struct scope* )  ;
  int (*cond_nothing_at )   (struct position) ;

  void (*action_symbol_at     )   (struct symbol* ) ;
  void (*action_token_at      )   (struct token*  ) ;
  void (*action_expression_at )   (struct expression* );
  void (*action_scope_at      )   (struct scope* )  ;
  void (*action_nothing_at    )   (struct position) ;
  void (*action_cast       )   ( struct expression* ) ;
  void (*action_struct_member)   ( struct expression* ) ;
} add_fce_t;

extern add_fce_t add_fce;

extern int fce_inline_deactivate;
extern int fce_use_deactivate;
extern int fce_def_deactivate;
extern int fce_dec_deactivate;
extern int var_dec_deactivate;
extern int var_use_deactivate;
extern int incl_deactivate;
extern int mac_def_deactivate;
extern int mac_use_deactivate;
extern int mac_defd_deactivate;
extern int mac_udef_deactivate;
extern int str_dec_deactivate;
extern int str_def_deactivate;
extern int str_use_deactivate;
extern int tdef_deactivate;
extern int tdef_use_deactivate;
extern int cast_deactivate;
extern int struct_member_deactivate;


#define add_function(x,y,z) \
add_fce.action_ ## x = y; \
add_fce.cond_ ## x = z;






void init_functions();
typedef struct {
  void (*add_function_use)    (struct symbol* act_fn, struct expression *expr);
  void (*add_function_call)   (struct symbol* sym, struct expression* ) ;
  void (*add_function_def)    (struct symbol* )     ;
  void (*add_function_dec)    (struct symbol* )     ;

  void (*add_static_variable_dec)  (struct symbol*, struct symbol* )     ;
  void (*add_variable_dec)    (struct symbol*, struct symbol* )     ;
  void (*add_variable_use)    (struct symbol*, struct expression* ) ;

  void (*add_include)         (const char* , struct token* ) ;

  void (*add_macro_define)    (struct symbol* ) ;
  void (*add_macro_use)       (struct symbol*, struct token* ) ;
  void (*add_macro_defined)   (struct token* )      ;
  void (*add_macro_undef)     (struct token* )      ;

  void (*add_struct_dec)      (struct symbol* ) ;
  void (*add_struct_def)      (struct symbol* ) ;
  void (*add_struct_use)      (struct token* , struct position* ) ;

  void (*add_union_dec)       (struct symbol* ) ;
  void (*add_union_def)       (struct symbol* ) ;
  void (*add_union_use)       (struct token* , struct position* ) ;

  void (*add_typedef_)        (struct symbol* ) ;
  void (*add_typedef_use)     (struct symbol*, struct position*  ) ;
  void (*add_cast)            (struct expression* ) ;
  void (*add_struct_member)   ( struct expression* ) ;

} add_functions_t;

add_functions_t add_functions_backup();
void  add_functions_restore(add_functions_t);

void (*add_struct_member) (struct expression* ) ;
void (*add_cast)          (struct expression *expr);
void (*add_function_use)  (struct symbol* act_fn, struct expression *expr);
void (*add_function_call) (struct symbol* sym, struct expression* ) ;
void (*add_function_def)  (struct symbol* )     ;
void (*add_function_inline)      (struct symbol* )     ;
void (*add_function_dec)         (struct symbol* )     ;
void (*add_static_variable_dec)  (struct symbol*, struct symbol* )     ;
void (*add_variable_dec)  (struct symbol*, struct symbol* )     ;
void (*add_variable_use)  (struct symbol*, struct expression* ) ;
void (*add_include)       (const char* , struct token* ) ;
void (*add_macro_define)  (struct symbol* ) ;
void (*add_macro_use)     (struct symbol*, struct token* ) ;
void (*add_macro_defined) (struct token* )      ;
void (*add_macro_undef)   (struct token* )      ;
void (*add_struct_dec)          (struct symbol* ) ;
void (*add_struct_def)          (struct symbol* ) ;
void (*add_struct_use)          (struct token* , struct position* ) ;
void (*add_union_dec)           (struct symbol* ) ;
void (*add_union_def)           (struct symbol* ) ;
void (*add_union_use)           (struct token* , struct position* ) ;
void (*add_typedef_)            (struct symbol* ) ;
void (*add_typedef_use)         (struct symbol* , struct position*) ;
void (*add_symbol_at)           (struct symbol*);
void (*add_nothing_at)          (struct position );
void (*add_evaluate_expression) (struct symbol*, struct expression *);

#endif
