/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

//#define PRINT_DEBUG
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <errno.h>

#include "sparse/symbol.h"
#include "sparse/expression.h"
#include "sparse/token.h"


#include "base/gf.h"
#include "base/xalloc.h"
#include "base/macros.h"
#include "database/memsave.h"


#include "interface/add-fce.h"
#include "interface/output.h"


static xalloc_context sparse_context = { 0,0,0,0,"sparse context" };

#define MASK_ONE(i,mod,mod_mask,pos_mask)\
  (mod & mod_mask ? i | pos_mask : i)

#define DEC_MASK(line,is)\
  ( is ? i | MASK_DECLARATION  : i)

#define CAST_MASK(line,is)\
  ( is ? i | MASK_CAST   : i)

#define MASK_LINE(line,is)\
  ( is ? line | MASK_SYSTEMINCLUDE : line)

inline int MASK_POS(int pos, unsigned long mod)
{
  pos=MASK_ONE(pos, mod, MOD_REGISTER   , MASK_REGISTER );  
  pos=MASK_ONE(pos, mod, MOD_STATIC	, MASK_STATIC   );    
  pos=MASK_ONE(pos, mod, MOD_EXTERN	, MASK_EXTERN   );    
  pos=MASK_ONE(pos, mod, MOD_CONST      , MASK_CONST    );    
  pos=MASK_ONE(pos, mod, MOD_VOLATILE   , MASK_VOLATILE );    
  pos=MASK_ONE(pos, mod, MOD_SIGNED	, MASK_SIGNED   );    
  pos=MASK_ONE(pos, mod, MOD_ADDRESSABLE, MASK_ADDRESSABLE);
  pos=MASK_ONE(pos, mod, MOD_INLINE	, MASK_INLINE   );    
  pos=MASK_ONE(pos, mod, MOD_ATTRIBUTE  , MASK_ATTRIBUTE);    
  pos=MASK_ONE(pos, mod, MOD_UNSIGNED   , MASK_UNSIGNED );    
  pos=MASK_ONE(pos, mod, MOD_NOCAST      , MASK_NOCAST);      
  pos=MASK_ONE(pos, mod, MOD_NODEREF     , MASK_NODEREF);     
  pos=MASK_ONE(pos, mod, MOD_ACCESSED    , MASK_ACCESSED); 
  pos=MASK_ONE(pos, mod, MOD_TOPLEVEL    , MASK_TOPLEVEL);    
  pos=MASK_ONE(pos, mod, MOD_TYPEDEF    , MASK_TYPEDEF  );    
  pos=MASK_ONE(pos, mod, MOD_STRUCTOF   , MASK_STRUCTOF );    
  pos=MASK_ONE(pos, mod, MOD_UNIONOF    , MASK_UNIONOF  );    
  pos=MASK_ONE(pos, mod, MOD_ENUMOF     , MASK_ENUMOF   );    
  pos=MASK_ONE(pos, mod, MOD_TYPEOF     , MASK_TYPEOF   );    
  pos=MASK_ONE(pos, mod, MOD_AUTO       , MASK_AUTO     );
  return pos;
}

extern int show_active_file; /* cmd.c */
struct symbol_list *spot_sparse_symbol_list(const char* str);

void consider_symbol_list(struct symbol_list *list);

/* void sparse_build_run( const char* str) */
/* { */
/*   top_level_symbol=1; */
/*   xalloc_context* c = xalloc_get_context(); */
/*   xalloc_set_context(&sparse_context); */
/*   spot_sparse( str); */
/*   xalloc_set_context(c); */
/* } */

/*
  inicializuje sparse
 */
void sparse_build_init()
{
  xalloc_context* c = xalloc_get_context();
  xalloc_set_context(&sparse_context);
  spot_sparse_initialize();
  xalloc_set_context(c);
}

/* 
   dealokuje vsechnu pamet kterou sparse v minulem behu naalokovalo 
 */
void sparse_build_clear()
{
  xalloc_context* c = xalloc_get_context();
  xalloc_set_context(&sparse_context);

  clear_token_alloc();
  clear_ident_alloc();
  clear_symbol_alloc();
  clear_scope_alloc();

  clear_ident_alloc();
  clear_token_alloc();
  clear_symbol_alloc();
  clear_expression_alloc();
  clear_statement_alloc();
  clear_string_alloc();
  clear_scope_alloc();
  clear_bytes_alloc();
  clear_basic_block_alloc();
  clear_entrypoint_alloc();
  clear_instruction_alloc();
  clear_multijmp_alloc();
  clear_pseudo_alloc();
  
  reset_parse();
  reset_scope();
  reset_symbol();
  reset_preprocess();
  reset_tokenize();
  //  reset_storage();

  reset_allocate();
  reset_lib();

  xalloc_set_context(&sparse_context);
  xfree_all();
  xalloc_set_context(c);
}

void clear_and_reset() 
{
  sparse_build_clear();
  sparse_build_init();
}

struct symbol_list *spot_sparse_symbol_list(const char* str)
{
  xalloc_context* c = xalloc_get_context();
  xalloc_set_context(&sparse_context);
  extern int verbose_spot;
  if (verbose_spot) printf("build file : '%s'\n",str);
  { 
    extern FILE* sparse_output;  
    if (sparse_output) 
      fprintf(sparse_output
	      , "\n-----------------\nFILE = %s\n"
	      , str); 
  } 

  struct symbol_list * l = spot_sparse( str);
  xalloc_set_context(c);
  return l;
}


void spotsparse_file(const char* string)
{
  static int cleaned;
  if (!cleaned) clear_and_reset();
  xalloc_context* context = xalloc_get_context();
  xalloc_set_context(&sparse_context);

  char* str = malloc(strlen(string)+1);
  strcpy(str,string);

  /* struct symbol_list * list =  */
  spot_sparse_symbol_list( str);
  add_symbol_at(0); 

  free(str);

  xalloc_set_context(context);
  clear_and_reset(); cleaned=1;
}



static void func_null(){}
static void do_evaluate_expression(struct symbol* act_fn, struct expression *expr);

add_functions_t add_functions_backup()
{
  add_functions_t t;
  t.add_function_call = add_function_call ;
  t.add_function_dec  = add_function_dec  ;
  t.add_macro_use     = add_macro_use     ;
  t.add_macro_defined = add_macro_defined ;
  t.add_include       = add_include       ;
  t.add_macro_define  = add_macro_define  ;
  t.add_typedef_      = add_typedef_      ;
  t.add_variable_dec  = add_variable_dec  ;
  t.add_variable_use  = add_variable_use  ;
  t.add_struct_dec    = add_struct_dec    ;
  t.add_struct_def    = add_struct_def    ;
  t.add_cast          = add_cast          ;
  t.add_struct_member = add_struct_member;


  add_function_call = func_null ;
  add_function_dec  = func_null ;
  add_macro_use     = func_null ;
  add_macro_defined = func_null ;
  add_include       = func_null ;
  add_macro_define  = func_null ;
  add_typedef_      = func_null ;
  add_variable_dec  = func_null ;
  add_variable_use  = func_null ;
  add_struct_dec    = func_null ;
  add_struct_def    = func_null ;
  add_cast          = func_null ;
  add_struct_member = func_null ;
  return t;
}

void  add_functions_restore(add_functions_t t)
{
  add_function_call = t.add_function_call ;
  add_function_dec  = t.add_function_dec  ;
  add_macro_use     = t.add_macro_use     ;
  add_macro_defined = t.add_macro_defined ;
  add_include       = t.add_include       ;
  add_macro_define  = t.add_macro_define  ;
  add_typedef_      = t.add_typedef_      ;
  add_variable_dec  = t.add_variable_dec  ;
  add_variable_use  = t.add_variable_use  ;
  add_struct_dec    = t.add_struct_dec    ;
  add_struct_def    = t.add_struct_def    ;
  add_cast          = t.add_cast          ;
  add_struct_member = t.add_struct_member ;
}

void init_functions()
{
  add_function_use    = func_null;
  add_function_call   = func_null;
  add_function_inline = func_null;
  add_function_dec    = func_null;
  add_function_def    = func_null;
  add_macro_use       = func_null;
  add_macro_defined   = func_null;
  add_include         = func_null;
  add_macro_define    = func_null;
  add_typedef_        = func_null;
  add_variable_dec    = func_null;
  add_variable_use    = func_null;
  add_struct_dec      = func_null;
  add_struct_def      = func_null;
  add_cast            = func_null;
  add_struct_member   = func_null;
  add_evaluate_expression = func_null;
  add_evaluate_expression = do_evaluate_expression;
}


#if 0
static void build_action_macro_undef          (struct token*  )                      ;
#endif



//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////




#include "build/struct_content.h"
#include "build/functions.h"
#include "build/during_sparse.h"
#include "build/visible_at-skeleton.h"
#include "build/visible_variable_at.h"
#include "build/visible_macro_def_at.h"
#include "build/visible_fce_dec_at.h"
#include "build/visible_at.h"
#include "build/called_by_function_at.h"
#include "build/used_by_function_at.h"
#include "build/macro_use.h"
#include "build/local_var_dec.h"
#include "build/local_var_use.h"



/*
  nema-li provadet kontrolu vrati 0
  vrati 1 pokud je systemovy include
  0 jinak
 */
int system_include(int s)
{
  if (!s) return 1;  /* co je builtin je systemove */

  struct stream* stream = get_stream(s);
  return stream->sys_include;
}

extern int match_some(const char* file);
int inside_project(int s)
{
  return 1;
  //  return 1;
  //  if (s==-1) return 0;
  struct stream* stream = get_stream(s);
  if (!stream) return 0;
  if (stream->project==PROJECT_YES) return 1;
  if (stream->project==PROJECT_NO)  return 0;
  stream->project=match_some(stream->name)?PROJECT_YES:PROJECT_NO;
  if (stream->project==PROJECT_YES) return 1;
  else return 0;
}
