/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

//#define MEMBER_DEBUG

#ifndef SPARSE_BUILD_FUNCTIONS
#define SPARSE_BUILD_FUNCTIONS
int inside_project(int stream);


static void build_action_typedef_use       (struct symbol* s, struct position* pos);
static void build_action_function_use      (struct symbol* , struct expression *);
static void build_action_function_def      (struct symbol* )                      ;
static void build_action_function_inline   (struct symbol* )                      ;
static void build_action_function_dec      (struct symbol* )                      ;
static void build_action_variable_dec      (struct symbol*, struct symbol* )      ;
static void build_action_variable_use      (struct symbol*, struct expression* )  ;
static void build_action_include           (const char*  , struct token* )         ;
static void build_action_macro_define      (struct symbol* )                      ;
static void build_action_macro_use         (struct symbol*, struct token* )       ;
static void build_action_macro_defined     (struct token*  )                      ;
static void build_action_struct_dec        (struct symbol* )                      ;
static void build_action_struct_def        (struct symbol* )                      ;
static void build_action_typedef_          (struct symbol* )                      ;
static void build_action_struct_use        (struct token* , struct position* );
static void build_action_CAST              ( struct expression *expr);

#define SMALL(x,y,l,p,typ)                 \
xalloc_context* c = xalloc_get_context();  \
xalloc_set_context(&sparse_context);       \
add_mem_ ## typ (x,y,l,p);                 \
xalloc_set_context(c);


#define LARGE(file,ident,l,p,df,dl,dp,typ)        \
xalloc_context* c = xalloc_get_context();         \
xalloc_set_context(&sparse_context);              \
add_mem_ ## typ (file,ident,l,p,df,dl,dp);        \
xalloc_set_context(c);


static void build_action_function_use (struct symbol* act_fn, struct expression *expr)
{
  if (!expr || !expr->symbol) return;
  if (expr->type!=EXPR_SYMBOL) return; /* spatne zavolano */
  struct symbol* sym=expr->symbol->ctype.base_type;
  if (!sym || sym->type!=SYM_FN ) return; /* spatne zavolano */
  /* dava mi vsechna pouziti dane funkce --- predani adresy*/

/*   if (!strcmp(sym->ident?sym->ident->name:expr->symbol_name->name */
/* 	     , "build_init")) printf("build_init\n"); */

  SMALL ( ( stream_name( expr->pos.stream ) ) ,
	  sym->ident?sym->ident->name:expr->symbol_name->name,
	  expr->pos.line,
	  expr->pos.pos , function_use);
}

extern int system_include(int s); /* config.c */
static void build_action_function_def         (struct symbol* sym)                       
{
  if (!sym || !sym->ident) return;
  if ( sym->type!=SYM_FN && sym->type!=SYM_NODE)
    {
      fprintf(stderr,"database save function_def says : not a function...\n");
      return;
    }

  SMALL (( stream_name( sym->pos.stream)) 
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include( sym->pos.stream) )
	 , MASK_POS  ( sym->pos.pos , sym->ctype.modifiers )
	 , function_def );
}
static void build_action_function_inline      (struct symbol* s)                      
{
  return build_action_function_def(s);
}
static void build_action_function_dec         (struct symbol* sym)                       
{
  if (!sym || !sym->ident) return;
  if ( sym->type!=SYM_FN && sym->type!=SYM_NODE)
    {
      fprintf(stderr,"database save function_def says : not a function...\n");
      return;
    }
  SMALL (( stream_name( sym->pos.stream))
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include( sym->pos.stream) )
	 , MASK_POS  ( sym->pos.pos , sym->ctype.modifiers )
	 , function_dec);
}
static void build_action_variable_dec         (struct symbol* fn, struct symbol* sym)
{
  if ( !sym || !sym->ident) return;
  //  if ( !(sym->ctype.modifiers & MOD_NONLOCAL)) return; // lokalni promenne
  //  if ( fn ) return;  // lokalni promenne

  SMALL (
	 ( stream_name( sym->pos.stream))
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include( sym->pos.stream) )
	 , MASK_POS  ( sym->pos.pos , sym->ctype.modifiers )
	 , variable_dec);
}
static void build_action_macro_define         (struct symbol* sym)
{
  if (!sym || !sym->ident) return;
  SMALL (
	 ( stream_name( sym->pos.stream)) 
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include( sym->pos.stream) )
	 , sym->pos.pos
	 , macro_def);
}
static void build_action_macro_use            (struct symbol* sym, struct token* token)
{
  if (sym && !sym->ident) return;
  const char*s=sym?sym->ident->name:token->ident->name;
  SMALL (
	 ( stream_name( token->pos.stream)) 
	 , s
	 , MASK_LINE ( token->pos.line, system_include( token->pos.stream) )
	 , token->pos.pos
	 , macro_use);
}
static void build_action_macro_defined        (struct token* token)                      
{
  SMALL (
	 ( stream_name( token->pos.stream))
	 , token->ident->name
	 , MASK_LINE ( token->pos.line, system_include( token->pos.stream) )
	 , token->pos.pos
	 , macro_use);
}

static int begin_double_dot(const char* filename)
{
  if (*filename=='.' &&  *(filename+1)=='.' && *(filename+2)=='/') return 1;
  return 0;
}

static void build_action_include              (const char* filename, struct token* token)
{
  if (!filename || !token) return;
  const char* f1=stream_name(token->pos.stream);
  const char* f2=filename;

  /*
   * pripad kdy sparse/lib.c (f1)  includuje ../cspot.h (f2)
   * chci aby se zobrazilo ze sparse/lib.c includuje cspot.h  (chci aby ke vsemu byla relativni cesta z adresare projektu)
   */
  if(begin_double_dot(f2))
    {
      char buf[2*MAX_LEN];
      int l=strlen(f1);
      strcpy(buf,f1);
      int j;
      for (j=l;j>=0;j--) if (*(buf+j)=='/') break;
      strcpy(buf+j+1,f2);
      f2=get_filename(buf);     
    }

  /*       included           including           where included     */
  SMALL (( stream_name( token->pos.stream)) 
	 , f2
	 , MASK_LINE ( token->pos.line, system_include( token->pos.stream) )
	 , token->pos.pos
	 , include);

  add_mem_included ( stream_name( token->pos.stream) , f2 );
}


static void build_action_variable_use         (struct symbol* fn, struct expression* ex)
{
  if (!ex || !ex->symbol) return; 
  //  if (! ex->pos.stream ) return;
  // test zda jde o lokalni promennou
  /* lokalni=ve stejnem souboru, a funkce zacina drive nez je promenna deklarovana*/

/*   if ( fn->pos.stream == ex->symbol->pos.stream */
/*        && ( fn->pos.line<ex->symbol->pos.line || */
/* 	    ( fn->pos.line==ex->symbol->pos.line && */
/* 	      fn->pos.pos<=ex->symbol->pos.pos))) */
/*     return; */

//  if (!fn) return;

/*   struct symbol* sym = ex->symbol->ctype.base_type; */
/*   if (sym && sym->type == SYM_STRUCT) */
/*     { */
/*       printf(""); */
/*       if (sym->ident && sym->ident->name) */
/* 	{ */
	  
/* 	} */
/*     } */

  if ( !(ex->symbol->ctype.modifiers & MOD_NONLOCAL))
    return;
  LARGE ((stream_name(ex->pos.stream))
	 , ex->symbol->ident->name
	 , ex->pos.line
	 , ex->pos.pos
	 , (stream_name(ex->symbol->pos.stream))
	 , MASK_LINE ( ex->symbol->pos.line, system_include(ex->symbol->pos.stream) )
	 , MASK_POS  ( ex->symbol->pos.pos , ex->symbol->ctype.modifiers )
	 , variable_use );
}

static void build_action_struct_dec           (struct symbol* sym)                      
{
  if (!sym) return;
  if (!sym->ident) return;
  SMALL (
	 ( stream_name(sym->pos.stream))
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include(sym->pos.stream) )
	 , MASK_POS  ( sym->pos.pos , sym->ctype.modifiers )
	 , struct_dec);
}


static void build_action_struct_def           (struct symbol* sym)
{
  if (!sym) return;
  if (!sym->ident) return;
  SMALL (
	 ( stream_name(sym->pos.stream)) 
	 , sym->ident->name
	 , MASK_LINE ( sym->pos.line, system_include(sym->pos.stream) )
	 , MASK_POS  ( sym->pos.pos , sym->ctype.modifiers )
	 , struct_def );
}


static void build_action_typedef_             (struct symbol* s)                      
{
  if (!s||!s->ident) return;
  if (!inside_project(s->pos.stream)) return;
  SMALL ( stream_name(s->pos.stream)
	  , s->ident->name
	  , MASK_LINE ( s->pos.line, system_include(s->pos.stream) )
	  , MASK_POS  ( s->pos.pos , s->ctype.modifiers )
	  , typedef_);
}



static void build_action_typedef_use          (struct symbol* s, struct position* pos)
{
/*   printf("typedef use ? at %s %d:%d\n" */
/* 	 , stream_name(pos->stream) */
/* 	 , pos->line */
/* 	 , pos->pos ); */
  if (!s || !s->ident) return;
  SMALL( stream_name(pos->stream)
	 , s->ident->name
	 , MASK_LINE ( pos->line, system_include(s->pos.stream) )
	 , MASK_POS  ( pos->pos , s->ctype.modifiers )
	 , typedef_use
	 );
/*   printf("typedef use %s at %s %d:%d\n" */
/* 	 , s->ident->name */
/* 	 , stream_name(pos->stream) */
/* 	 , pos->line */
/* 	 , pos->pos ); */
	 
}

static void build_action_CAST ( struct expression *expr)
{
#ifdef MY_WARNINGS
#warning vypnut cast nebot se sem dostavalo i n asledujici misto typedefu rikalo ze je tam struktura
#endif
  return;
  int debug=0;
  {
    struct symbol* cast_t=0;
    const char* name="";
    int sys=0;
    if (debug) printf("cast at %s %d:%d\n",
	   stream_name(expr->pos.stream),
	   expr->pos.line,
	   expr->pos.pos
	   );		   
    if (!expr->cast_type) 
      {
	if (debug) printf("no cast type\n");
	return;
      }
    cast_t=expr->cast_type->ctype.base_type;
    if (!cast_t)
      {
	if (debug) printf("no cast base_type\n");
	return;
      }
    while (cast_t && cast_t->type == SYM_PTR)
      {
	cast_t = cast_t->ctype.base_type;
      }
    if (!cast_t)
      {
	if (debug) printf("base_type se stalo 0 !\n");
	return;
      }
    if (cast_t->type == SYM_BASETYPE ) // zrejme primitivni typ
      {
	sys=1;
	name="void";
	
	if (cast_t->ctype.modifiers & MOD_SIGNED)
	  name="signed";
	if (cast_t->ctype.modifiers & MOD_UNSIGNED)
	  name="unsigned";
	
	if (cast_t->ctype.modifiers & MOD_CHAR)
	  name="char";
	if (cast_t->ctype.modifiers & MOD_SHORT)
	  name="short";
	if (cast_t->ctype.modifiers & MOD_LONG)
	  name="long";
	if (cast_t->ctype.modifiers & MOD_LONGLONG)
	  name="longlong";
	
	if (cast_t->ctype.modifiers & MOD_USERTYPE)
	  name="usertype";
	
	return; //nechci ukladat tyto casty
	goto print;
      }
    if (!cast_t->ident)
      {
	if (debug) printf("no cast base_type ident\n");
	return;
      }
    if (!cast_t->ident->name)
      {
	if (debug) printf("no cast base_type ident name\n");
	return;
      }
    else
      {
	name=cast_t->ident->name;
      }
  print:
    if (debug) printf("ident = '%s'\n-----\n",name);

#if 1    
    if (cast_t->type == SYM_STRUCT
	|| cast_t->type == SYM_UNION )
      {
	SMALL( stream_name(expr->pos.stream)
	       , name
	       , MASK_LINE ( expr->pos.line, sys )
	       , expr->pos.pos 
	       , struct_use
	       );
      }
    else
      {
	SMALL( stream_name(expr->pos.stream)
	       , name
	       , MASK_LINE ( expr->pos.line, sys )
	       , expr->pos.pos 
	       , typedef_use
	       );
      }
#endif
  }

}


static void build_action_struct_use           (struct token* token, struct position* pos)                      
{
/*   printf("* %s at %s %d:%d\n" */
/* 	 , token->ident->name */
/* 	 , stream_name(token->pos.stream) */
/* 	 , token->pos.line */
/* 	 , token->pos.pos); */
  SMALL (stream_name(token->pos.stream)
	 , token->ident->name
	 , pos->line
	 , pos->pos
	 , struct_use
	 );
}

#if 0
static void build_action_macro_undef          (struct token* t)                      
{
}

#endif


static void build_action_STRUCT_member_error ( struct expression *e);

static char* build_action_STRUCT_member_basetype ( struct expression *e)
{
  struct symbol* sym;
  sym=e->symbol->ctype.base_type;
  while (sym)
    {
      switch (sym->type)
	{
	case SYM_PTR:
	case SYM_TYPEDEF:
	  sym=sym->ctype.base_type;
	  break;
	case SYM_STRUCT:
	case SYM_UNION:
	  if (sym->ident)
	    return sym->ident->name;
	  else 
	    return 0;
	default:
	  return 0;
#warning default... mrknout se sem --- generovani jmena struktury
	}
    }
  return 0;
}

static struct expression*
build_action_STRUCT_member_get_upper (struct expression *e
				      , char** instance
				      , char** type
				      , struct position** pos )
{
  while (e && e->type!=EXPR_SYMBOL)
    { 
      switch (e->type)
	{
	case EXPR_BINOP: // pole struktur !!!
	  return 0;
	case EXPR_PREOP: e = e->unop; break;
	case EXPR_CAST:  e = e->cast_expression; break;
	case EXPR_DEREF: e = e->deref; break;
	case EXPR_SYMBOL: break;
	default:
	  {
	    build_action_STRUCT_member_error(e);
	    return 0;
	  }
	}
    }
  if (!e)
    {
#ifdef MEMBER_DEBUG
      printf("e == nil\n");
#endif
      return e;
    }
  if (e->type==EXPR_SYMBOL)
    { 
      *instance = e->symbol_name->name; 
      *type =  build_action_STRUCT_member_basetype(e);

      if (e->symbol->pos.stream)
	{
	  *pos=&e->symbol->pos;
	}
      else
	{
	  *pos=0;
	}
    }
  else
    {
      *instance = 0;
      *type = 0;
      *pos=0;
    }
  return e;
}


static void build_action_STRUCT_member_error ( struct expression *e)
{
#ifndef MEMBER_DEBUG
  return;
#endif
  printf("at %s %d:%d ",stream_name(e->pos.stream),e->pos.line,e->pos.pos);
  switch (e->type)
    {
    case EXPR_VALUE: printf("EXPR_VALUE\n"); break;
      
    case EXPR_FVALUE:printf("EXPR_FVALUE\n"); break;
    case EXPR_STRING:printf("EXPR_STRING\n"); break;
      //	      case EXPR_UNOP:printf("EXPR_UNOP\n"); break;
    case EXPR_PREOP :printf("EXPR_PREOP \n"); break;
    case EXPR_POSTOP:printf("EXPR_POSTOP\n"); break;
    case EXPR_SYMBOL:printf("EXPR_SYMBOL\n"); break;
    case EXPR_TYPE:printf("EXPR_TYPE\n"); break;
    case EXPR_STATEMENT:printf("EXPR_STATEMENT\n"); break;
    case EXPR_BINOP:printf("EXPR_BINOP\n"); break;
    case EXPR_COMMA:printf("EXPR_COMMA\n"); break;
    case EXPR_COMPARE:printf("EXPR_COMPARE\n"); break;
    case EXPR_LOGICAL :printf("EXPR_LOGICAL \n"); break;
    case EXPR_ASSIGNMENT:printf("EXPR_ASSIGNMENT\n"); break;
    case EXPR_DEREF:printf("EXPR_DEREF\n"); break;
    case EXPR_SLICE:printf("EXPR_SLICE\n"); break;
    case EXPR_CAST :printf("EXPR_CAST \n"); break;
    case EXPR_SIZEOF:printf("EXPR_SIZEOF\n"); break;
    case EXPR_CONDITIONAL:printf("EXPR_CONDITIONAL\n"); break;
    case EXPR_SELECT:printf("EXPR_SELECT\n"); break;
    case EXPR_CALL:printf("EXPR_CALL\n"); break;
    case EXPR_LABEL:printf("EXPR_LABEL\n"); break;
    case EXPR_INITIALIZER:printf("EXPR_INITIALIZER\n"); break;
    case EXPR_IDENTIFIER:printf("EXPR_IDENTIFIER\n"); break;
    case EXPR_INDEX:printf("EXPR_INDEX\n"); break;
    case EXPR_POS:printf("EXPR_POS\n"); break;
    default:printf("EXPR not handled\n");
    }
}

static void build_action_STRUCT_member ( struct expression *expr)
{
#warning zatim nepouzivat
  return;

#warning tady opravit STRUCT_MEMBER
  /*
    expr->deref->symbol_name --- jmeno promenne
    expr->member --- jmeno polozky struktury
    expr->deref->ctype.base_type->ident  --- jmeno struktury
  */

  if (! expr->deref
      || ! expr->member
      || ! expr->deref->symbol // nedeklarovana promenna
      ) return;

#warning struct member - promyslet co s globalnima a lokalnima...
  //  if ( !(expr->deref->symbol->ctype.modifiers & MOD_NONLOCAL))    { return; }

  //  if (!expr->pos.stream) {return;} // ?builtin

  //  if (! expr->deref->symbol->pos.stream) return; // sym_uninitialized 
  //  if ( expr->deref->symbol->ctype.base_type->type != SYM_STRUCT)  { return; }

#define MAXSIZE MAX_LEN*2
  char instance_var[MAXSIZE]; memset(instance_var,MAX_LEN*2,0);
  int instance_var_len=1; // terminating null
  int instance_var_error=0;
  char type_var[MAXSIZE];     memset(type_var,MAX_LEN*2,0);
  int type_var_len=1;  // terminating null
  int type_var_error=0;

  const char* member          = expr->member->name;
  const char* dec_file;
  int dec_line, dec_pos;
  struct expression* e = expr;
  {
    int member_len      = strlen(member);
    strcpy(instance_var+MAXSIZE-member_len-1,member);
    strcpy(type_var+MAXSIZE-member_len-1,member);
    instance_var_len+=member_len;
    type_var_len+=member_len;
    //    instance_var[MAXSIZE-instance_var_len]='.';
#ifdef MEMBER_DEBUG
    printf("\n--- %s\n",instance_var+MAXSIZE-instance_var_len);
#endif

    char* instance;
    char* type;
    struct position* pos;
    e = build_action_STRUCT_member_get_upper (e
					      , &instance 
					      , &type
					      , &pos );

    if (e)
      {
	{
	  if (instance)  {
	    int   instance_len = 0;
	    instance_len = strlen(instance);
	    memcpy(instance_var 
		   + MAXSIZE 
		   - instance_len 
		   - instance_var_len
		   - 1
		   , instance
		   , instance_len );
#ifdef MEMBER_DEBUG
	    printf("-- %s\n",instance);
#endif	    
	    instance_var[MAXSIZE-instance_var_len-1]='.';
	    instance_var_len+=instance_len+1;
	  }
	  else instance_var_error=1;

	  if (type) {
	    int   type_len     = 0;
	    type_len = strlen(type);
	    memcpy(type_var 
		   + MAXSIZE 
		   - type_len 
		   - type_var_len
		   - 1
		   , type
		   , type_len );
#ifdef MEMBER_DEBUG
	    printf("++ %s\n",type);
#endif	    
	    type_var[MAXSIZE-type_var_len-1]='.';
	    type_var_len+=type_len+1;
	  } else  type_var_error=1;
	}

	if (pos)
	  {
	    dec_file=(stream_name(pos->stream));
	    dec_line=pos->line;
	    dec_pos =pos->pos;
	  }
	else 
	  {
	    dec_file = "uninitialized";
	    dec_line = 0;            
	    dec_pos  = 0;       
	  }
      }

#ifdef MEMBER_DEBUG
    printf("... %s\n",instance_var+MAXSIZE-instance_var_len);
    printf("*** %s\n",type_var+MAXSIZE-type_var_len);
#endif
  }

#if 0
  LARGE ((stream_name(expr->pos.stream))
	 , name_var
	 , expr->pos.line
	 , expr->pos.pos
	 , dec_file
	 , dec_line
	 , dec_pos
	 //	 , MASK_LINE ( ex->symbol->pos.line, system_include(ex->symbol->pos.stream) )
	 //	 , MASK_POS  ( ex->symbol->pos.pos , ex->symbol->ctype.modifiers )
	 , variable_use );
#endif

  //  int   member_len      = strlen(member);

#if 0
  char name_str[MAX_LEN*2];
  int   struct_name_len = strlen(struct_name);
  strcpy(name_str,struct_name);
  name_str[struct_name_len]='.';
  strcpy(name_str+struct_name_len+1,member);

  LARGE ((stream_name(expr->deref->symbol->pos.stream))
	 , name_str
	 , expr->pos.line
	 , expr->pos.pos
	 , (stream_name(expr->deref->symbol->pos.stream))
	 , expr->deref->symbol->pos.line
	 , expr->deref->symbol->pos.pos
	 //	 , MASK_LINE ( ex->symbol->pos.line, system_include(ex->symbol->pos.stream) )
	 //	 , MASK_POS  ( ex->symbol->pos.pos , ex->symbol->ctype.modifiers )
	 , variable_use );
#endif
	 
}

void sparse_build_functions()
{
#define log_init_fce(x,y) add_fce.action_ ## x = build_action_ ## y;
  log_init_fce(fce_use, function_use)
  log_init_fce(fce_inline, function_inline)
  log_init_fce(fce_def, function_def)
  log_init_fce(fce_dec, function_dec)
  log_init_fce(var_dec, variable_dec)
  log_init_fce(var_use, variable_use)
  log_init_fce(incl,include)
  log_init_fce(mac_def,macro_define)
  log_init_fce(mac_use,macro_use)
  log_init_fce(mac_defd,macro_defined)
  log_init_fce(str_dec,struct_dec)
  log_init_fce(str_def,struct_def)
  log_init_fce(str_use,struct_use)
  log_init_fce(tdef,typedef_)
  log_init_fce(tdef_use,typedef_use)
  log_init_fce(cast,CAST)
  log_init_fce(struct_member,STRUCT_member)
#undef log_init_fce
}

#endif



