/*
 *
 *	(c) Petr Susil 2006 <susa@atrey.karlin.mff.cuni.cz>
 *
 *	This software may be freely distributed and used according to the terms
 *	of the GNU General Public License.
 */

#include <stdio.h>
#include "sparse/symbol.h"
#include "sparse/token.h"
#include "sparse/lib.h"
#include "sparse/allocate.h"
#include "sparse/token.h"
#include "sparse/parse.h"
#include "sparse/symbol.h"
#include "sparse/expression.h"
#include "sparse/linearize.h"
#include "sparse/scope.h"

#include "output.h"
#include "base/macros.h"

extern int MASK_POS(int pos, unsigned long mod); /*sparse-build.c*/

extern void (*add_function_use)     (struct symbol* act_fn, struct expression *expr);
extern void (*add_function_call)    (struct symbol* sym, struct expression* ) ;
extern void (*add_function_def)     (struct symbol* )     ;
extern void (*add_function_inline)  (struct symbol* )     ;
extern void (*add_function_dec)     (struct symbol* )     ;
extern void (*add_static_variable_dec)  (struct symbol*, struct symbol* )     ;
extern void (*add_variable_dec)     (struct symbol*, struct symbol* )     ;
extern void (*add_variable_use)     (struct symbol*, struct expression* ) ;
extern void (*add_include)          (const char* , struct token* ) ;
extern void (*add_macro_define)     (struct symbol* ) ;
extern void (*add_macro_use)        (struct symbol*, struct token* ) ;
extern void (*add_macro_defined)    (struct token* )      ;
extern void (*add_macro_undef)      (struct token* )      ;
extern void (*add_struct_dec)       (struct symbol* ) ;
extern void (*add_struct_def)       (struct symbol* ) ;
extern void (*add_struct_use)       (struct token* , struct position* ) ;
extern void (*add_union_dec)        (struct symbol* ) ;
extern void (*add_union_def)        (struct symbol* ) ;
extern void (*add_union_use)        (struct token* , struct position* ) ;
extern void (*add_typedef_)         (struct symbol* ) ;
extern void (*add_typedef_use)      (struct symbol* , struct position*) ;
extern void (*add_symbol_at)        (struct symbol*);
extern void (*add_nothing_at)       (struct position );
extern void (*add_evaluate_expression) (struct symbol*, struct expression *);

void output_macro_define  (struct symbol* sym)
{
  WRITE4(sym->ident->name,
	 stream_name(sym->pos.stream),
	 sym->pos.line,
	 sym->pos.pos
	 );
}
void output_typedef_      (struct symbol* sym)
{
  WRITE4 ( sym->ident->name
	 , stream_name(sym->pos.stream)
	 , sym->pos.line
	 , sym->pos.pos
	 );
}
static void write_struct_content(struct symbol* sym)
{
  struct symbol* s;
  FOR_EACH_PTR(sym->symbol_list, s) {
    switch (s->type)
      {
      case SYM_NODE:
	//	fprintf(stderr,"namespace=%d\n",s->namespace);;
	if (!s->ident)
	  {
	    write_struct_content(s->ctype.base_type);
	    continue;
	  }
	fprintf(stdout,"ident = %s\n",s->ident->name);
	break;
      default:
	fprintf(stderr,"default at output struct dec\n");
	break;
      }

  } END_FOR_EACH_PTR(s);
}

void output_struct_dec      (struct symbol* sym )
{
  if(cond_output_struct_dec && !cond_output_struct_dec(sym))
    return;
  WRITE4( sym->ident->name
	 , stream_name(sym->pos.stream)
	 , sym->pos.line
	 , sym->pos.pos
	 );
}

void output_struct_def      (struct symbol* sym )
{
  if(cond_output_struct_dec && !cond_output_struct_dec(sym))
    return;
  WRITE4( sym->ident->name
	 , stream_name(sym->pos.stream)
	 , sym->pos.line
	 , sym->pos.pos
	 );
  write_struct_content(sym);
}

void output_macro_defined (struct token* token)
{
  WRITE4(token->ident->name
	 , stream_name(token->pos.stream)
	 , token->pos.line
	 , token->pos.pos
	 );
}
void output_include       (const char* filename, struct token* token)
{
  if (cond_output_include && !cond_output_include(filename,token))
    return;
  WRITE4(filename
	 , stream_name(token->pos.stream)
	 , token->pos.line
	 , token->pos.pos
	 );
}
void output_macro_use     (struct symbol* sym, struct token* token)
{
  if (!sym) return output_macro_defined(token);
  WRITE7(token->ident->name // 	 sym->ident->name
	 ,stream_name(token->pos.stream)
	 ,token->pos.line
	 ,token->pos.pos
	 ,stream_name(sym->pos.stream)
	 ,sym->pos.line
	 ,sym->pos.pos
	 );
}
void output_variable_dec  (struct symbol* fn, struct symbol* s)
{
  if (!s) return;
  if (s->ident)
    {
      WRITE4(s->ident->name,stream_name(s->pos.stream), s->pos.line,s->pos.pos);
    }
  else
    printf("--- no ident at var dec\n");

}

void output_variable_use  (struct symbol* fn, struct expression* ex)
{
  if (!ex) return;
  if (ex->symbol->ident)
    WRITE7(ex->symbol->ident->name,
	   stream_name(ex->pos.stream),
	   ex->pos.line,
	   ex->pos.pos,
	   stream_name(ex->symbol->pos.stream),
	   ex->symbol->pos.line,
	   MASK_POS  ( ex->symbol->pos.pos , ex->symbol->ctype.modifiers ))
      //	   ex->symbol->pos.pos)
#ifdef PRINT_DEBUG
  else
    printf("--- no ident at var use %s %d:%d\n"
	   , stream_name(ex->pos.stream)
	   , ex->pos.line
	   , ex->pos.pos);
#endif
}

void output_function_use (struct symbol* sym, struct expression* ex)
{
  if (ex->type!=EXPR_SYMBOL) return; /* spatne zavolano */
  
  sym=ex->symbol->ctype.base_type;
  if (!sym || sym->type!=SYM_FN) return; /* spatne zavolano */

  if (ex->symbol->ident)
    WRITE7(ex->symbol->ident->name,
	   stream_name(ex->pos.stream),
	   ex->pos.line,
	   ex->pos.pos,
	   stream_name(ex->symbol->pos.stream),
	   ex->symbol->pos.line,
	   MASK_POS  ( ex->symbol->pos.pos , ex->symbol->ctype.modifiers ))
      //	   ex->symbol->pos.pos)
  else
    printf("*** no ident\n");
}

void output_function_dec  (struct symbol* sym)
{
  switch (sym->type)
    {
    case SYM_FN:
    case SYM_NODE:
	WRITE4(sym->ident->name,
	       stream_name(sym->pos.stream),
	       sym->pos.line,
	       sym->pos.pos
	       );
	return;
    default:
      printf("*** function supposed\n");
      return;
    }
}

void output_function_def  (struct symbol* sym)
{
  switch (sym->type)
    {
    case SYM_FN:
    case SYM_NODE:
	WRITE4(sym->ident->name,
	       stream_name(sym->pos.stream),
	       sym->pos.line,
	       sym->pos.pos
	       );
	return;
    default:
      printf("*** function supposed\n");
      return;
    }
}
void output_function_inline  (struct symbol* sym)
{
  switch (sym->type)
    {
    case SYM_FN:
    case SYM_NODE:
	WRITE4(sym->ident->name,
	       stream_name(sym->pos.stream),
	       sym->pos.line,
	       sym->pos.pos
	       );
	return;
    default:
      printf("*** function supposed\n");
      return;
    }
}




#if 0
void output_functions()
{
#define log_init_fce(x) \
add_ ## x = output_ ## x;
  log_init_fce(function_use)
  log_init_fce(function_call)
  log_init_fce(function_inline)
  log_init_fce(function_def)
  log_init_fce(function_dec)
  log_init_fce(variable_dec)
  log_init_fce(variable_use)
  log_init_fce(include)
  log_init_fce(macro_define)
  log_init_fce(macro_use)
  log_init_fce(macro_defined)
  log_init_fce(struct_dec)
  log_init_fce(struct_def)
  log_init_fce(typedef_)
#undef log_init_fce

#if 0
  if (!macro_use_on) add_macro_use= __null;
#endif 
}

void output_init()
{
  cond_output_typedef_      = 0;
  cond_output_struct_dec    = 0;
  cond_output_macro_defined = 0;
  cond_output_macro_define  = 0;
  cond_output_include       = 0;
  cond_output_macro_use     = 0;
  cond_output_variable_dec  = 0;
  cond_output_variable_use  = 0;
  cond_output_function_def  = 0;
  cond_output_function_dec  = 0;
  cond_output_function_use  = 0;
  cond_output_function_include  = 0;
}
#endif
